#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
#  This file is part of the `pypath` python module
#
#  Copyright (c) 2014-2017 - EMBL-EBI
#
#  File author(s): Dénes Türei (denes@ebi.ac.uk)
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  Website: http://www.ebi.ac.uk/~denes
#

#
# this module makes possible a
# dynamic data integration, download
# files from various resources, in standard
# or non-standard text based and xml formats,
# process them, sometimes parse html
#

import bs4
import os

import pypath.common as common

_fonts = open('fonts.css', 'rb').read().decode('utf-8') \
    if os.path.exists('fonts.css') else ''

_ebi_logo = u'''data:image/png;base64,
    iVBORw0KGgoAAAANSUhEUgAAAL4AAAA7CAYAAADFCTD2AAAAAXNSR0IArs4c6QAAAAZiS0dEAP8A
    /wD/oL2nkwAAAAlwSFlzAAALEwAACxMBAJqcGAAAAAd0SU1FB+ADCg4dOsFzszIAAAAdaVRYdENv
    bW1lbnQAAAAAAENyZWF0ZWQgd2l0aCBHSU1QZC5lBwAAIABJREFUeNrtvXecXdV19/3dp9x+7/Re
    NEVlNKMuVEAFdapEBxtjg3mMYydx8qY5tp/ksZ/ESRwnb5wnTvIkccEJwZjYBoMpoggQQgWhrlGd
    ojK9lzu3nrLfP869d+6MhCyC3/9m8RHT9zlnn7XX/q21fmttIaWUzMiMfByRgJCAREoBQoCUCAQI
    kFKCEAhsJIrzqwDC+dA3epFXm/+DQx1vU1/YxI7Fn6O+eIkzbHp4AGxAybpm+nticrDMD2Xm+3La
    T5kcZUZm5L8nNmALiUQACkIA0sootqN0KbWTChLLWSSpbyWNKG+2/JS9l15FCoWzg8384uQPmIiN
    ZJRd2M5HiZIZO712pBRZy2NyRUmpkDbp4ir3PaP4M/KxLL0iQUmptgMeBEIqCCEySidFtsKpU1Rx
    PDbMyEQ/FiYuXUMKm5HoAH1jlydVVHE+CgChAiZSWGkVn2Ld0zcmhI0QMnVP9hW3rs28vRn5b4uj
    7SkoM6nsKMJRNqmAAGFLUAQWFgdbX+Jc72Eq8mZzY/2d5AdKKfAVY1uSsVgYRRrkeXIpCc3CtA0+
    aN/Jma4DVOfPYU3DvXi0IEJqiDTiEUpm0WWuT1rZ1Q+99RnFn5GPZ/SFQCCQGcurOKBEKhlLLxTn
    k3dOPcVLp39Gwo5gXtrNWDzMrU2fYuP8B5BYnOn5gPK82Wya9yBul4d9rS/zzOG/R0iV93v3MxQf
    5N4bfg9dcZTdxkZIFYRECAkSB3KJ9B4kHOh1FbAzo/gz8jG9WgdnCyFACqRIwQ8hpqibaSY41Pke
    MXscTdVImHHO9+5nedVN1BYv5e4lv8mtCyLoqougp4ChiW6OXNxJ3Izhd/mwpc2RjvfYuuCz5HsK
    QKRgk0hjehBCpnzslIN9DZnB+DPy8bCOBJkJ08iUnVXoH7/MB22v0NZ3DABVdaEpOknLRNgSy04i
    FBeK4gIgGh/l8uAphicGnd8XGm49SMJMOgvHMnEpCi7hBgFdg+c4fPFVeobOY9kWQohURCkdXWKa
    wztj8Wfk16n66YAKEoTEtm0uDZ3mv47+I0ORfnTdy+1z7+amufeztelh4scidI23UR6oZk39Diry
    5nCqey+vnXmartHL5PuKWDf7TtbU3826hnvpG++gP9pDga+I2xofwe8OcPzyLl458ywjkQGCnhDb
    mz5DY+UaXKon4+dmudZwlYDmjOLPyPWjmqvGzUUWqhDEjHE+aN9Jc88h3JoXC5PXzvyUxvKbaSy7
    kaArl7H4EB49SE3+fJJGnA8uvkVz9yE0xcVorB9FUZlfegPzy1by6Rv/mHAyjEfzUFfQiCUN3j7/
    HO1DzShSpSt8kcILr1CWV09JqOYqEEdm7UWTUf0ZxZ+RX6Hrqf8LiZACmTbxEoQQWLbBWHQAt+bD
    78nFsg2GIv3Y2PjdPpJmkqHoAAkrQZ6iUVu0kEh8FL87CEJlZKKX8fgIlmWR4wkSNsYYjY8TS0RQ
    gyr1xUuJxofxuEMoQiNpGYzEBlAUF17FTdSOMxIdJGFEAUgYUcKxEYKeXNwuf+YJpru4M4o/Ix8O
    Y2Q6Bp+CC2kILRyMMzzRzQcX3+T8YDP5viJW1WyktmARCypWc7L3ANFkGKRg1awthHz5DE/08G7r
    C3SPX6TAV8jq2jspy61ldtFCWvuPMxIbwOvy01C0kMJgNcPRPvacfZ7O8Qvk+ovZOGcHZTlzWFyx
    loGWFxlPjOIVGg2lqygKVjIwdpH32l6lc6yNypxZrKjZSmV+wzSQI2YUf0Y+yiJQskymwLQSfHBp
    Fy80P8l4MoxHdTEU6ebx1bXcULuZuBGja/wCAVeQG+u349N97Gx+kjfP/pyIOYGUNnEjwV2LP89N
    9bejqzqXhlsp8pezYtYGdFXjnfMv8fyp7yOlQFUUEolRHln9VTbMeQBd9TIw0UWRr4zVs7ejKi7e
    Pvccu1p+Tsya4GS3n3AizIMrvoRHC844tzPyEaBOtpMowLIsVMVJCoVjo7QPnmXMGKXUX0o0GeHi
    cDt9oxepL1vO1gWPkExGcLn8AFi2wYnufUgpKfQUMZQYpX3gBP3jl5hbegNbGx/BNA00TQdgINzB
    qa4DWEJQ7C0kYcU53vU+9yej5AfL2L74CZLJCVyuAAC9I220DJ8haZkU+SoYjg/SNnyGofFuKvLn
    AWDYBrqizyj+jFxfyNIwE1wcPknPaAdBdy6zSxbjcQUo8JfgUdyMJ8fAluR7Cwh4CrCsJJeGz9E3
    fhm/O0Rd0SJ8eoDyYC29Y51EzDC2NMj3lxL05GPaSdr7TzEw0UuOL5e6gvl4NB8loWoYOEbYmEDa
    NnMK5qKpbhJGlPahU4xG+8jzlVBb0IjXE6LQX8IFRWUiMYEqNAr9JfjdeSSMKG0DJ+gPd1EUKKG+
    eCnqN77xjW/MvNwZuVYg53TnPn5y5B94p/UFTnbuwaVozC5eTJ4nn1hijIgZoyw4i/X1tzO/fDUt
    fR/wo71/xu62VzjWsQeP5qauaCEFvlKGon0YtkFZsILNDQ8xu3gphy+9ybNHvsOu8z/nXO8hVCFp
    KF1Frr+Y4XAHloCiQCV3Nn2aivw5HLzwKj8+9Pfsa32F5s73CHnyqCteilf3MpEYImElqc1vYOPc
    u5lVMI9DF17jP97/Fgc73uJ03weE9FzEDC15Rn6V/POuP+Bo70FcuodYPExlqIpPrPg9GituIm6M
    0zd6CZ8rl6KcKiYSw/zXwe+wq+0lCnxFGKaBR9P46m0/ojhYwXh8hKGxTgr8JYQCxYzHhvnPg3/N
    wUtvkOstYDwxTmmojt9e903K8+pJJMbpHm0jN1BGnr8Uy7b4i1cfoWOsE7/mZTA+wKLi5Xxq5Zep
    yJ9LON7P8EQvub5ScnzF9I9d5On3v82hngMUe4uIWglyPKEZqDMjv9riSwV0oYE0QQVLCOyUvbSl
    IOQvRVc1bGkhpIIiQVddgI0qJKriQtgmUpqoikpeqAwhtBT2TyKEQBE6Ugo0RcetathWHAATyAtV
    oQkN27aQ0kQTbjQhsKSFprgQippJ0CrCTY6/FFVxO98Sjn+iqW4s28alqAjxMeL4U9lwU2dLZoe9
    yKpTYJKiKrJn9grJKji46tcf4a1d+1uTP+DDfvjRL/trGUpmxRQ/4miZTTxNDU4NMX0kmZ3lvMZl
    VlRupGO0jeHYAAoai8pvpLawiYGxS7xy8kd80PkOFTmzuLXxMyyuuplFszZxpOcAccsAabOuegN5
    gXL6xrt47sh3Odd/nIpgNXct+TzzylfRWLKc1sGTDMVHCWleGoqXUF7QQN/YRZ478o+c7T9ORbCK
    +2/4HeqKl7CycgO9E11ErQRuxc3iirUU5VTRPtDMKyef5Hz/YeqLFnH7gsepKZzPkqoNnBs8TZIE
    muJlQ912RCwSlUKIjAcvP+wFpOZRAooQ6C5XptLGMExsy8KWNqpQ0D1uFCGuWAgSmSEVZc+xZVoY
    hpGZeF1XUVU9c2nbMjANa8qLSlf1TF2EThZRUVV0TZty/RSLCWmbGIaFbUk0TaDpGkIoKaLVr9dU
    Os+VnMwmpui5YtoCkVO1Fk3XUVXHMtm2iWGYSDvr2a9yrfStq5qGpmlTFTmt+NLGNJKYlo2qqGia
    jlDSBLMrFd+5XcfonO49QGvPEYqClSysvhmP5uO5Y//My6f+E6/mxbQNKnLq+J0N36YgUMK57iOc
    HthPsbeS1XU7UDSNf3v3qxy+/C6KqjGeHGdl5XoeWv7/UBaq5Uz3fs73H6U8t4YlNVtIJuK8cPxf
    efHcUxR4iolZE8zLmc/vbvlHPO4ARy++ScfAOaqKGllQuRrLMvjRgW9x8PIu3C4viWSMxeU38cTa
    b+DWvZzu2k9b/zGq8htYXnsL2u9+4QsZ5VEUBZniVzukH+lk6oSjqgqQSCSoq6/nN37rtykuLyUa
    jfLUv32fvfv243a7cXnc/Pk3v0l+aXHmRaZL0MRVTEoiHufdt97hP374JB6PB0VR+PRjn2btxk0Z
    fdn/3l6e+c9niEQmcOv6VKtlO0UGQlHxuDwUFhayfOUNbNi2FY/Pi0AgbIlMKdzQwCB/97d/x/mW
    VjZv3sj9Dz1AUXEpvzadl3bqvhUOHtjHj37wQ5LJJLqmoWZlEdNW2VFuO7N7RiIR7nvwQbbdfjv+
    YICTx47zX888w6VLHfi8bhQx9W/T78vj9hHKy2XxksXcesftBHNykekZTz1ceGyMnz77LC+//ApN
    85v49GOPMnd+w4c+iyJSpYLA/JJVzC9Z5eweimA40s1YZBBQ8Og+kkaEmBmlf7yDgkAF88qXMbd8
    BTJVqJKwIvRNdOLSAyiKJCD9jE0MMR4boiy3joaKm2goX52iGKuMmQN0j57DrwfQhYrQAwxG+xi3
    JlDxsqRmI0tqNpM2owPjHYzGhxFC4FNdmGqS8eQYQ+EuqgobWVC5ngWV61PG20T75ZtvOKvbtDBM
    M8WrZtIy2SmLgHSSDvE4S5csZcf991NcXkoymWT3u7vZ+fbbuL0ejESStWvXcc8nHsTtcWe42NN3
    kPQO09XRyfe+/2+8sutNAl4fmqKybNVy1m7ahI2Fgsr5U2fYuWsX4fA4ArBMM7MgHcvvjKsgEKqC
    5z89LGiYzzf/7M9ZfuOqKRTZweEh3tq9m6PHjuFya2zZtoWi4tJMhlJ8zCUgxCQku3i5gxdffgVT
    2igITNOcUo2UrbxpxR8dHaWqtoY1N6/HHwxw+dJlXntjF63tF/F4XBiGCQikIjILQBEixUeXeF1u
    /v47/8BfffuvuXHtGvRUXBwB42PjHDz0Ab98+SUGBwbZtHnjpOKLq8BXWyAVsCyDA60vcLJrHyU5
    1aybdz9FgUrKcmahC42xeBgFi1nuEqry5mFYBu+3vMjxrr0Uh8rY0vQZ8nwlNOQv4PL4K2BKokaU
    ioI6ikOVxI0oB1pf4lTXPirz57Ju7j0EPEFmly7naO9RNKGTsOMsKLuZXHcumEneOPk054dOUVc0
    n3Xz7qckp46yYBXn+48zEo8gJJT6KyjLrWciOc57Z39GW+8xaooWsGXBI2guXce2TcqLy6mfMxu3
    7kJaFgiBisCSEinsDAsvkTCYPWcOpSmLLgG3242u66m/tXnu+ee4bcedeDzua0SHHent7eXk8ROE
    vH7cuo6mutBVB6YoqQoaTVXx6hoJRaVx3lzKSkszu5MQzsJMJg2Ghke43N3JyMgIzc3N/OGX/4gf
    P/ssZeVlma1cU1RcmorP7caruVBSiurgYMGvE+VrqoKuq7gUjfpZtVTVVGNbNto0QJkdWJuYmGDp
    kqV4fD4H9mkqHl3D7VKpr6lm9uw6kKCiOn8nJZaUjI6Mcam7k8HhYVouXuAPfv/3+cGTT7JoyeLM
    LQmh4NI8uNzeDJy61iKUimPv3zr1NL849wwJI0y8czfD0UHuW/JFNsy9DwU4cultynJr2Dz/Ebzu
    APtaX+SHR/8WDQWzx2Yg0sPja/6c7Uu/iKZ6ONd/mMq8ejY3PEKut5hfnvgBr555mrhMcnzgCD0T
    nTy25n+xpeFhpJHgdP9hKkN13LnsN3GpHp59/9vsansFWzE52P0uE0aUOxZ8ljsWPoZP93Ku9xD1
    xYvZPP+TGJbBm2ee5rnm7+NRAxwfPs5ovB9NlSpxI0FDYxN/+Td/TV5uHrZtT2btpnlrUoKiKnjc
    7ikKLITArWrgcnH8VDM9fb3k5OdNh5mpzx3LOjI8zN49exgaHiGQE4IUJLnSigosBNFEkk999rPc
    fffd6C4XUtpTHDkBtJw/z7e/+Ve88e47tF28yH/86/f58v/+00mYjURVNSzHoGUs8BSn+2MmfCY/
    U7BNidvr5tZbb+cPv/LHRGOxa64vKSUulxvdpaegHFgSTFuydv0G/vpv/4aJiYlpPQVspC0ZGx3l
    T772NfbsfpfW9nZe+MUvqKiooKCoMLMzKCk8rwglXcX6ocEKB9xKmvuPYhsGXtWLRKF18BSXRs6z
    pHojtyx4jC1Nn0ZBoCkuxhPDnO1+H2FDQA9iCJvzA82MRgcpyankruW/hZWCg7riZjDcTcdIK4YV
    J8+TSzQZoXv8Mn1DZ5hVvJTty3+bO6SNIgWaqpMw45wZOoqmgkv4GVclbQMnGIp0UV3QyD3Lv4SU
    FkKoaIpOz2gbLb3HcOHCp+rYQuFoz340RUiwJSqSQMCP1+/7iGltga0IDGlRWVVJMpHkTFsLLz3/
    ApXlFQRzQlPVPQvrd3Z08uZbb+Fyu2iaPZsLHZeZiMZRpnnYtnBerhBg2SYenwu3++r3uWTZMv7o
    a1/h4NHDRGJRDh36wKm+TK08hRSmBlQhUFLRol8Xxs9e4Q7+tkGRCAUUTSUQDHykfcOBIKTu2fFn
    AoGrjxEM5fDV//mnnD33GUbHwxw7eIjhTw5nFF+mFr5UBCjKlAWYKQ5PWf10OxBQCLpDJElgS5W4
    MYEvWEfAlQtA3/gl2vqPkecvpaniRjTFRcBbSDQZxqd5iZpxij35uHUfApWOkfNcHjhFWU4t9cWL
    CbpD6LqXqGWgmwksy0RRFHw+5577Rtpp6z9BcU41DeWr0BSVoJLLZeMyQoGEEcPvysGtOfpwefAM
    HcPnKM+rZ07JMlyqF787RMSK4ZJuTMumyFOAZgmQiqMMtm195NickBLVkggkBTl5zJ49m3MXWtj1
    2k4e/exjBHNCU1tMZEl7aysnT5ygclY1W2/Zxr98/3vOixHTnSxSlsp2HFXr6lYqHegpLi2hOK+A
    1nCYcCzqqHaWRU9HsaQtP6xA59dh8FOcRicsIKT4SGHhyShmqnpUCieykzX65GY3CVFmz55Nbk4O
    aCrj4TGSRjw7EoCCRLGdghE7w1Wf+oqFSDvhjlHY2vhJoslRzvc3MytUx5aG+6grXsjJzr28cPx7
    dE5cJqgHubF6A/cs+xJr5t5Lz2gbF0ZbKQ+UsX3RE+T6ijh28Q2eb36SvsgAbkVnx4LHWD9nBxtm
    7yAcG6Rl4BjVefVsa3iIAl8ZzZff5anDf0fUiKLqXu6a90k2ND7E9kWPkzj6XbrHLzGvsIGN8+6l
    OFTDnpYXeO3MjxmM9VPgLWTz3AfY1PAgmxoeYDjSS89EJ1XBKu5Z+iU0ka6SB6dvSdakZsJc00Jj
    6fBgBgsKGykVTNvkvoce4PlXX+L0uXOcPXWaguKiDJbMfsm9XZ3se28PkWicebNns3rtOr7z3e86
    zrAtp+mSnLy8TcYBF1nhzDRV1vEJNBJmEgkU5uZloNWkMqbCq4rMKpv7OJGcLOWV0xdZanGltXba
    jiA+DPeIyWd3WtaIKWtUZhWAyGlzIS0byzApLC7B6/FNXYgyqxmHFB9q1yYbOUkq8+fx8KqvETfC
    aKqLfF8Z0eQ4x7r20DZyFq/qZSwywJHu/dxQeyvVBfN5dM3XiRtxNEUlz18KtsHbbS/QH+5AVVTG
    EmMcuvQms4uamFu6lEcD/5OYEcGtusn3lxKOj/Be24v0TXTjc/lIxmPsbnuBFbXbqC9dwufW/yWG
    FUNXXRQGKhkOd3Kk8x16whfxqD56wx0c7nyHheWrmVe6gt+4+S9JWnF01UO+vwzFxmn/oLp0PB5P
    5rEzL2R6kyoB0wvXZWo7TiQTLFyyhNk1dYzHY7y28zXGRkYys549THv7Bd55Zzf5+fncuGoVZeXl
    WKblvGNFuWLjkVJgS4E/FMTr81/hA4i0atuSnS+/QldvLx5d544d251lI6+amvjoSSGnhVLWP5nq
    35L6OmshZaZQCNwpn4irwItrR4lSda1S4vF6rraxTPn8uZ//lIvdl7FMk01btlBSWnbF9a43JZbO
    lyhCJd9XhEfzkeMtRNfcSAlRI4phJXC7XKAqxIwElm0AkB8ox6Np5HoLcGluQBKLjSGkQFfc2LZF
    wgxj2glAIddbgFfzEfIUoKkuDCtB3IhjSAtdcQMKkeQoEomq6OR7C3Erbgp8ZShCxbZtosk4pm3j
    Ut3YEqLJGLblhIrzfWW4hZs8TyGaoqEJbDRFYWx0mAN79hAMBrFsM2uLJRM6NE0Lr9dLbW0duYUF
    TGaEU7FyKdHdLjZt3MihE0d5+61dPPY/Hie/sNCp4MmKdZ85c5rW9gssXbqEzVu3kUgkruFkCRA2
    uiZoOXeGfXvewaW5sC3LccAEJA2DgcEhjp88wauv7sTj8XDPndu57a4dU6DOf1fhs+FEupA5/bkT
    WbpSDW0pUVQFW9pc7ujg+OEjxONxpwHHtOcUQpBIJigpLqaqth6315sZA+Hw0Xu7ejJjgJVJEhqG
    zfDgIBcvXuCnz/2M2HiETz70EJtv2YI/4J82j+IjLH6JEAoD4Q52nvkJl0bOEtRz2DjnHprKV7Oo
    bBVt/UfoGe8g4M5hcdkKynLqGAp38dKpJ+kcvYDfHeKWeQ8yv3wVa+q28+KpHzEY68OTyrhW5M6h
    e/g8u87/jPbhFkoDZayvv425patYUbuV1sFmxmJDuHUPa2p24NNzuDx8ip2nf0J/pJsiTzFbGx+m
    On8eS8tX0jvaxmCkn3xvAUvLV1EQKuPS0Fl2nn6KwYk+CgNF3DL/YTSEgu72cLallb/4y28hFAWE
    TMXvSXVMcVKviXickuISHn/8cW7bfmcGV8tUOwkpJZZlceeOHfz4J8/Q0t7OyRPHmVVfi8vlykz3
    +XPn2L37XSxp09TYyLymRo4cOuwoghBM582lGxa53B5e/OXLvPPuu5l2FiLl/ZmWRTg8ztDwMLa0
    uW3LNr72J39Cbl5uphPAdD/jel6+uCI7zJReLeIatAgHoSgYhslb777D8VOnUzvDZFIve+zIRITl
    y5bxB1/+MrNqazIDqoqCqml8cOgQf/hHX86E25ykqsC0LCKRCQYHB0lEY2xcu56vfO1r1NXVT7sn
    wRXb0XXInpbnOdD+KnErgmEaYJsU+stYPmsTHt1Lx0gbIU8+TaUrUBSNA+0vs7v1BVQ8WIqBZcSo
    zG9gRf2teNwh+iY6yfcV0VS+CiE09rbv5L22lzFsk8vDZ0mYccrz5rF81mZcipveSAdBTz5LKtaB
    Injz9LMcvvwOFgat5nE8uo+7lvwGa+fcRZ6vhL6JLooCZTSVryJmRHiv5Rfsv7ATXXXTOaZjWkk0
    iYIqBOFwlOHB01hpFGyL1AuSqTYlgng0ynBlBUNDg9PNYqYsTdo2c+bPY9H8JlovXuDVV19l5Y03
    UlldlZn482fPcfD9g5SUlLJ8+TKHjGSYjgJk+RxZbpzjaikqPd29dFy6RHZD0Mn+Lja2beN2u+nq
    6+Xlna+wY8dd5BcWXN3kX4OYejWLfL2+bXZDO0VRMG2bgYEBuju6Uws73fpFydy3IgTjo+PkBHOJ
    xxNTr5u69MjICD09PRnnfAodATBNE1VV6R8d4ZcvvcRDDz1AaXnF1HmU1/8saVNxof80tmUQ1EOE
    CdM1doH+cAfl+bNZWr2JJZXrEYqTexmLDXJp+BxJ26LYFyJpx2gfPEM0PkZJbg2r6m/FtOJoqgPb
    hiI9dIcvkTDjFPiLGU2E6Qp3MBLppaZoESvqb8Uy4qi68/sJK86FwdNoQsWnehm2xugcPs9YpJ9Z
    RQtYVX87lp1ETbUt6R5poXPoLFIohNwhkqbFuf7jaEJIzESSuXNnc/fdd+H1B1Isu8nAvUxNlGma
    BENBVq5cOTXqcpWOtDffvIG39+9lz3t76e7uTCk+xGMxmk+cpKe3l81bNrN+482TBkhM8y+ywYNt
    YyTj7LhzO0uXLMbt8ZDK32QWRiKZpKuri4Pvv8+efXs5fuIEvd09/OFX/tjhr1xpzq9p6acr/+jI
    CD/9yU8wDCOTOEuH/kzTICeUw4bNm6meNSujk5Zloesu1ty4mm3bbiWZTDr5B4TDk0lDIgSJRIJZ
    s2ZRWloy5eFt29lJl61ayZ3bt2fa9klFICwbhCBpmAwODrJ3316OHjvOkePHGOjt5ku///uUlJRe
    xWm+/thsTWETbSPnGYmPYkuT8pxaikOVmGac5q79XB5tIejNY0HpSvL85dTkN3Gw6x1G48PYWCws
    XIrPHSJmjHOiaz89Y5fJ8xayuGINQXeIskAFzaqL4fgQAoWyQAV5vhLi5gQnLu+hb7yLgC+PZZU3
    E/DkUVvQRF/HLiJGFNs2qcitJ9dfRCQ+zImuA/SFuygOlrGgfCU53gIq8+dyavA4I/ERVKGzuHgV
    GtgkLIPq6io+/9u/ia67r3NKUvZATs08pj/bcvst/Ojppzh69ChHDh9mwYKF+AIBmk8cZ9++/ei6
    zuKFC6mpm32F53yFRUp9nUwarN9wM5/6zGc+9L6iExF2v/02//xP/8T+gwf5wQ9/yJbNW1i+euXV
    F9R1wJz0DtDT3cPXv/4NYtEYiqpOSc7FolGqq6spLa/IKL5IEfDcHpUbbriBTz/+2EdIB2TPr41t
    WTQ2NfH455/40L8xLZP339vH//m77/Du/n38+7//BzeuWccd27dfScuU17d7Aayds52YFefS6DkC
    ei4bZt9FcWgWhy/t4sXj/0b3eAd+dw6dNa08uPx3WFm7lcFoN11jl/G6gmyb+wBBbx57zv+cF0/+
    iMHIAG7VxVi0n23zH+amuttJWnEujrZR5Ctlbd3t+N15vH/xZf7r8HeJJCO4dS/jkX7uWPg5Ns3/
    BIaA4Wg3+e5C1s29B787j11nf8Jrp59hODJInq+Aoci93Nb0KGvm3E3EiDIY6SM/WMItDQ+iSQsQ
    CrZlEZmIkpvnzmZic2VfEieMk90N92pSXlXJ0sWLOXmqmddf2cmGjZtoaGyk+eQJTpw4QV1tXQbm
    ZBZO2oR/yLYrgUQiSSwaxevzTXE80+IL+Ll1+x0YSYNjx44TjkT45S9eYMkNy9H0j1d+4Pf7Wbt+
    HYl4AkVJQS3pzFEikaSkuJj8/Pwr1qwtbcxEckqu4VpxfDmlv/VkUMBIGtOo2mLKItFUjTU3r2d4
    YIjz7W20trXxwYH3WblyJUWlJcjMf3AH9IoJAAAWTklEQVQ99UcOY1VSklPDQ8u/xEikF787F78n
    l3BsmOPdB7g4doFCXwnRZISjXQdYW3c7tcWL+eSNX6N//CJ+bwG5eg5SmrzX8hLhxDghb4iR2AjH
    Ot+lqXwl9cVLeSjn9xiZ6CfgzcfnDjIU7uJQ+5sMJ0bI8xSStBLsaX+RDfMeoKZwPo+u+DLjsUFy
    /IW4dT+DYx0c6dzHYGyQfF8+I4kRjnTtZ9WsbdQUNvLYTX/K6EQXQW8RXncQTRVO+01hZxsBkUk6
    Sa4MgUkxjfGa2rIdJuakbNiwgdff2sXRY0fp6e5hdn09p5qbGRoe5NZbb+WmteuuyBqKqzi3TpOu
    yWukAe6HYVWBYPHSJfg9XkYSES5euphhcTLF0bt+B08gqKio4G++/ddYltOyLk21FqlssNvtpqC4
    eNpfTdsVp/Hrr5q8mjIHKT9GTEK6NMCUU2ogJl33BUsXEQyFEKrg0sULjI+NUVRakllESpbfcO0d
    L73EbMZig0SMCQxp4tZ9KELBo3nQVY2kHUNi4tW9qKrbUYqL5ynsGkbkjiHnzEPqOm53AKnYJCyn
    +MSjedFVh5oxEhtkwhzHilm4NDeaquJx+VHRMSwLW1oE9BwUIbCsJKPxfqJmFDsqKQm5URQFr+ZD
    FYKkHUdB4FN9KEIFLEYi3UTMCHZMoLt8aHbaUZLiGs6anFIGIsgK7YmUE5TuX5ilrZu3bqHhqR/x
    6utvcPr0aYxojOMnmsnJCbF40ULyCwuvcAqn7CIp/rqTQbQnoz7C/pUvLTw6hiEthzPi8aXGSS9O
    x28R6QTE9dSyCFB1jVmpSMn10BZkxi47HXz5kAz2hy20tDWXmZ1QXhWKTI9XjY0MY5oGSInP67li
    pxNZ9IUrjFr6vWaZv66Rc7x45F8423eMnEARdy54jJU1t7Ck4iY6h8/QGe6iwJ/Pkso1VObWkTx9
    mL7/9RXi+86iVIYo/pOvErzjYdbX38lIfIDh2AgFWoDl1ZsoCdXR0neUV5p/yLm+o5SHarml8RGW
    ztrAqrrbuDDSStyIoGp+1s/eTsCTQ2vvMZ47/n+5PNJOeWgWdy95gqbym1hetZ6hiS6G4gOUhapZ
    VrmOwmAl5/s+4OeH/y9d4xeoDNayfdkX0JQU7UCooCriQxIk4kPxr0glJYWiTM0sShtfKMjixUvY
    9/5BXn/tNd4Phmhpv0BD0wIWL112BedHpuLhMmsnSXN10pZTUxV0Tb2m0owMDvP6ztcIR6N4XC4W
    L1uSYWFmLy5FFajXGEv86qKua64WkTV34iMyP7PDr5nwqKpe0zKDQyF+5623GRoeRlE1GpoWkldQ
    OGmwAEtK50CHaSeGXAm7nId5s/lpjvQeRBMqF0dbeOPMsxT5K1hUdTN5vlJa+4+Q5yujoWodIjzG
    yH/+C5HnDiHIh94wvV/+c7wr1nND7a3keovpGDpPaU4lc8tXYtkmu1te4FDnXoLuHNpHWnjl7NPU
    Fi1iSfVGcrx5tPcfpzBUx6KqdSANfnnye5wdOoNHddHcf5ics7kUB6pYN/duSoOVdAyfpyKvnnnl
    qxie6GbX6Z9ypO8IRb58WsZaeP7QP6BZ0rFGtg0jI6PYqUSUmAJmsiGQRFVV/H4/qqY506k4PdGz
    IUp6Ajdv3srrb+ziyNGjuFwuopEIK5bfwNLlS6dCGUDYNoqtpDKxWZt6KrJko2DZMDYaRnfFHa5N
    xjY5145Eo7yz6y3+9fvfQwBVFWXcft9dKKpIhUQnj46xbMH42ASR8ARJI3nV3Sd9f16vF4/Pe/1O
    adYYiiIwkgbxWJxYLHrNxWNLicfjweP1phZ+6ogdBLZpYZkW4+Njk/5X1kKzbZv33nuPp378NCND
    w9RVz+LGm1YTShEFhZ36l16QyodHskSawmJLwolx3MKNpqhYqk00GWYiMeb4crl1FAWrUBSBrqiY
    yQTJzi5UcgAnACDPm5hjI2hlldQULaSyoAFVqKiKi9HIIHEzhqrouBQPUrWwbJPx2ACFwTIq8xsp
    z52LIpyzswzDZDwZxqt50FHxaT7CyVFi5gQANcULqSpszPT/SVgxJhIjhHQfuqKjC43h5DCakBZu
    VeHc2dN87Y+/jCfFqU9nBe1UWj59xEsymaCiuprPffELzJ4719mKLQs7lcmchA7OrN645ibmz5tH
    a2srY4lxyopKWLZ0Kb5AYNLCCOeFp6uRrCw06+woDiXA69F55umneOvN19EVdcpJGEIRxOJxOjs6
    ab98kaRtUZBfwBOffYLKiqpJiygltrTxeTycPnGcv/zzP8Pv8yNtM5WOyD7CRmClYuP3PPQA2+++
    5/qBcereVEXBSCbZ9ebrnGs9iy0lmhRXDaoIYGIiwu133M72++8jv6AgQ4lQNIV9+97jc49+JkNZ
    FTgsU5G6Rk/vAOfbW4kbSbweD0888QRzGiYrrOwUN8lCYgr7OgI7NkJRmFeylNaRc8SNMKZMUFs4
    n4rcOqLxcfa1vcCRjrcpDlSxsfGTzApWEdy8lehPTjlli5rEfd983JXVjBtj7D71E870HaYqNJsN
    jfdTmlPLrLw5nO07yIQ5hipUykO1VOTMIRwbZvfZZzjTe4jy3NlsW/g4RYEyGoqX0Nf2CjFhYQqF
    +qLF5PtLGZro4d3zz9Had4yawiZunnc/OZ5CZhcvoXmgmYSdRLE1VlXeiGZYEoHK5d5eznVcTpUb
    Zm3NDoh3srhCEovGmDd3HrfceSez58514tC2jRU3nbrZFK5V0pZIVVi+dBnv7d1Le+clmubPZ/68
    hqlKkmJPGYaBpqqYpjllC7ZsG8t2rP7x5lMYx45PSY/K1PUUCSgKXq+XuqpZfO7Rz/DYFz+PhZ0q
    aXGshmlaSBQu9/TS2tmZTYFLZVbTDEWJZVroikZlff11Kb6YZv1N28Y0DZpbznL0TDO2BCG0q1Ih
    AMbGxsjNL2DD1q3kFxRg2c4Ylm3Tfvky51taAAUrdSaIFJNloxoabk2nuKCYzz/xOT796KNZtHDn
    9wzLxLIMTMvEtKxrOtYOHUSypfGTKJrO6a4DlAQq2dDwIIWhKl4++STPnfgeprQ5M3CKvkgPv7nh
    O4Q+8UVkcpyJt95Dm1VO8e/9L0Qwj1cO/R27zv4MG5sjvQeIyhj3Lf0tNjXch0fVaO46QGXhPDY0
    PIwQ8M7Zn/Pjk98noPo4PnSKcGyUJzb8BXcv/018ipe2oXPUFzWxqfET+NxBnjv6Ld5pfRFbKJwZ
    PM5IvJ/H13yDbU2PoCoarYMnqc2by7YFj6HdsnkTahZMmQIzJFjCTkUwnAlJJpNUVVdTU1UNgMvl
    YvXaNcQtWLK4CaEoV/RD2HL7LXT1dHP6zGkefPiTNC1dPDXAIaCgoIBbb70V27ZpbGycUvk/d34D
    WzduJDYxgculZ+pMlawoUJrTEvD7WbBwIbfdfTdl5WVO4UkqAiKA/Jx8bl67jvn1c1BVp4RPSKcT
    MFcJM1qmiapprLph+UcOf1ZXVnHHbbdhGgaaojqH/akKUlpXKH5aIpEIN62/kWBuDgBV1RVs2bSJ
    eZc78HvdmYSdojgGJ7tO2uP2MWfeXO69/z7KZ9U4OyjOswkBAX+AlUuWMritl4amJioqK6/um2R2
    LafmVlF1tjZ8iq3zPpV5X8MTPXSNtGHbJoWefCaMCGORYXqGTjO3fAV5n/86eZ+fDIubVpyWnsO4
    dQ1d8SAUje7BFgbGLjGnbAVbmx5ja9NjmXvoD1+ipecgfsVNwBXEg03rcDMj0VEKAsXsWP6lKe0h
    OofP0z3RgRAq+XoeYXOM3kg3XSOtzCpoZMeSL0xd4KZpSiGZhCpX4+HbU2m/6e017f1L23S4PYpE
    UbSrxqql5USQhSIm2ZfTKLrpMKGiOId6KaRfnI1t/+qqDaebg0OsUlQlGzhnaL0CMC1rqkM3JR7O
    FUSuNMpGEdfnpMrJGlbbsjNcJnEdDnLaeitCyYwhbTvTYeJDSyTS3RZSxmBK8i0rL2NbDqRThJqB
    iFJO5Q5dbdy2vqO09h2mMFRJQ8UadOHiuaP/xKtnfoxfz8Gwk5TkVPO7679FYaictoGTnOs+TEmg
    jMX129Clwj/v/ipHu/ejoTJqjLGych0PLf8dynPrae09Smv/MYpC1SypXks0Oc7PjvwrO1t+RoG7
    iJgZoSGvkS9t/Xt8rgDNXXvpGjhLZWEDc8tWYBoJnnz/mxzq2IVPDxEzIiwoW83nb/rfeNwBWnsO
    0TZ4kqrC2Sys2IiW5sor13ol6rW3dqEoSMUJOaZhUfZkCiEQmriyvkVMfWGTtAI7o4xKilGoqiIr
    YXPtzOt0xzTdYSHTgkNVs0gWk4eW/UqW5nUovcyibguEU4sgrp8KPK2KxYl0qWoqAiOv5IT/qtqA
    dFlmKv3hvO+pcylEqgZCTPXP0pN8uH0nz574VwYiPQihcdu8B7lzwaNsmHMXscQYJ7r2U5lTx5am
    hygKVHD80n6+t//rxGUcENwx0sr2ZV/k3kVfQFpx2kZOUZ23kFvmPUh5bj17z/+c55ufpD82QI4W
    ZP3gbdy/9He5rfERItFe2ofbqc6bzT3Lv4jPFeDtU0/x3Ol/J2zEcUmVh5b9Bhvm3cu2hk9gmAla
    B07SVLKC2+Z/Co/Lx57zz/PM0e9iAl7NzfZ5F7MbSn2c4jslo6ZXbpe/Ikworj7etb4WHwFnX82Q
    TQsSXtezfxSS2tW++MizK6Y9ueCj36eY+qfiKs8qssvcrmCvOjDpwOXXicTG8Gk+YmaME117aSxd
    TlPFGj65+is8aBuphJSficQYhy6+RMScIOD2I02TvZd2sqbhXkryqvgfa/8CSzjBE68eYDTSx8me
    IwxFBsnz5BBPxmjuOc7q2nNUFTTxuZv/yjnjCoHHFcAw4+y99Aa2ZRJU3IStMEc7dtNQsoLZJUv5
    jfy5WNJCESpe3U/f2AWOXX6LuJUk1xXEsm3euvDCzOFvM3KNRSTTETcwhYkqFEgxYNNBABUYnugm
    nphAEQoSG0VVSZhxVOHCEgLDMh3/RmhYwmJ4vBPbTDrus3DOvzVtAyltDGGQlEkU6UqFsnUGw92Y
    KWUWQsGWpuO7KSqWJbFswzkxHbDMJCPj3ViWOQmhkdjSREXBxsY2jZluyTPy4WKlQv1e1cPFgdMM
    TvTg1nQ2zL2bVbW30TfSxn8d/gd+efrHNHfvR5EGc0puwKN4aO8/xmBsEEUKbm98hIUVa7g03MxT
    7/+/7Gp5nsMdbxNy51GVPx9bJukZbaFnopMcdz6b6u5g6azNdIyc58n9X2dX60sc7XibAk8+pbn1
    qKi09B1nLD5ASA9wS+OnaapYw+nuAzx7+B94/dxPOd9zmIDLT33JIqSEtt4jjCZH8epe7l7wuZmm
    sTNyLQDrAJ75Fat51BOiZ+wyOZ486osWYkuLvRdf54OOd1FUhUh8iN1tMLf4BuaUreALN3+LvrEO
    Au4gtcULUYTCrtPP0DZ4Aikt+sKdvNfyC0pDVayouYVCfzkD4S5yvAXUFTYSNcbYfe6nNPcdw6t7
    CUf7ePnM09QVLeWG2m3kBysYjfaQ7yuhtmAB4fgwe9pf5XT/ETRV4ezQIK42D7MK5nNDzVYK/CX0
    jndS5C+lvmTJjOLPyLXcDMfZ1VU3c4qXMqtgAbrQEQqMRvoZifYTtaOUekuJJGIMRYcZjw5RkltL
    XdFCqnLnoKcKSCwrQff4JWxb4nX5idsmg5EewrEhynLrmVOylPrChSipZmID4cv0jl/GAtyaj7iM
    0DV2CcOMk+MpprFsBUkjgkt3SiuHIz0MRfqwbJtcdy4j8WEGo/1MJMbI8RfTULaK+uIlKVKcMqP4
    M3LtyJLTzcGJSblSTEqAoCeP2rx5HO18l56JHjzCw/ziRZTl1hFPjrOn5RdcHr1AwJ3Durk7KA/W
    sqDsRrrGf85gbADLNqkpaKIkt4bx2ADvt79G+8hZigOVrJ61xQmblt9Ac/8RBmN9KBbcXLkerzvE
    SLSPPS0vMDDRRbG/nJtmb6cwVEF9/hwuDp6id6IPj+qiNm8O+cHJYntddWcebEbxZ+T64qpZHF0h
    QVV1ltdsRgpJy2Az+d5CVtZswuP2s7/tZV448UNMaSBRCCeG+NSKL7Nx3gOomo+esTZyPQWsrr8D
    nyvErjPP8trppwknRnHpAcZiQzy07He5qX470hZ0jreS5yllw5zt6JqLt5qf4e1zv8Q046AILGmy
    tfFTbJx3Px49SMdoGxU51ayYtQWvHiCbzyVmTj2ckV+p8tmc7Ksc8FwQLGdTw4Msi27ErXsIeYsI
    x4c43f0+48kJCnz5JE2DYx172b5wiJJQDdsXfpZoYgyvHkTVdEYivZwfOMlQfIgCbxHh5DhnB44x
    GL5EVWETOxZ/nnB8CK8rhKbqmFaC4517MWQSryvAUGKYM70fsLx6E1UF89nW+DATiWGC7ly87lxS
    DRazwpczBzzPyHWg/Oz+qVfpoIKueSgKTZIAVaGT5y1CCknEiGLZJhXBalyaDyktLgycYDDST8id
    S13xQry6j4A7hEAwYUxg2yY5egCvO4iN5MLAMQYnegl4cphTvARNuMj3ldAV7iBqRDFtgxx3AS7N
    Yc763CF87lC2jk8SDzMnlsxY/Bm5ptpfqejTk4mZzg2pHcKj+7mhbhsXwu2MhLtx6162zr2fPG8R
    zZ3v8ULzD+gebSPkLuD2hY9xY91trKzezGi0n+5UEfpN9beT6y/jdM9+njv0T/SOXyLkL2J74yPc
    NOdeNsy9l5gZZyw6QKW7jtV1tznd2kjvSCClkmkDM/U50otg5vC3GfkY8N9Oab0yjeTXPXqBjsHT
    5PkLmVu2CmlbfOeNL9E8dAyPcDGaGGVR6QoeXvEHVBc00DfSTudoK3n+UmqLFzEe6+OZD/6W3a1v
    kOfLJ27GqQyU80e3fp+gJ5eLAyfpHb9AWaiWioL5aIqWOpQj3bslC9xcpa55xuLPyMfaEhQ5WROc
    OfYJSXluLeW5tZkVkrSTJOwkLkVDERouzY1pxTCsmBMl8hUyS3Ph0jyp7hQ2sWQUl+ZCUTR0Tcew
    DMxUY+OaooXUFC3MUm4rxf26klh2NbrJjOLPyMcz+SLVyTnrRBkrDYjSxD4Jbs3Lyuq1DJ3rYcKY
    wK25aShdSUloFn1jF3m75eec6z1KRaiGjfPupzKvgWWVG2kbOou0DXTVw9LqdQQ9odQCsxEZaqVM
    Kb0yhSQ4Y/Fn5P9HL8BhdTqOo1Mco6YP2hNKps0hNmye/yia4qN14ASVOTXcOOcu/O5cfnn8SXae
    /TFC1WgfPsdIYoQvrPsWa+fuACFp7TtBWW4lGxs/k+ILMQXKOG1eVCabR0w9Wu9q/MIZjD8jH9fm
    Z0jdGcs67UhHKSQ2InM4R7YMjXfzzOH/w+Gud8jz5BNJhikMVPC5VV+jtmTxlRdLiSWcc8VECmp9
    1FOcZtiZM/LxkI6cbEGThhO2yDrcQ9gIaaagjxNVzz7Uwu/LId9fiK5oJOwkQqgU+IrID2RFaWyc
    o5vSlxWpUlI52ab9ozZ/n1H8Gfl4SCdddCMdJXWsv5g80BsF0DPaNtmJz9FVj+Znw+y7WVm1GcUW
    zC1YwB0LHiXHX+KoslRAcejPSrpFBIoTrkTNNF2d2ozPnrIQrrYk/j+2JV8XG2BgOgAAAABJRU5E
    rkJggg=='''.replace('\n', '').replace(' ', '')

_rwth_aachen_logo = u'''data:image/png;base64,
    iVBORw0KGgoAAAANSUhEUgAAAL4AAAAmCAYAAAB6QSGzAAAAAXNSR0IArs4c6QAAAAZiS0dEAP8A
    /wD/oL2nkwAAAAlwSFlzAAALEwAACxMBAJqcGAAAAAd0SU1FB+ADCg4mH7bFiE0AACAASURBVHja
    tZ1pjGTXdd9/99631NrVXb1N9+wz5AzJ4SaapChKUShHkm3JRhLZhmIgsIPACYIsX4I4iJEgH7I5
    QD4kcBYHTpxESewYXglHtiVLifaFIiWSIjnkDGcfTu9b7fW2e/Phvnr1qrq6Z8ZQCpjprnq37rvv
    3nPPPed//ue0MPYFgBCC/O+D1+Azw/A1aCuEAJO/kjUY+b4UAp22H3xP5K7nP8/6zb0Gn2ltkHJy
    m/G+QKSjzo1l8BRmMEQxPly0Ngwey97D9jFok83HSB8HtLGfHjDO/D0Ou87IMwzGe9dnNZPnygwb
    jKznpJfOrfG9zfd+ERhpk34oJq1t7h7igHGCGPnyoJ0xJp3ng8dJ7jmc8YfOC3/+syje5HbHITQy
    +9wREk9JOlHMQD5cJZEIQp3YgRtQUuBLRTeJs+/6SrFQKFB2nQOFfXwMjXbAt76/yo88dxIhJ2+M
    fB9vXN3i9nobeUi/BkPRczh/cpp6rUCSaL726ko66SJdfM3yXJmHT9cpeA7dfszLF9fpBXFuCQ2e
    p3j2kUWqJQ+Ay7f2uHxrD0dJfpAvKeHJc/Ms1ksAtHsRb1zZZrcVZM+qjWGxXuR95+ZR6f21gZXd
    gE6YZAIzXVTMT3lMmqHNVsRuJ8ree45kecbHU6PykmjDbifivd2QRjcmTDTVgsPilMvytI/vSmJt
    2GyGNHsJAFNFxcKUh5J2zZr9mLW90MqVFNTKDrvtGD1JqeZey9Me682IRNt2rhIcqfkUPfvMUaJZ
    3QvpRxqAo9M+5YLCMRN2/STNf/m9v8fbvY9zpfsY2ljNNuN7hImmnySE2qAEFB1FQSlakZ0AAVQ9
    F0cKOlFMrDXzxQLPHZmjWCmDyXTwPuEdH8PmXo9/+9uv8/HnTtqFO+B0Grz/H3/4Nr/++Uu4rhw9
    riYcTtMVn7/08XN8+mPn+RefeYVrdxqodHNFsebJ8/P80t98nicenGdtu8sv/odvcmejPXJSlIsu
    /+xvfIBPfeQBksTw4leu8q/+x3cpF90xFXbQQO7SJr3uuYr/8Pc/wsfefwKAlc0O//IzL/Py2xt4
    jl3wMNb86HMn+Xe/8ALlYioEccLn3thhtxtlp+eDi0U+emGGasHZN5/futLg8loXEBgMtaLDjz5W
    58RsIdWwEBv4zrUmr95s0w01A20nhEBKOFH3+egjdTxX8vL1FhdXuoDhsWMVPvLQNCVPghDc2Qn4
    7OvbCARlX/K+U1W+9W6TROtUHhk70QzawF94ao7/c3GXTmA3VNFTfPh8jceOVQDoBJqvXNpjdS9E
    CPjJp+c5XSgiJx0J4xp/bef32Wp+nXn9q0y7bYQQlBxFmGhCrSm7Do4UFB0HgF6cUFASR0LJVRgM
    u/2A2YLPo7PT/MyDp7lQnzlwncfHkC0IUPCcycfY2E9jQCqJ40hcR+I6Kv1nP3OyzyTKkTS7Eb/8
    W9/nF/7t1/iJD51GSpF9r+grLt/a5ZW3N7KN4iiJ44hhn64kijUvfuUqcaKRUqR9qLQfieOo9N/o
    /fNtXEfhqnyb/dcdJRk/xJSSuNl4FK6SKCVG5rMdaLY7EY4UKGk16243ZrcTj08oYaxZb4QoZdsq
    KQhjw1ojHNEYr91s88r1Fv1I40jBbNVlYcrDc+y9b22HfPnSHmGsUVKghEBJYU8mMZBju24yvaak
    NVqUxM6jELiOPXE8R+A5At+RuEoghf2eTL8XxJrbOwFaZ+oUicjubS0Tg5PXrnltOfg9iNZ4b/u/
    EsYNZLLNycIX6elP4SvFTj9EpUdd2XEAQ5AYwkRT81x8ZbV/M4xwpOT0VIU/u7w4crIcZNYcqJqZ
    bJceZCoN7OPM3iQ1nA1DAVVWO5ULDt99ewPfVRxfqLC23UUKEFLSDxO+8foKn/zgKaS0/cmBn5Kb
    s8u39vjOW+s8//hStq4CkSrC3GYWIAe2u7F9DK5q7PgGfQ9+ps322fxDu3j4nIPLgxPUGLi13R8R
    lARDq5+w1405MTs6n+vNgF5ktbpU9vthrNlohlm/7SDh6kaPOLFzvzTt8uFz07hK8PKNFm+vdBHA
    Vivi5lY/NWuGKyiyGRoK/gRbFCnh9HyR+YpLogFhEMbOR6XgZI87sFS22xGb7YjFKc/eY2xOILXx
    DxO+7eaXieIdCt6yPc7FN3io9hPc7PhM+a61/aRkynVphhFhoik4ishojpSK7AURUggenqnxwSML
    hzqvHOQ8j3hlk4V9v28ihk5s+lNrw+x0gfdfOMJ8vUijHfLtN1bZ3O2N+CNvXtnmQ08u87+/dh3P
    VYjUvn3l7Q3eurbDmaNT1jvIfMrUeQP2WgGf/fp1nn98KTURrL1d8h0++swJalUfYwyJNvyvz19C
    itFNcfLIFM89egTXVUgBL721xpXbjezI3+/mMuHcNCPAweCzG1t9lLCacq7qst4ICWLDdicijHVm
    JgHc3g5IjBWw+YpLq5/QjBP7sxczVXTYaIS0+4ndlMbw1MkqyzM+AE+fqnJlvUecQD+yJ40Uo2tm
    cltgaMiYCT6N4MJymbMLxX3X+lEy9g3BXifmzk7A4pQ3Ni1DM8kxhzgP7f477HW+hTExQrjppz0W
    +F1W5M9hkqFGEunmlelxolIv3WA4Ui7y4eVF3AMc0nE/YxxlOuxkGBf+fUCAsBMsECTGsDRX5q//
    xUd57IE5okTz2a9d55/+2ks0O2GKNFl79sET08zPFNlrBRZIkLDT7PO1V+8wU/WRUqb3tMs3uEcY
    aV69tMk7N3dwHWuSaAzlosff+unHeeD4NMZAP4z5jc9dwohhH1obLpyp84t/5RkqJTvf/+TXXuLG
    SpPEpIKSajrPHQqp5yqkHBemoRlhjCGIDWvNEDAUXMXZhSK73ZhelLDViugESSb4xsCdvdBqeyk4
    u1Dk+lafZj+hE2g2WxFTRYdGL6Efa4wB3xHUy0M/oehJPEfSjxJEDP1Q4zpiuGZYGcmEO/1cyslr
    HSeT5XQoOvZUUhKC2LDaCIlTdG5cXQghhqbOfhgrYmX7N9hqfBFDTBTvWeXmLtMLf52a8zDvBE/h
    ypiSo+h1rXMRG0OcaGqex2q3R9lxeGZhlqrnTnSiJ5k24yjTgYjMATCoMWYUes0ddyLXp6skf/GF
    s/yb33yVZifMwZyCYwsVPvrMcX7985co+taM81zJ51+6yeMPzmVwqMmDlcIu3HvrLf7w6zcoeGpE
    BofPa0+WgemS08+pUziGson8cxi0MXzxO7dY2WpjDKxud9na61kTbOy0IzV11pp9eqE9NUq+4kjN
    o1Z0aPYSdjsxjV7CTNmuUasfs5OiOVLC8bpPO0i4tR3QDRM2mxFnF4oEUZIJl+dIXCUzs7LgSj50
    rkYYWXOy5EmubvTSjS7ox5qNVkgxPVEbvXhoBomJB3eGVuXXNb/OniOol11W9gI2WiEbjZCpopOt
    TXayGDM0dcaFZ6/9TTYbf0gQr1Hyz5DoHq6aIdFNtO5QT36VWf9X6CUSJSWNIKToOLhS4kpBbDRB
    ojlR8TlVrUwU2INs84OuC8FEG3/SBhJDWczFG8i08+D16qUNuv04m12DQUnBY2dnma76vPjlq8Ta
    ICUoKVnZ7PC11+7Q6UV26vN4fXr6dfoR335zjaPzZeuImoEuNiOayhirwUUutsA4uGDSzZiOTQhB
    Emte/Oo1PvuNGwggSTT9IElRqNG4ymAr3NoKMo1YLShmKy4zZYfbOwHtwNr5A6h7tRESpPCfqwTz
    Ux5b7QgwRAnsdCLixBDr4a0GJ/0gvuAqyYXlUrYmnSDh8novNSdhZS+gcTFGSrvxw0gP5d2MrqPW
    hldvtriy0RuJoZR8xbOnq9k0eY7k6IzHbidir5Ow1gwp+2qCbZiz8fNCH8U7rO78Fp3+FQQOQbhK
    wTuKNgFx0gKjMeYdTvi/yS3x87TCJgBBHFNyHZRQtCOLrZ6fqeEIOdGWv5utfzds/6CA1wCqG1q7
    QzNACkEUab766h1eemuNL71ym0Y7HOnnkTN1zp+sUy66PPXwAt94fRXfUxYNUJKvv75CuxdlDoTW
    Q5PPYLX+5Zu73Fxt4joCbQZO9X7H2wozGGNNrEmAukgNqaFcCLrdeLiRxo5zk/MgjbEnxJ3dwJ4m
    AmbKDgVXUi/ZkyzWgq1WRD9KKHqK93aCgf/PXMUidtNFF8+RRImh0YvZ60b5fYlSAkeJiebrwNzK
    P38/0vRCnfUhBzC2GVVcBoEG1poRa80oQzSNNtRKDk+fqo5M2UzZYWnG4/Jqj9W9kIWqOyGAaCbj
    +N3gGlHSpFZ+FoGDMRG10vvohbfxnKUhhOZuYdjlWlJASTvB9YJPQSnKkaLoOJyqljM8/H60+/0K
    //73ZlRlIjIHtOg7fPONVf7nH7+Tau4h+jFd9fkHP/cMAAszJT7+/pN8/bWVrF8prAObYigYY5if
    KbLbCtKI72AOI7pBbN9rkw8Vjwn1QEjMIUG2/ElnlzHSSSacFgbMIVj51kLQ6g0hSyVF5vRNlxxc
    JYkTw0bLBnmKnmK1EWYO9/JMIQs4lX3FXiemkfYnBxvOGGRqvu1fk+Hc5x8jD1sCJCn4oHMrZje0
    PRVjbVIzZ3Bya6LEjIY4jKHsKY7N+Ly71mO1EXKiXtjvMIsJqE4vvM3tzf9Eo/0dEtPDmAhXzbLZ
    /DyOqtELrwMaUBRNFy/5ZQz/gFbYx5WSnSDEE4JmFHF+uoan1D2hOAdh84fGe8Y0izkoBC9GNY9S
    gg89scyLX75KsxNmeDfAVNnj0dOzmdP47IVFHj5T5/LNXVxH5RzvgR1l8f4PPbHE/33lTqb1yKBH
    kdqX+WHkHVAzFP4DAhsi76Ck/b7wQ8c4tlDFGMNeK+Dli+tsNXqo7OQaCs7KXkiQGCQG31VMl6wT
    6ruSakGx143Zbse0A02sraM7CPEfmXJJUju+XnbYa0cpShMTJkPlEiWGKNJ4aujXxNpgUjNxEFkV
    whAn8MiREh96sEbJV4Dh0mqXL7y1l26moelnDLgOfPhcjbMLxawf63/Y2IXJrb3nSJanfSoFRbMX
    s9oIiLTJAmuDaczh+GBMzHbzS6zt/j6eU0dRJNIBSvoE8QbaBChZJk4a+O4S/eAqkm2W3O9xJX4S
    JTVaG1omQSBYKBVwxmgEh1Ej7uf63TdNTkpyQK4QaST23DyPnp3l1nqL3FxaG/71O/yZJ48CcHq5
    xgvvO8ZbV7ZxHTX2LMPff/YTD/MnL93CUc6E5xUTQch90mzuEl7OBav+8o8+xEeftZHbd27u8Iv/
    /pus7faw/qXIaXy4uR3YPZVi77/zyqZFmzSEiUYKQTdMaHRjNuIkQ1CEgT95azcbWhgblJLEiabR
    jW2ASYBO4wR6DIV58dVNtlq23eKUlznbpCanqwRuOl4lReZjGDFmnBhBpegwnTrfedSvFyYjSsZg
    N+jRGZ/Lq13u7NqTTI5ZmiOR225wnTvbn8GYmDjpIISg4B0lTloYHRMlDVw1g1JVCw/qCGl2qJnf
    Yd6PUEISJvYIUkJQygnBaFR1v2a+2/V9Aj3BTDoI78prWpMeqQA//PRxahV/KHBpUOu//9E72ddL
    BYcPPLbEkbkySWJSxGhosw4U9TOPLPLwqTpJrPe1YcyUyVM0Bn1gzEQy28A1MLnnEGNktEHgR2QR
    bpM9dxAZNpphei/7/L1A0+kndENNkpg0GAdrzYCrG32L1KTj7AYJ3UDTDTVxYg0RKQV73Zh+ZDLz
    Jk50dgIIYaHjTqDpBAmdIEEbMsLcgUpggPOPoXJmgiWw34IYXi95iqPTHgjBXi/OeDp5v0cOWYQh
    u+1v0gtuUfCO4agaBe8ERf8kjprCc+fxnXlr+9Y+gevMMlV6jHLhPFNqj3Pl15ktlpkr+swWPISA
    GLOPgzMJqsw/zGFQpjgkcpvv/8B7ZMiDSAX/GKeXp/adLq9d3uDd23vZ+wtnZ/nAY0eIE51h/SIX
    SbUbxOUnf/hBosSMtAEbAxCTIq1C5qKxB2j8VIBH52ryopNjSYr0WXd7EXvdOBuTryRFT1LyFRVf
    4jsqRXskt7YC1ptR5qx7jqDkKQqupORKip6ym8xYqkOiDb4jwQi6oRXywTh7gd1YUgiUkBRcYXF8
    xMDyS83AHPiQMVVHsOcMKcqfnuk+zp5dIEamb37KY6bkYHQ6v7lTN2NnGmPo9C5zc+NXSHSbWLfS
    ULGH0l3CeIvE9MEYPGeWXniLbvAuie6lbDofKb5IEJ1kLZxFkOBLyX+7vcrfKvicK5cOpTtn2PsE
    R5VDFn1SoGvcuU117/Cz3ORUSh4f+aFjXLy+Q5AemQDtbsxvfeEy//CvPps6uUU++PgyX/neHeuw
    5qDLvG766DPH+S9/8CabuymmnkZPjB4HM/MR1mEvRkxWgyYXmhrf/ONMU7uZTGZODGBMAMeBH3+y
    zvyUZ30dKXh3vcfXLu/RDTWNXpLFEhJt+PD5Go8sl4cQdzfmC2/tsNGM6IaWe+O7Anp2LS6tdqmX
    HZQQXN3q0w0SpBD4nmS+6rHVitJnNPuDbXn0SpghhSc9pdabEZ7Ts5SFPHav0hkcm7v5isvStMde
    t4s0YuR6CggItOlyc+Pf0eu/i6NqgMZRZbTuE8VbSOkjEEjpoFSVvfa38Zwj6SRJhCxjwm9Q54/w
    pUAhaEvBf7qzyn++eYdgjHWZ/3nQ7wefAGaflp8YwBoDdQafmTGz6FMvPMBcrTi8rg1RnPCVV++w
    tt3J2j39yAKPnK6TxCYLpJg8AgEsz5f5kedOEsU65zdNUs1DLNqORwzHOVGcBzcSWfR2kvkvTM58
    SjXgzZ0gG6MrJcfqPoVUe3uOZK7ipMzM1BlOTzIpBCdybQuupJZSibWxzqoxhpmSg+vYE+PNOx2+
    8OYuX7y4yzcvNzM6xlzZ4XjdJ8kgTTHqtI+ZdYPo9NBsgu9ca/Fb39nkd1+x/37vlU1+55VNXrvV
    nkBrsRj/YkqWM/ttTGvjb7e+zNrubyOkAhPhyDJSumjTI4w3kEIhhMR35oniTSAi1jt4zixCuAgM
    YbxKlS8z71xEKY/P7TboRzGdJNlnf98PonMvbScGsFKMPL8JJjFPF2dLfOiJZZSSWcTXGFjb7vK5
    b9/M2j14fIZnLxzB91WGTubtZoCCp/ix508xVfEzwcBYZGP0tiKHbZssaDXpWQ2MXDdmEL00+9is
    Ojf+QdBoqxVl0c56xRk5XUlJXhVf2e+mfRtjqBZVRgAb3MN3JXMVi/JIAZvtiAcWizy4WMRVFrW9
    vN7j4p0O/dj6A0dqLu87WWGq6KD1UPFkPtCYXsg2L6PBRkcNWJnCRonT3x0lcsHB0TjI0rSNTutM
    6Q03nhPFLVa3fo+CexyBg5A+vjtPknSJ4h1cOYsQPqXSWbSWYFZw5SxgKBZOY0xEEK5RdM8hRcgZ
    7x1WwvM0jOH0VJmfOXoEX8r7wu3vB8Of1G4QwPrIU0epVwuZA5Zow/JcmfmZ0sj3fvaTDzFbK6TO
    23DjzNWK5FOtPvH8KQqeotOLckQrC4EO7nvhzCx/92eeZHOvn91XG8NUyWO2Vhgm8SjJ3/ypxzPH
    dIA6PHyqPsLD+fATy9RKHlGiM4FWSnB6uZa1mZsu8lM//AAffHwpo0IYbTgyP8Wp01WkkmhtOFLz
    9/kZZV/x2PEys1UXKYaPWytaLT9yugInZ32ef2AKR9nT9tiMz8nZAsvTfmYCaWPNoHrJ5fSCvRYl
    mvNLRbv5sChPwRv2P1t1ef/ZKaQAT0mWpj2ePV29awrDfMXjSM0jiDWekkOKAob5qsvTp6upj2Of
    azoN2olGZ8f8y8/+a9qBjey50mGmPE0v6tPstdKUQc2x+nE6/QaNXifTNnOVGTy3xO3t2ygpMTpm
    cf4B3p1+gj9px/yjB0/x8yeP4U4Q9IOQncOuX7vT4Bd++Wv89i99ckjKOjCAde/Q571srnvlDR30
    ihLDXjdGa5NGRN1UW9nXXi8mik2mWW1MRTPQGTMlFzfXvtVPUmfS8mIcaR1MUh67NoO4mcFTIoUC
    TeZA9sLE2vQCVEou1Gb0ZNJjz+RIQcVXtIM4s7UHEKQQFu40xkZPHWU58xliGCaspWzOxBibZVV0
    WKp5uI4EDN1A0wqSlItv4c5enspwl5dK+xxGkKHTT+jkIsQzJQfPkTgFb4py6QV+87vfoui6FDyX
    qu8RJR7toEgQR0yXSrx8S1MrLvDeXgOjDUpKjtdrtPsB3WiZVq9PtVgg7tVodht84Ogin1iYy4T+
    fl9/Ghx/8gkwRn29S9bWePvJ4xhDViaYcuOc+Tu7fV6+3qYfaRJt+PPvm2Om7KRUD8233m2y2Qox
    Bs4vFcEILq93s1zUx46VefJ4BSntafb2Soe3Vzs4UnK87lN0JW+vdm1mXMkliDX9KEEbwXRJ8WOP
    zzLYN4k2XN3o8cqNNlLCVMFBSbuZtDYZOiRyebzaGOaqLo8eLfPy9RadMEEOroshBu8qwVzF5amT
    FXxHoo3h9nbAd2+02OnE9GMb61FSUPQk81WX585MMT/lcWWjZ212Kah4knrF5cZmenKOpxSbsXEK
    g+8onj5V4fR8MduIr9xsc2urD8JmZ/3YY3U8BxzPUfzcB5/i6+++w9pek4ovaPXbVAo+taKkGwjK
    Hux1+3RDzXKtyHqzxcnZGuvNPTCGeqmEjg3F6TJXfBfZD/jrx5c4WSoemCV1N17+ZKrxZFTnIF9g
    Mip0r9fNxBTMPL/9sOsml4I30N5b7YhuYANE1nSx1+PEsNOJWG9GGGNYnvYwCDaaUSromtduah4/
    VkamUd5mL2GtEeE7kqmCIvIV640QhKDoKvqRpQ9rA62epBcmVH0FQhDFNg91rRGiJBQdSQCsN8LM
    yU2SAT9omOSPgWDBsNWKaPXtaRMnJpPHQfBsdS9ktxPxiSdm6YeaL7+zx1Y7yjaHEoIosRtztxPT
    6if81NPztIMkHZMgKNp8hJXdAKWGiT8jJD+GhQyEsDyqG1sOp+eLgGCrFXJjs89GM0JjeOqE3YzG
    pFydpdoUf+3PPs8//YPPEUQxUZLQ7gdUCz7z1Qq73R6x1jR6fSqFArPlMt0goBOESCFxgoCluVlu
    Fcus9vr8/OnjPD1XP5RPcxgd+fDr5p7py4fRne/l/Q+qj7yNnOfV5FmnUggkNgoqc1FMkfoDO52E
    9WbE0Rk/4wypLBdi2LdJHcGFopsmfhv6kaHRjammzmqUGLY7MW6aEVWvuPQiCz3qlJG5POcNx5cK
    2UzZngwi9UmkFDx5osJC1WWvn/DOSodO33qwK3sRb7zXQQIbLZuMNF1SPHWywvK0x9urPS6udAhj
    w+peyPXNfpZ6OPg5V/U4v2zvE0Z2wwUp/39+yqFatP7CRjOk09ckmtTPSCilnKNGL0lNSsGjx8rp
    3Kc4vhSCF86f4/Un7/DVS1eoFQtoYzg6M43vuCi5S5gkmQHwyccv8Nrt91iarqU2q2FuYRFRLON0
    ejx/ZJ6ZHP/+MBPjoOT2e8HxD+PkH4b1HzSuw/q433vszx4zQ36PMAfQFgaoxnj2ub1wcaXD0Rk/
    QzrMRODT3nthyuP6Vj+Npho2WxHH6jZJvBsmNLsxCIFUVqBNx66jNtYk+PiFmVxiiu3XUYLVvSAH
    4xpOzPqcXbBw8Mm6z2df2yGIbVbUnZ0glyppOD7rc+FYGd+RlAuKWzt9tpqRZV82wsyZNhgcJXho
    qcSzZ6oIIdhsRnzx4g79lvURLhwtc+FoGTC8s9bji2/uIhE0ejF3dgNOzhZYbYSEKbT84GIxdWxT
    f2WwODOlIh+78BB//MZFdnux1T7aUC0W2G53CGIbh52vVPjW1eusN1t0Apu57pdKvDczz6X2Lhvd
    Ln/nG9/lxRvv8f6FWX723BmOloujuQUT4Udx6PVJGv+wRJSDHOfxNne7fpBA3/c9Mv6MGIFX95ls
    OdqOYJjSqKTk2maPfqSzRG4xRgEatE0MLE65eErQS32NtUaUwVWdIKYb2ShrxVPUig573TjjMgkB
    vqtGnOnxewyo0JkJIgQnZgv4riBM0nFom/kl03VTUuCkoETJU5w/UmK+as2xmbJDN9Q2hpA+lKNE
    mtxiA1UqN4eOFHhKIKXk/GKJl6426QaW6ryyG1JwJGuN0OYWJ4YLx0oUPJWxOGX+gR47tswnH3+U
    fhRT8n0inbDZalMueEhpE05cR3Flc5PZSpnEaJRSyOkZvtfpE+uEiufSiSK+u7nDP375Db67ub2v
    Nsq92uZ3w/EPSz0cL1Mynpt72PVJfRxGk7hbH6PUWfYVcprYJhVSrQ1TBYUEOn3Nze0++byV8cJM
    g/jCTNmllGWAWR9Ca0OSaHY7CUFkEaSpoqJaUMMA3oTNuA/dyjEn8+dNnGibe5AKlu9aesRgXHd2
    +ry3E0AaNX7/mSqfeHyWH39ilsePVzI+0licb1TR5a4P1tJ3BY8ul0m0IdaG9WbEpbUeja6ty3O0
    7jFXcTOhF6lZmb1qxSI//sQFzszNZtUTmr0+jlQ4UrI4VWWj1UIg2Gy1WKxWcAsFtkoVOkFIbAwF
    pZgrFtgOAh6t13h6fhZ1F5z+IEG/n0SUSRsiT4OYpMEPun5Qm/u5xz5EKA3aIExWDGs8mGT2ZWkZ
    EuD0XCErkPT2SmcUUDqAvQ9wZNrLTpu9TkyUGDSC7bZ1mk3Ks68W0qBcqsWjxHB9s8+V9V7279pm
    n81WOFIlASGyKG6i4c07XXpRkqUsnpkvsFizkV4hBdudhD9+Y4cvvdNgqxUR69wmy4MCYjIdw6I4
    wzTSwbxJIXhouZTl9K41Qt5e7WQcn3OLRSq+GtnU+xJRHlle4tPP/hCf+ca36YYBRc9lr9vj5NwM
    /chCWJ5SxFpTLZUw1Rov9UOUknSjmHq5RMV1uBbHhFoTJMlI+Y0fZCLKYfb4QRygu13/Qfcxqs1T
    gtZYPni+PAhilKFstIUR93ox3SBhdS+0AZlBGqmZ7C8k2rBU83hDGaEiuwAACcpJREFU2LzcSBu2
    2iH1sstWK8qCVVNFh6KnMlqEENANNJ99fXsfLeLC0TJPHCuPjPn1m22urvfY6yasNy0q5CrB2fki
    jxwts9OOeHulY1EgBN1A890bLd54r8PZhYJ1jqe8UbPK7GN9p2k/o9dFjsBWLSjOLZZ4606HJMu9
    gNmyy9EZf181O2dcyIqey49ceIjrW9u8+d4KrrIMwoeWFrmz26DkDp3Wan2GaGaW+uqmraoAHCkV
    mPV9ikrRSxJutDocr5RwpfyBJ6L8/wg23a2Pe7nH4WMXGKPHiDbjubij30g0PLRU4r2dgH5kuLbe
    yygZIA/M2lqetiX6BgnhG82Isq/YTdMGy75K0w8HQm8yNMmRo5lcSggcOc4aFdzaCXLZafYbS9MF
    PvDAFFJYZOaTj8/y1UsNdroRYUoRTrR11i/e6fDM6SrvPzuV+QxCmH3HmclIeqMlSAb3dhQ8vFzi
    0lo350PZejz1yjB4l7EzJ2mxI9M1Ts/N8uL3vk+YxCxOVbm9s0etWGC1Ycu6ea5LUqpyY2UDYQyb
    /QDfcXknaeGpDq0woheG/NJrb3F++jmWy6W7autxTbnPzrxHATusoOz9RnQP5n/fPUI8sY0YDaaN
    05EngQBaG87Mlyi4Dbqh5vJ6j8UpN4eYMHJiDL4zVXSoFBT9lk07XGtELNY8OoFGCRuFHVRWGHBF
    wUZ6Hz5aQo1VpzhS80aSTQyGYzM22ymIbdygG1o/5Atv7fHJx+uUfFtv888/NceltS63tvtstyOa
    vSR1dAUvXWvZSK0jJqzz/mChmbAWEpiruhyv+1zfsoWzKgXFibqP74gsYj1YT+cgaO9jFx7iW1ev
    8fL1W3hK0ez1MRhmy2U2W23qi4t8JTYYnbBUKtCOYyqeIko0u0HMfMGjnyRcabTpJ0nOhpvMvT/I
    th+HBPdX5P3T0wgOih/cDcu/1yDYvvRHMyr8Y2mo+5vmyFoFV3JyrsDFlS5b7SjjmJt9iRyjfS1O
    eeykxVfXmyFbLS+bw7KvmM7VwhHGoI2g6Es+8tDMxOe9sxuMHFZPnarwwEKRKDF8/3abb15pohGs
    7Abc2gl4aKkExtbZefJEhceOlVnZC7m40uHKeo8wNniO4O3VLg8uFrO4wf4sNzFC2puUxFJwJMdm
    fK6s9xDS1vmZLjsjkz8Qfpm3T/PCuDxd4yeeeIxziwu0+gHaGLqBhYeW5+rcKFUJo4hYa5phxMlK
    GUcIWlFMFNsiskdKBZ5ZqDPn+xOTRCY5g4dWbp4QwDITEJKDqqsdlKgyvhnvp4/x02nSPfKRRovn
    5xKn8xtkzLY1DPvVxvDIchmROpObzShNCpnMnBj0aXH/tNp0L+baZi+tSWkdWxvJHIVrjLFVhsfZ
    n4wyvTPHUgjLmHx4uWxpwGla2s2tPq9cb/H1dxt8+2rT1uKUguN1n488NM3ytJ+VUQxjY3k6QhxQ
    aWLogwgxSWasbDhqWKZQZubXfl/MOUyb/blHznNrZ5cvvPUOURrAKroOZ8+d4+p2k7M1BwOUHcXJ
    SpkrzRa9xB6jsdE8MztH1XUpu85EDXo3usK9aPzDosM/uIoNd39/mIM+IFDJbH0EvUCPJGBEiU4r
    0Vn8Oo5N3irCGGuz1ysOW+1omCc8OQU4ey1P+yhp/QRj4PZOiEyzq+plZ2Ifg8jsQZDsKKtpaFB7
    jkh5NbbFbjfiykaPTqDxXXvPxZplsnqOpFZSiJ3hMSXvRkcT5vA1E2JsE3NglT7nsKhqwXV5/oEz
    fPXyFTZbLRt4WDzCFxtdJILNfoAUkGiXtV6fXpwQJja/cr7oc6XRZqPX55+/epEXlhY4PVXGlTax
    YD2KebfTZTMM+XNzdR6pVibyeUYebkL20b1EYg+z3ydr5/uP9t7NCa6VHCsUxmqiKxtdTs1bmvJm
    y6YHSiEyBmRTJzBSTcdungcXi2y2YuvTprUtx7Vi/v6VgqJaSANUxlY+II3OzlbcMVvLOrSxtlWR
    B0jIADv3HKuZ96EsqdCJ3J/BEAZ8R1FwNf3IEtO2WhFRonGVpBdptlMTzGDwHEmloCZEJyaYg2b/
    iS0mPMdBNJisktphWvCJ40d5/Ngy765v4EnFVdfje9u7HK+U2QsClBAUHcUbO3tU0kpq7ShGILjZ
    7jBb8PmVi5f5tUtXeaI+ze1OF0dKpmanebXb44X6ND+9tHhXngtwqD64W/BoUpu7Xb+XiPK99EFK
    Q64VnRTWg0trPcJkC0dK1hthRvaqFh0Wpzxa/d5EhOeBhSKv3GiltOAJJVbHJsmyJa3gD+p7amEo
    uWIo+HntKAS9UPMnb+4OsnkyE61Wcjg5O8hvGKdNmH1/uEQKWJr22WnbimlXN3r0Qk2loNjtxKyn
    lZcTDWfmixRcmauDP2HBTZ6aMUH5CZMby8E8MCHE5IJS46+/8sHn+Mblq8SVKl/qRxitudPpcnqq
    wkY3oB8nBGGEBDylWCoXaUURYRyz2tUcKxdZ6/ZJ9C4rzTYoxWIcMVWv8amlBY4U/Hsis02Cre+F
    V3MYKnMYTfl+uDuHU52ttn7mTJWt79u6NVob3l3rp9WU7T/flTyyXGZ5xuPSapfEpAWWcqmT9YrL
    fNXl9k6QkdIGGU2DmjN6LIVxadrnndVeVhteCFtGsOzLEYRGp6mXWsBOJx6rRWqDVUvTGm0MyXje
    JYNkcki0RgtJEBuePl3mxlafdj/GRIZrm/3MBjfGlno5Vvd47myV77/XIdFpyZIJNbYMgkTbsiiT
    wJDByRRrUJkfNVlBOvcSCV2arvG3P/oC//H2Gmfbvew4Olou8li9xqW9FvOp8C6VixSU4k6nR1gs
    2DonvsfzR+a4uNNkanbamkyOy6ePLvEzx5buGQuXQqQFXLl7WF0c7gvcL45/GEfnbmMfvD85W+BT
    PzTHKzda3NwOsvo1roDFKZ+nTlY4NVdACkG94nBm3hZQqpVU5tApKXjqZBUlJFLaOZmf8ig4kjPz
    RYyBhdogacX2f2quwJnUrDJpQsmAujvQJLMVl9NzBQ7yogxY+LPkcLxeoJ4mwdiCUMP5uXC0xEpa
    adlzJI4SfPrZeb5zrcmt7YAgNhlprVZ0eXi5yKNHyxRcyUzJ5cx8ASFsBpij8ieXTSUsuBJt7Mk4
    aV1qJcXZeTuHR2peFnvYB1CYewS3B+HzSX+Ia/gXS0RmEI5O4FiQJv0v+8sY9+icamOyJIb7cUL/
    NBHi+31/P98xBhJj6IdJRrjyszS/YZt8UaiRevditJTGiLOZIipy39zlZj+lEIhJqNMhDnr2dywG
    jQflWib0M+TJDz+PtSGM7Inhpn/BZbyNGYsqiAkc/CHAN2ENRkR0fx+D1/8DhDCvobavS0IAAAAA
    SUVORK5CYII='''.replace('\n', '').replace(' ', '')

_ei_logo = u'''data:image/png;base64,
    iVBORw0KGgoAAAANSUhEUgAAALUAAAA7CAYAAAA96csBAAAAAXNSR0IArs4c6QAAAAZiS0dEAP8A
    /wD/oL2nkwAAAAlwSFlzAAALEwAACxMBAJqcGAAAAAd0SU1FB+AHHRQOHFtumB0AABJxSURBVHja
    7Z17fB3Fdce/V5IlZBtjG9uIIAw2BAyGuDwaSFEgIUDMo7SZoaVhaIBSSoACIQGSJiF8GvIgtAWX
    RygEEpN0iQmZDY8AxoFSiHi1gYTwsIlfGMsgxwZjGVu29cofcxaNlt37kO6VhLTn87mfe+/u7Ozs
    7G/OnNecyTU3NtQBuwPdDE9qa2ppXU9GGRVJNcABwG+GcRtvBs7PXlVGxVIV0DPM29iVvaaMSuXU
    I560MjOAsQOoogdYasOgY4jaXwvMAHIyyFfZMNjuna8WEXIHKbPGhkFbBuqRTTcARwyA63cDhwKv
    D1H7dwMeA6qBTcBxwArv/HjgdhElxwJ/D9yTgXpk007AxAFy6qohbH81MFXeV13Ce8sBk4Fp8n+H
    TPwYXbQdWFciSLuHgWzfM8DzGahHMP0/cCSwY4mAeneIAd2TAToDdRp12jDoBjYOUHnDhkHJ5ZKu
    08oAJNYXnSuRqhLqydkw6Cm1/ZXol6S2RM+f0F99ymagrowlAhsG73W8VuZoUSL3BMYBHcBq4Ekb
    Bo/EX5T8/4gorjlgiw2D+V79c4GTgEbgHhsG8xOA3VMMt9bKNAD/IO0bq5XZBrwC/MSGwStxEMYH
    l1ZmOvAJYH+R16uBd4CXgQdsGKyJXyPP1wicIqLeNuAOGwadWpnPAZ8GJmtlOoHF0pYX/bZoZcYB
    5wGHA+O1Mh3AEuCnNgyeLzRwMlCXSNLpOwPfAC6QF502AN4E/tKGwXMxAB0NXCfF1gPztTJ7AL8C
    PuxVsayfIkanVuYUYEFC+04GvqKVuRU4z4ZBd8JAnQtcDcwpMMBvAT6fALD9vOfbADyvlbkXZ3b0
    6STgUq3MFcB3pS0G+O+E20Vl59kwuCTfrFCVwbR0Tg18G7goH6CFdgV+o5U5swAA9wGeiwF6IO/n
    fODuAu37J+DWhOf7MPBQIUALnQv8TiszJU+ZccCDCYD2LTffAr6ulTknBdA+fUErc2c+bp2Bungg
    +9PsXZ41ZBUwDzgL0AKoR2Pc9QatzCEpL6EWZ4PeOXb8VeCNIjl1XJE8Sr7bgCuBY4ETgJti152t
    lTnJb5cNg6VAs1fmQeAy4FTgs8LBfaV5DvAdcQClPV+D/L4TOBGYC9wSK3cl8J/e/58Kdz4W+PdY
    2b8V8aZi4kc7sEVGXCVoc5nrmyAybTEexhzwpg2D12IiyGNamYXAVTYMnk247matzDHAw1LHeOA0
    rczzCUrPZO/3KuBiGwb3xgZUf6wbi4EjbBhs8I49pJW5G3hAOCgyIH8Zu/ZqGRhfS/CiLgD+RStz
    v4AOAdjXcKbStEE314bBIu/Yw1qZRYD1GGx9VJ8NA+uVfUQrc5vI8tXy0VqZe2wYdFaCU/8QOAQ4
    rAKfw4H/KDOoDwJeAJ4u4vMU8NUUS8VJKYCOZL1H5EVH9FHyO0XuA/b3AZ1iEekpwqzXA3zShsGG
    uJJpw+BxYL53aC+tzC6xtj9gw+ByGwYd8eu9/+d5M8kkYN88z3aNDYNF0bVameg+IXBHrOx3bBhY
    /75S9lXgKq/coTILVIRTb2hqaV1Vyem/ubGBppbWoZI+tiUoiomA08rsYMNgq3f818BWAfNuefr7
    LeACGwZb+mlqiwP9RzYM1uYxJ94oSm5EewNr0xRjubYW6I44ow2DFq3Ma8CHPGbRnFDFFuDeBAtJ
    dP4W4Ayvr228b73fT4l1aYzct/oDa/0oM6DbgBeLUPKQMitTLCB7amVOBP5OZpUxHieLzFWve2JZ
    fR4RLRCQlMt2/FTSoPP+/yFWfoeYSS0HfFKeba6v5Mnz/RH4P2CPmEKYRNvpG6cSB/abPoMkv/+g
    TZjEGFy4QC5TFB29YMOgSeyvhT4fB66PT71iLntCOF5TBOgYszhQlKK6Itp0fYqoMRA9J59ZsjuP
    uXKsyNgPA+ekWC2miTy9W5GzSEeeWcAHZhf5wxG6i9EvRqOduls6dFupF8pLnyFmpwisHTJ1PyaK
    Ur1wsEOACfSGg+ard3kRXLrirnCxYNwolpLonpuB54Hfywy0MzBbzI9ji5zxBpUy50sJZj0B3bUe
    oNcCZ9kweCjlmqNEAZyQp+q3h9FjNoppMgL0bcCXYxYU//l+BRwz3N5VZqcugUsL/bV3+NoI0HFt
    XWgjhdd+dpVZ9BgIzfV+Lwe+GllQUp6vczi+q3KAuuC02NzYUNLx/pYbBG4dV4ZeyGN6A5iOi+Wu
    yMxRAfGk0fu90YbB+ujZEiw99Z7lY8SJH3s3NzZ8KkFZimhpU0vr8jSrRnNjQz3OHp2mUK1qamld
    XG6Zup/cenMMTEfhnAh9goJE9h4DfJfyOKW6RXavxy12GFchzu5bIqZqZabaMFjnPdN7ANfKnC6y
    9YgE9Wflk0ZX4Hz7aTQN5z5NY8c3AheW8Zl30cqcTGlrFnuAx20YtAILvWn6Qq3M4wLsnAC/Rytz
    kJSbVqY2bwZaPdncamWOxwU8lTO1xf30utKnAz8QS09n9HxamSpc7MuXh6uoOFwUxcEMct8fcQaU
    SCfgAn0uEQ5dj3N/L8TZvV8CarQys+UePpcdqJi3HufY2Ef+z8TFhoBzs5dLtn0T5208U97JX+Hs
    wotwDqJdcAss6rz3lhtuoB4timKuDIMuUuiWAJ+JDcQDZbb6Gw/QncDllGHFjMSLXML7nRjtKZw6
    15/+EW/h+UDo1VENHA+cjgsuigB9B24VERmoh4ZW4zyDywbw2eLJzA+L+HILsAZnltso36txscS1
    OCfGCrl+pQfADWJdWEGCxzJJKbRh0GbDYC9c/MNKnDlxjdTZDbRInWtwK84L0TqvDVs85bPdhoEW
    IP9Wyr0jn3XAM8BxNgzOlNlimdzbN/ttEXFpubQ1n0OlA+d5XQ68VmDW2Sb1LZdPouiVa25sOBgX
    y1spuqKppfVbeSwbe+CCh3ZNk6mbWlr7LVP3ZzlQAXDFV4vU4Lxu9SL7vp7vfinLuQq2MSGOYyxu
    1UzR9/GtJklLyuLKrvyeAkyRmWmdDYO309pcaFnaYJktRzyoK2ziG/QXOFjgyHefwQRoBuqMMiLz
    KGY0Amm05NKrKXIAdyetpBgp4s1QiQ1paSEq1ZbREtB0HXBwghbuKzrVIgZdNlwaLUppFc4u3Y0z
    s71hw6A9DTw4h0+UqKdNlDu/vgmi+OVwsc6rKw10uW8DvTHXW20YrMk49cBoDvAXRZTbPAzbPh5n
    OmsTIJ6Mi+VO1JGAfwVOk//zcfZt3/SlcJGGVVLv4QyO82seLr68B3gWZ/POQD0AapNPl7z4iR6n
    3oSzjdZImWGnzAsAJxb5zsbS605PCgWow60phPwhsZUYnOO93x9ombq2CGU130PWlaENF3tTci1u
    uVM1zgV8Ic7JkKM4p8VwpzEpvzNFsQS6C+cyrUnhMksKXN+Ky5dRl3L9yoE0ThSS5d7/HbzpthtY
    GaW9Gkplqoz0S9waQmTw9mSgLp2WNLW0PpSvQNpqcDnejku3VfL1xSop/SkveeSiGeRdGwavy/F9
    cKuvd5JBsR5YUoziIzHIs3GBQbUiw68Fltkw2OwPKq3MjjhPZTyGe7pWZl/PmpMD1toweEsAvBB4
    Us5tikAtac3q6RsNWQfsp5WJp1f4g8xkewpGqnA7KWzP82zVwCxPfl9pw2CrVmYWLkbFn43rtTK7
    x47lovx+ceYizztTRLBuXDjCMhsGK5MYUTlAXXCNWhogiwXqEKVHuBaX8y4HPKaV0cDPcYtxJ3vP
    vR34o1ZmgQ2Dy9Kyi8r1V+EWq44XoHTgAp7Wa2UeBi73LBtH4lKHbUlQuLbTN+Dom3I8B3yd3pQD
    twNfESDcAHyMvrlHpgOPx+rvxOXw2AkX2NQgg2EOLt4ijSbgnHiRCHcMbhHFY4IzX36fjds8qyrG
    YCfF+mwGbknZR+Rc1OcdwAatzPPAOfGV+JnzJT2ibSfpyInCWZ8Ry8HU2ECuxa0YuVQrc28E6Fh6
    2s/JgNhPXm6VJ/NOwi1i/Wdgi5d3r07AtHNCu6aKWW6KXD82ppBNlo/PCSdK+fExhjQl9pkqfVIt
    dUyR2aK6iH6s8+qJGGbU1toYgKfF7jsxBujP4IKtjpbz1TFdYRourn2ZVuZTfi6RbOFtceSbA1eJ
    SWqbiCEf886drJW53obBRd7UOZu+WYjexAXjb8Ithzoct0lRRBGI3xXRpF3EAP/69hhANhTxDG8I
    SCI7dTTLtMTKdVHehQcrZBDv6g2+bbhoxjinjgA9V2YpvHb+QkyQtTifw3He4F+olTlO0sFloC6R
    LsGtDIkSMuYE1L7d+FQBdpSG90veudX0OlJ66M0LdxguIeJaGwZRmrVHBcw7iTId0dlyzudcxewa
    doa09x9FFEGU8IMTlMmtCTNEf2mODYN2rcyDuHBWcLlXDhMdIy6bT8Tl+qvyDAmfENGnU56hRhjN
    I/K7BrhPK7O7DYN3MlAXT/NsGMxLCNH8tSwPu8+bbg+iN7f0AV4dt0ny8aiOTq1Mpw2DJ7Qye0VK
    oZzvArq0MnGr0LZ8ClseBXib1N0RUwrbk8Je+7mDQdJ92z1LE/7vFM/odR6gO4CmWF6UHq1Mhw2D
    x7UyR4qFB+m7LwFXZDJ18XRV3JriyXEPAUs92XKmBwpf0TtXK7NvwiJdJAffWyPApDgQ82sDvWmI
    EcX5fYl+vD57mt40vzngWK3MjmWzfjQ3NtRW6Fm7m1pahzrI6FUbBm8nAc4D9mJ6k6ZPkE7uwe1n
    GL2oDwFLtDJ34RYjL9HKdEdpwEY5oMFlttrVs8IskGC0xJlDjv1MxMJq3FK6CeUA9bnNjQ2ayi3A
    /Al9U7gOBa2Jc+kY9dA3O2qVB/p5WpnjPHkSXALzU0VJWyjy5gM2DLaPVmALc5jimRy349JEF9IX
    xskAqMZ5jceVA9SR+ahSNO2DzH0EoBr4Iu9PFdEoituZwHKtzPE2DFaOYo7tmxvHxhhB0XVkMnWF
    uU+kENkw+La8tCtxyRbXecpTDc7hscKXuYepiJBG5Zipt3q/u2SGXF3CZwPQnlk/Bpdrb8Z5/74p
    Gwb9GS71wMn0mghv0sp8Wqwfw4lqCgy2SWUYMG/RGzHZiTPlvVNCNTngrQzUgyN+vO+YDYOlWpml
    Ngzu1spcRO8mPjNwHsw3hrj5HSLXRjQHtwdj2jOeWAaZugVnl27EOVWOsGFwR6l9PNrFj1wFFdzo
    Re0leedyKedzuLRrvjXJd6zEvXu1SeJACXZlfyevOkmTllSuTThnRF+M38szSY6nr5Mp7b6+vPze
    3i9eSMFrwO+8ct/XyuxWQBSq0cp8TzZpHR2gjnLAeUDaGuvoLSXIjaXeuxb4EfBjYJFWZnaCSbAH
    F/cR0Sa8LSJsGMRjvE+IDYparcx+Jcjh0WIJgBlamf1j+URmaGXGyH1/6113qFbmhzFdIXpGS/o+
    iRH5XtF9tDK7RMkmpZ7I7HlJDPz/K6EGSQxhJs6reDlwv1bmAq1MbsSLH8KJbsBl9u+KccJ64Hat
    zEY59pQNg0vLdOse3FZsH5f/xwAvaWVeFM68WMxRZ+OCdiJ61IZBW2xqfRbnSgeXlHK2cLSZuHzZ
    F0h9xdArMpCjqLmXtTJ34uJMPir9tIuIH9fgkrBH/XWWJKZcgItBOVB0gmLoF2LpAWe2W6yV+Zm8
    kxNxnthxNgyWaWX8jUv3ln77Oc5ru0FEtFNwkYx473IWUDVaZOqD6Rt45M9UB3r/y7lGMYcL67wV
    t7tsRAfiUvwm0ZO4UNE+1hNcOKkfc350bCD0lDDIfy97IPpT0mkxTh4NqKVamTPouwttA/CFhKqD
    WJ3x+z6olXnaew+TcDvmJt33BxIDco13/hTSNwTtAL5hw+Dq0SRTFys3+/1RyrKoKMjGl4uRhQPn
    A3vhIvPy0c24HHVbE/YQfAS3b2Ea+ZF2NSm/fdn1dJxjI4neoe+q+wDnEc1nhbg4QaauShDpVExm
    9mllTLT5N2EALxfot//BhfR+L7rXaOHUn8fF6/YUAObbMcVocmQmKlB/F87+/H35/5p3ry4bBitw
    Yak7ikx8sEy3W0UcuMeGQUsCh/bjHP5LK7NAuOEsqf9l3HZ173o7417jKZ6r/WeOxZycrZWZh0vX
    u5vI8s8AYcIehk8Ak7QySgA+DRf++hzwYxsGmyTJ/Fx6I+lejd9X8nsfJB7Wo6R/1wGLbBg0+wss
    pI0vAQdoZf5czJ6zxCqyTgZHGK048uPXByPt2EBpoAkiS/LOFZNAsRLtyJdcsdD15fZA5tlUtKT7
    l6ueUvvVB3X3MAX1TU0trReRUUZFUg1uJce5FBdoPhS0OHtNGZVCfwLteMXEzr1bnQAAAABJRU5E
    rkJggg==
    '''\
    .replace('\n', '').replace(' ', '')

_valid_html5 = u'''data:image/png;base64,
    iVBORw0KGgoAAAANSUhEUgAAAFAAAAAPCAIAAAD8q9/YAAAA6ElEQVRIieVXUQqEIBAdl74D57cD
    BB6mY3iojuFhgg7Qrx+eYD/eIjJF2rYVrI9B5KXjPGdUUlprqgkNEXnvnw7jJjBzg94y9OfddW4O
    1pz3cxHacSKi19Nh3I3qBXduXvc7N8PESMEUAqUlOjDBbw4WUwSfxXaGhZJl6GHpFgjmDHDy0/MP
    GftigjWwQ2s1heMgbH23fXfbZdMSrGnHCe2+k6sEQxgSm/JrpgSIMpvAkq/Yl/KlpWBUaWxBxrqN
    TCzmn7xnhUhTGjfraIaV1tp7X8k7zMzVP0t/j09JPx3GTWBmVdvf0hvVzH8tNHhSIAAAAABJRU5E
    rkJggg=='''.replace('\n', '').replace(' ', '')

_valid_css3 = u'''data:image/png;base64,
    iVBORw0KGgoAAAANSUhEUgAAAFAAAAAPCAIAAAD8q9/YAAAA7ElEQVRIieWX0Q2EIAyGC2EBuoAD
    MIcrOIejOIcrOIcDuAAjmHvopWlaLycS9YHvwdQ/hPSnrUQXY4SWCACQc347jYdAxECRH5f67fap
    T/NWv89NrEMHAP7tNJ6mOcNBve9Tz+3N8T71pMjOJ7FmFqjH5BSQIkWl2AWlaMOEtA3ClTwCFZSy
    Dh17UIFdw8iDuOb52LDlVz3rv3acd5o3VXOrwFFfFHF2hv24+HHh3masUgp3KQCkeWOT55UidIXJ
    FT9JtDPM5i9X2Fbv7wAfKqW4GGPOuZF7GBGbu5aaM/xt6bfTeAhEdK39LX0ALc+PofdKE1EAAAAA
    SUVORK5CYII='''.replace('\n', '').replace(' ', '')

_favicon = u'''data:image/png;base64,
    iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAAAXNSR0IArs4c6QAAAAZiS0dEAP8A
    /wD/oL2nkwAAAAlwSFlzAAAN1wAADdcBQiibeAAAAAd0SU1FB+ACBBQjLyFlNIEAAAonSURBVFjD
    rZdpdJTlGYavb9Zv9sxkTyYbCQlJCGEJxLAoBKVIASsVrW3Rql3sOfVYa6vV2hbFLqetLVr1SCla
    awtiq0DF3QpEwlLIRiKU7CSTSWYmk8nsM98sX39gyaE/WuX0/f2u1/vc9/M8gizLMlc49nftoHXg
    IGtrtrB6zs1XtIfi00yWZRl/1IsnNAbA0f7XSaQkusaOkpbTeMMT+KNePs2bhE9DYCrsYm/bdoa9
    57hr2VYmQ0763V3U21eg1xh58cRPydBl882rf4ZBa/5Ee6o+yaQux1E0Ki0GjYVx/zDxZBR3YJS6
    wqUoBAWzc+rpGG0hFPcTivuRkTk3cZpoIsS8wmWoFOorv0CP8wR7Tj+BQWPhhvqvcftVD+MNT1Bv
    X8FTh+7DE3Jy3tXOLYvuRasSMYs2psIudp/6FbKcJhIPsrxiw6eLAU9ojHfO/hl30IFCUBBLRAjF
    p1EKSpJpiWDMh1YlEooHkJIxwvEAaqVIWAqQltOolRqCMR+huJ+UnGQiMMLbZ/+EP+r93zEgpeL8
    pe0pTl14n2xjIQ+t3cmApxtBUKBV6Xjh+Damwi5unH83C4tWcd7VxtyCJtpGDvGX9qcAeGzDHqYj
    k/hjXoqss3np5M8Z8HSzpHQNmxfeg0qhRhCEywn0ubvY1/kcE/5hzLpMlAoVmYY8pGSckalexv1D
    yLKMWcxEEBSYRCuu4AiO6QG84XEMGhNalQ6zaAPAMd3P2PQA8UQEi5iJgIBVn8101MP+rh20jRy6
    PAZ2Hv0hUiqOO+jgjqZHKLTMYm7BVZwZa+X17l0oBAW3NT7ErYu/w2RwjNqCq7j/1fWk0gn6PWf4
    7rVPo1GJ5JiKmI54eKXtyUtUb1p4D3WFS1lQdA0vHN9Gl+MoAIuKV81cYFFxM4OTPczOqaff0023
    8xhSKkZhRjn2jAoMWjN6jYke5wnGpgfIMRfRUNLM8ORZ6gqacAVH6XR8iFm0saT0Osqz60ikJAos
    ZZyb+AfdY8cwiTaqchfhCoxiM+RdHgPJdIJBTw/FtkpePr2dTkcLJtHKtg0v4/QPISCQTCd4ruVh
    wlKA9XV3sKryJnrdHVTnLeZw72vs79oBwLYNLyOl4gSiXrKMhew4+ggOXx8Liq7htsaH6HN3km8p
    wyRmoNy6detWgL2nn+RI3z4AqnIX4A46WFm5CY1KZGfrjxiY7KYiex4alUhaTrO2dgv7u3bwYf/f
    iEgB5tuvxhUYYU5eA6WZ1exs/TE9zhPkmovIMdnxRTysqb6VXncnB878jl53B0tK18x8QfvoIRIp
    iQHPGZqrNlOSWY1Ra6F14CBTYRfheABPcIz1dXciJaOIagNtI4dIyynOTpxm3dyvcHvTw+jVJnrd
    HYz7h1AICoYmP2LDvLuYb1+BXmPicO9r+CJufBE3wAyBDF02ofg0zXM20+loYceHj+AOOriu+lYm
    AhcoslZSb1/O88e28deOpynPrqM8u45gzMfa2i/jDU/wi3fv5tTwe6yrvY1Q3I9JtLKq6iYO9jzP
    H0/8DJ3GxOLSa5nwD9NQsprK3AUzBGryl1Biq0Kr1tMx2gLAqK8PpULFDfVfR6VQMx3x4I9OXpSZ
    r5+Gkmspsc3Bosuk/WNZTUVcxJIR1tR8kbScAmAyeDF5jU3301DczC0N9yGqdJcH4fYP7mPYe5ar
    Z3+Oq0o/Q9voYWrzl5BKJ3nmyIPoNSa2NH4fKRllbHqQ1XNu5tkjD3Jh6jwNJau5sf5ujg2+gVWf
    g91awa/fvwcpFefWhu+QY7LT4zzB0vLPcnLoHd49txuLLpNH1++eMaKoFAQgloigUekwaszo1EYS
    qQQAyZREWk6h05gwai2oFRrCH6+JSiFUSg0GrQWD1kJaTiOl4gAk0hKi2oBBa0ZU6Ykno5fWXEbA
    F3HjDbuw6rP5+z/3cmzwTQoss/j26t8wKo2hQkkqEuHlU0/gDjrYNP+b1BUuZTI0Tr6lhO6x4+xt
    244gKPjRuheJSiGklIReY+SvHc/Q62pn6azPcn3tFtxBB0athVxz8QyBcf8FjvYfIBCbQqcxAVzU
    aUrAev/vyXhsNwZZRKXSAmAz5OH0D9HSv5/J0Pil/K9ViYhqAx2OFo4PvYmAgF5tBMCiyyKWiNDS
    v5/zrvbLCTyw7wakZIya/CXctfTHuIKjFFjKCP1+N76vfQ/BZMD2zE9JbVlHPOQny5jPd19dTzKd
    oMhayf3X/hZ30IFZtDHq6+OZIw+gEBQ0V93Murm34wmNkWsqYlfro3Q7jwGwffM7MwTm5C7CJFqx
    Z5Qz5D3LG90v0DbWgnpBLcr8HDRVs0iW5XNy4C0Odu8iEJuiKnchZtFGWWY1voib987t4UjfPqz6
    HLKMBdgMeRRklNE91sqBrp2M+vqYlVWLWbRht87+DytOSfzT1U5Fdt0lKzZqM3h8416mBj5CkRbw
    F4jseP8BIlKQ9XV3snL2Jj4aP0m9ffl/WPFeookQvoibwoxynmt5GMd0P/PtV/OVph/Q7TxOsbUS
    iy5zxogO9jxP+8hhEKDYVkUkEWJJ6XWIKj2v9D9Pf2IQu6EEvcaIRimyvGIjh3pfpdPRQiwRYXZO
    PYHYFBXZ8yi2VXLgzE4GPGfI0GeRZcwnkYqzvHwDI77zHOndx4ivl3r78hkjOtJ3gFQ6gVal46vL
    H0Wt1DAnt4G2kQ/on+hAISiYn7+MZeXr8YZdFFjKONL7GjIyiVSMlZWbWFn5ebKNBUyFJ+hxHgeg
    NLOa1XNuwabPoyZ/MbtaH2XIe5Yh71m2ND44QyDfUgLAyspNnB75gH2dz+H0D9JctRkpFacydwFl
    mTW81vEs757bQ5G1gnr7cgQE1tR8Cad/iJdO/pz20cM0V21GoxLJNRXRWLqGtz96ide7d6FSamgq
    W0skEaJp1vWUZlbPEJhXuIx5hcuIJ2McH3yLRCrO6FQfWpWOZeXrL5ZrySjuoINUOsG4f5gVFRsR
    1QbKMmtoHXidsBQgLAVIpRM0lq4hFJ9GpzEyERghLadwTg+yomIjtzU+9N/7glFfH8cG36SxdA3J
    dIKnD38PUa3njqZHCMUDXPCe44b6b/CLd7+BKzjK3IImtjR+n/fO7SbXXEyprZqfvH0nADcvupd8
    cwn/GH6PVVU3kWOy/++yvMg6m1sW3QtAp+NDlAo1GqVIVApTYClDrzEBMik5hUK4qGRBEKjIrseo
    tRCWAujUBtKyTDwZpSyrlrKs2ivrjGRZ5sTQ22jVOrIMBfzh+ONMRVx8oeE+SjNrGPB0U5O/mF53
    J3tOPQHA4xv30uvqJJoI01DSjPbjzHdFjYkgCDTNuv5iWg05MWgtTEVcmLQZuAIjOKb7yTYVYtJm
    8O+6AmBh8cr/f2+YltP4o5PEk1HyzCU8uO9G4skIZVm1fOuaXzIZdqJWaMjQZ1/6mv9rd6wQFFj1
    OeSZL0q2Nn/JRa3bqlEqlOSairAZcj/x4QD/ApX6khedphFNAAAAAElFTkSuQmCC'''\
    .replace('\n', '').replace(' ', '')

_omnipath_logo = u'''
    data:image/png;base64,
    iVBORw0KGgoAAAANSUhEUgAAAPEAAADICAYAAAA5vgdsAAAAAXNSR0IArs4c6QAAAAZiS0dEAP8A
    /wD/oL2nkwAAAAlwSFlzAAALEwAACxMBAJqcGAAAAAd0SU1FB+AHHRQNLFaa+3IAACAASURBVHja
    7J13eFzllf8/987c6UUa9V4sS25y7zY2NhBa6CVZCFkINX1T9rckIRsIKWTTNmWTEEijm96rAWNj
    sMG9yEWyep0ZTa93bvn9MaPBPRQngUTneXgeLF3d+n7fc873NEHXdZ1xGZdxyUvzrT+gfcTLry+9
    mM8vW/qhv19x/JONy7hkRdM0ADwWC6UuJzaT9JG4b2FcE4/Lv6qs7TjAnpERGjweVjTUk8pkEHSd
    tR0HMNvtNJeWUlNYMA7icRmXD6tccde93PPaOlZOn8bLX/wciqaRTqcxSxJGk+kj8xzj5vS4/MtK
    TYGb5vo6plRWgChiNBoxAKLR+JF6jnFNPC7/dKLrOp2jAYYjESaWllDqcBzV/x0NBnG4XGRUlS89
    /Ch/WbueRz53PRfOnP6Ret5xTXy8xQComoaq62i6TlyWicsy2vi+96EWRdP472eeY+m3b+HhrduP
    ekwkEqHQ7cYqSRgEAUXTQNPQ+eh9W+P4Jz+2+GMx/rjhLcqcDmZWV3H5X+5BFASevuEa6go94y/o
    w2peCgK1hQV4qipxm0y57VjI/z4ei2GxWABIJpPEYjE+M28upzQ3M6+udhzE/0zSPRrgxt/cjrW+
    jgevvhJfNAboqNq4Jv5QL2pR5OsrT+YTk1oocbuIxeKkNY3+aAxJECgxm7DoOul0GqPRSFlJCWUl
    JR/d5x3/5MeWqgI3n/3kJXjMZqaVl7HqM/9OMBikwu0efzkfcim0Wck4HRQWFGA2m9nb1c0n77qX
    icVF/ObC8yhwOJDM5n+OTWv8cx9bKlwufn7h+cRiMewWC/VFRfh9PqzG8df2oeYydJ14PIHNbkdR
    FMxmM4oocqC9g1g6jdlm+6cBMIyz0+/eP/b7KS4uJhAI4Ha7MRgM4y/lQwzi0dFRiouLCYfDOBwO
    fH4/BUVFCIDJaDzIQ/7oyzg7/S7FbDaTSqXweDwEg8HxF/Jh1kyCcAhIvV4v5WVlWIxGzP9kAB4H
    8XsQp9NJNBpFVVUEQRh/Ie9BFE0jIcvIqvo3v9ZgOMy6jgPsHPECEIvF8Hj+uSMJ487de5CCggJC
    odA4iN+DaLlc5D+8uZFPzZvDcCRKOJXk5IlNzKyqOuHX+9OGt7jp7vuobqin7+abKCsu5p/d8RkH
    8XsQSZIQBAFN09A0DVEcN2TeDYg39vRy3xNPMbGkmF+tXUdgaITbrrjsbwLiCYWFnLFkEZU2GxgM
    vNrRyZ1r1zGpsoJbzj5zHMTjAgVuN9FgEEFR4COUJP+PEoMgcNakFgzXXc05s2YSiccJyjJzaqpP
    +LVkWeZjLROZWVZCeWkpADsGBnhw8xam+Wv/aUE8zk6/R0llMlx1173EZJlbzzmbmdVV4y/lr0ha
    lkkmEhQUFDDi9VJUXIxBEE6oW6JpGrFYjEQiQXl5edYXT6dZvWcfr/f3U1VQwGeXLh7XxOMCqq6z
    aWCAkWiU0Xh8/IW8G9F1DAYD6XQah92O8W/ghkSjUZLJJOXl5WiaRjQaRRAEzpg5nTM+YgUN4yD+
    W/vFBgNXL1pIIBaj0uX8cGIGSMgyBkHEZDQg/gOJOF3XSSQSFBYWEgwGcToPfWf7vT56AkGmV1VQ
    5nS+bwCPhf9kWSYajeJ0OjH9i7g74+b0+xVFIZ3JYLZaP9BpegJBtvb347JYWNk88YTcWlyWuea+
    VUytKOfaxQvfNzhOlJkbDAZxu92kUikcDgfDkQgD4Qi6rvP9F17i8Wef5/f/+VWuXbzwvbs3qRSB
    QACn04kgCKiqivtfLC12XBO/38UpiqRkGaPJ9IGyt9YdOMCXHn6MppJinrjuaipcrrwGO9hnPLQO
    5/iSlDM88OTT1E9v5cIZ048A8Qc593sVURTRdR1ZlvNs/uM7dvGj1S8zpbycZQ31CGedTqHNyg9f
    Wo1NMnHlwvm4c1VGx5NMJoPP58NisSCKIhaL5V8yk248RvIBFqcoiiiK8oHO47RYKLLb2N4/wJKf
    /eIdE0kQ+NRd99B48/e4440NCDmwvbtzmvnllz7PD886neqCI7WSIAh84aFHqPvv7/Lrta8fFcCa
    rvPx393BzNt+zAt79r7v5xsdHcXlcqEoCtac1TIQDtPdP8DG7m6e3L2HZ3bvwWQw8J1nnue2l14+
    7vkSGZkOn589Xi/t+/djMpkoKCjAZrP9y6bCjmviDwJApxO/34/JZHrfTOt5rdNIKwpX3nMfPYFD
    0zl7RgN0DQ/nCTQB+OVr63hg8xauWrCAa5cc3fw0G41cPX8uuq5jP4ZG6wuG6B0exheLHfPeDvhH
    6RodJZpKv+93pKpqXhuPvaNbzz6TXr+fNq+P7f39yNEYDrOZOVWVaIJAwddvpNzt5snrr2Febc0h
    59vWP8Cn/nIPXcMjRH70fZzjFWXjIP7ApEIu+eP9aoGMqmIQRW5YMJ/JZYfWtH5l5QpOn9jEKQf5
    yvtGRnizbS9LJzQe1w9NJpMUFRUd85jrly5mTkUZH5syCYDn9+xl1Zat/Nuc2ZzSPBFREPjmaafg
    DYWY9S5jumlFYWNPD/3BMAvq65hQXJSrKIofQWjtGvGyta+fS6ZNYevgMH/csJGbTjuF5ZMn4/zS
    VzAbDUc1E0UE7CYzRot5HMA5Mdx88803j7+G9y82m42RkRGcTifBRJI17R0k5AzFdjui+Ne1cziV
    4tN33ctTmzZTV1jIWTOm5/3UKeVlLGueiEPXsORMUUXTcFosnD51Cs2lRy9kV1SVQDSKxWZDPEo8
    VgeaS0o4qaGeMrsdTRD4y8a3+fUDD9FQV8eypkYMosiM6ipmlZdSWlBwVHM7pSjIqopkMCAA/lic
    z9x7P79a9RAN9XVMLS7CYDCg63relM6b2aEwk0uK+cLSJewLBHnw1TXUVFcBOqIgcsWC+Syd0IhF
    kg7ZnKyiwEWzZvCF5csotNnGF+C4Jj5xQM7IMpt6+zjrm9/m5JUreOaGazEa/nqIY2N3D75YjMKS
    YmY0NhILh3G43Yf4qcl0GncOfOe1TuOM+jrMx2Gcdw4Nc+szzzGrvo7rliyi/LBjx84tmkwEhocp
    Ky/nlJZmfOefw0mN9RhyBJSiKKQyCkfj30eiUW5f/yZtQ8P8+YrLsEkSFqOB01uaKbLbmVZZQSwW
    w2W3Yz/s+ul0ms8vnI+SyeBwODhr8iRqPIWcPKGRM793GxR5OKN1KgUHAV9RFBKJBG67nWJJGl90
    4yB+9+KLxdjU24fFKLGsaQKGo2hXl8uF3++nqrCApctOwmGSeGzHTubW1NBSVnrc87+8bz++4WHO
    njWTi2fPJBmLcXhvRovFQiaTQcot3mg6jcnhOKYf3uH38/iatWxtaebimTOOAPGYNhYOAvUpzRNZ
    UVeLLor5uLKYa+N68LXHJBBP8KcNG+ndtZvbP3kpNknCBHx52VJMFgs2sxnv0BCbe3oJ6tBUXMTU
    imwmVSQSQVXVbME+cNnM6VgdDlLxOIvnzKKmqAin2cxQJIJVknCaTMTjcRwOx3gd97g5/d5lY3cP
    n7n3ft7s6ubKhfOPmm00Fp/02G1cPnsWj23fyY9ffY0yl5OTjuO7Auz1+tg6NMTyulrOnzMbQRCO
    AI3JZCIcDmPLmY+JRAKbzXZMENtNEh67jWtPXsa0inLMuU4ktz7/Ipf84c+IosjihvrsAjAYSKVS
    mM1mYskkQg64B/v7iqIckThhFGBikYeV8+ayoL4OTVVJp9MUut2YjUbUTIZwMMiqfe184aFHcFrM
    nNrSDGTLAyHbpM7tdiMajUSjUeKJBFcsXcKls2fy1K7dXPyHPyPLMvMrK7LHjRecjGvi9yOyqhJI
    JBBzjdUsx2jN43A4CAaDFBUVYTKZUOJxkpnMXz3/l5afxOWtU4lFo3nARqPRPGDHwPRecnLKbDY+
    NXsmjQ0Nh/zcG4vhGxk5JF3UarXi9/mOIJ4O0dpHubbVYODMyZNw2O1Z89rvp6ysLP/7WO4ZUoqK
    HosRT8uHPE8wGMTlcuUzulKpFBUVFfljUpkMiUCQ0XicgoKC8YV4PHJ1PGPr3Us4HEbXddxu95Fk
    UW4xGiUJPSNjMZuRMwqmd5HRFQqFCAaDlJSUYLVaiUajhyxcXdeJRCLYbDYkScLv91NUVHRMTTw0
    NITVYqGgsPCQn7/W3sFre/dx+vRWFtTV8vqBTlZt3UY8HmfRxIlcNXsmqiBgPqj/VDqVQtNUrDb7
    If5pPB7H5XIhCAJer5fSXNVQMpPhL2+9zQtbt/H9Sy6i2e3GmAtzjZnwe/fuRVVVGhsbicViqKpK
    QUFBvo2sLMt0Do8QkGVqi4uoHgfxuCY+UeJ2u8lkMgSDQex2+yGLXRAEJEkilUohCAIGHQLhMOXv
    AsS6rmOz2fJ+n1GSWLd3L1FNZ0JRES1lpdjtdqLRKIWFhdn2M8eJSwu5+9naP0B/MMTKlonYTSaW
    T2xifnkpVmc2K2xr/wC/fuEl0HS6I1GunT8XNdd9Q9U0YrLMrr5+BkMhTp06lUKbNW8GWywWBEEg
    HA4f0jkjrSjc89Ym1r+9iSuWLKHSZKIgB86xOx4z35PJJKIoYjabsVgs6LpOKpUiFotRW+Rhkt0+
    vujGQXziRZIkCgsLSSQSpNPpfM4ugNFoRBAELBYL0WgUk8mEqqrvioyRJIlkMpld5KrKjU8/xxtv
    b+bq88/hzn/7RH6jAEjIGfRYHJNkwGU+MpnDGwzy2zVreXjHLvztHez61c+ZWpGt7knImTzbPLWy
    nAvmzAZ0zmmdls0+y11jv9fHt55+llf3txPq7OKFW2/mY5NbyGQyqKqK2WwmkUhgsVjyPjSAVZK4
    Ys4sFtTXUSoZjrAGUqkUoVCIlpaW7PAyszmf0ZVKpZBlGYfDccgGOS7jID7xPoggZMNKmQyhUAir
    1ZrP3x3TrKqq4vF4CIVCx026GDve7XYzODhIVVUVdquVM6ZORtD1QwgoSZJA03hg124e27mL5ROb
    +OE5Zx+RC+0Ph/ndcy+A242puuqYfvzC2lqqPnYqnsICPBYLsqYh5jac3mCQx15+FamshEUL51OZ
    S98MhUIUFxfnN5zDwWY2Grl+8aL8ZnA0l0QURdLpNHa7HY/Hg6IoyLKMLMtYLBbMZvM4iTUO4r8P
    kE0mE0ajkWQymW9p63Q6CYfDeTLq3eRWSwYDBoMBR27wl1WSuHb+PG5YvJh0MpG/3hhQu4Mh3jzQ
    RV8wxIyqSj45exbPte3h4W3bSWUUvtw6her6euZUV/H5hfOpKTh6ZlMqHqfAZsXtdLK+vZ0Cm42G
    khLMwIyqSm7//PVo0ShnzplNbWkpo6OjeDweNE0jlUpReJiWBYimUnx+1cM8uXMXT9xwLcubJhzy
    +1g4TIHbTTqdpr6+HlmWSafTKIqCzWbDbDaP9zAbB/GJEV3X2dzfz0NbtrGieSKntjQfNbwkiiJ2
    ux2r1YrP56O4uBhN0/B4PPh8PgoLC/P1rQdL92iAhv/8BlablZf/44sscrsPyWpyWa0YDAbC6dQ7
    1xIEunw+ZlWU0VxeSrvPzxudXXxy9iz2DI9w36bNpFJpfn3mx+j89jdIplJoqpqP+yaTyTybDNkM
    KASBkViM5f/zM1rr63jouqtpsdkod7m4cv48uru7KXM60XM9xQwGAz6fj5KSEgbDYTpHAyiqyskT
    m9CBtoFBtvb2EQ6FSMhZRtoXi3H3W5uIZzJ0dHZyaetUzp4xA1mWSSQSiKKYt2aOJWlF4bHtO9k2
    MMA5rVNZVF//D62THgfxh1hUTSOYSKLqGusPdPE/qx4iev65LG+acNyOFKIo5ntSm8xmuodHSKZS
    eIqLyRwt1CQABhGL2YzRmDVhDZJEV08PDXV1PLu/gx88/wJxReGN66+hqKQEi9XKfa+s4aannwVB
    4Lwpk5mbGwDWWlnByqYJDEeiKDltNrbEjQYD2wcGOPd3dzAciyOPeLn/q19iQWEBDRMmMBAKg8OB
    y2LBILzzjMFAgP60zNW338kBr5fdN32TYDCYJ7Ie3radLz/0KMRi6H++k3Qmw182b2FXZycXL12a
    T+4YCoe55bkXiEQioChcvnIFmqoSj8ex2WzomkYGiMZimI1GXIeB+cYnn+ZHjz4OZjOk07gsFhbU
    1Y2DeBzER5f+cJgz/u92ZFXhKytOZsmc2bSWl+dTEY8nBoMBj8dDx9AQl/zlboajUVZ/6fPU2+1H
    EFyFNhtfPetMnEYDZQ4n3miU37+9mZHhYeZOnkxfMMhILI43GsVpt+cTQKqKimiuKCeUTPGbM0+j
    ckITA6EwHpsN0WBkS08vn7jrXl754ueQjEZQFHoCAd7q6aN3xAuSBMVFSIpCLGcRuK1W9n3rv4hH
    IlQWuEmrKuFIBJPVSlJR2NLbRyIQwBsM0FKRHcgN2cFlFW4XhpxZHU2lKDIawWqlssBFMBbHCiRj
    cU5vbuLVA534gyE6hodZXFGO1WpFURTsdjvffOwJHt66jfNaW/nVJRce8l7LnU5KK8opsTsod7to
    Li1hHL7jID6maLpOLJ0mJcucP6OVLyxbiibLecLn3YjdbkfWdNIZBaOmY8/5yW63G0XTCOdCKz+9
    4FwyiQSxRIKucJhv378KDAbK9rdz52WfZE5NNZt6etni9THRYaOouIRLZs2gxiShG0RiikpaUfjD
    hg1859EnQNOYVlvDgtoaUFV2Dw3xdlcPT+3fzxtdPRisVuZWV9EVDPLknr10rg+iCAL/depKzp/e
    Stxkwmax8NvX1/Py7jbOmjmDOeVleCQjCUni9e4eHtixC7vJxMcmt9Du86NpOjcsns+jW7Zyz+at
    THU5OX1iE7986hkefHszJ01oYGplJQ9+6t8470938+TQCO3+UcwmExlNw54z7xPpNNFU+qgJMhfP
    mkGtxcykmmqmVFePL9JxEB9fKlwuHr/2MyRTKWw5cy2WSmHLpSN+46lnsJskLpwxnSm5roqHS5Hd
    zqqrPs3Tu3bzoxdfYllLM5+ePQvItuO5/C93YzeZuOXsM1lQWYFRkkgNDXP+0sXIqTQzqiqxCAKd
    o6MMjY7yXNseFp5xGuv37ec3G98mHAhw4cwZzJxcTiAYpDcYhlSamVOn8I1FCygrLWE0HueVjgPc
    +PRz+bTQmgI3TqOR0ViMuza8BYoK6Oyb3sqjW7fzwq5dfPbk5TyybQcvb9+BZjBiFg34UmnQdbqC
    Ib7/5NMYrFZqCgsYCIUZCQT408a3SGQyeEcDrClwM7WigjPmzmb/aJCH3tjAzsYGbly0gJOrKph+
    yYXMLClGNBiwHWQ2XzN7FlctXECx48jYcLXbjVxSTOM4gMdB/G7EYjQyp7YG0PH7R/ME0Fhi223P
    PI/BbOLHq1/lghmt/PzC848oiTMZDEyrrOCp3bv508a3ORCOcFZLM6VGI3FZZuPefRhsNvzxeDZj
    CZhcU83vqqtQFYV0IsG3X32N3QOD1LicnNLSDKKBUDzG07t2E4nHebG7hwWV11NdWsI3Tz+VcrOJ
    nkiUb7y0mqSm8+kF89B1HU2WsyY08OVFC/n6089iMpupq6jgxlNXYtR1ZjTUcduLL/PQps3ct30n
    Z0+dwhdPWcGsujpe27GDdDTGKbNmZJsDmM18bdlSQskUI6EQNrud7tEAH58yiQ6DyN4RPxMnT+Ln
    l17M/sFBdg4O4TCZCCaTXDh7FgaDgdKSEgwH5YaHw2FmNNQfUWQxJh0HDtDY2Di+OMdBDLKisLm/
    n3haZk5tDYXHzaYSsNvtJBKJbC60oiBJEouam0grClsOdDISjaIeJ2u1pqCQ0sIC1rbt4Ycvrubn
    l17MhOIi1vzX19FVFTESRhdFinJZWCMjI8REkZZbvg92O4giffEE6w900j0a4N9/fyclpSVYzCZS
    KZnH2ju4yiSRstp4bM8+2gYHQRQxiiLDgQA2SXonXisI3LttO4giC8tKOHXaNPYMDnLurJlMK6+g
    qaSYQqeDYDzBqi1bEYCbRJGbzz+XgUSCZ/fs5eX97ZgMBn509pncu6uNtXv2gsUCgoA3FOJry5bx
    o7XreHzPXkxPPcPugUHMRiM/PfsMNLK9oEqKijAdBNZkMonRaDwkYeRgSaVS+VZI43IMUvWjeNPf
    efZ5Wn/wP3ztsSfe09/54nHO/u0dnHbzrbzd03tcsIdTKdKaRiaTwWaz5Stv5tfVsqC+jquWLuaq
    WTOxHsdX/tS8OdywdAmSxUKx200mncYiiiyorKClwE2fonLfth14c+cuKioiGAiCIHLupBaaijwg
    ioQjEe55exMIAl9cuoQFtbW4bVYyue4dJoMBm2QEXafIbueK6a1cu2A+1y9dwo/POh2nScJtMbN1
    eBiTJHHp9FbWtrfz8+df4undbSDA/zt1Jee2TgNVy99/ZypNTFXpCkfym4HLYmE4EkFCZ9nkSUwr
    KwVBYJs/wGXz53LapBaC0Riburp5o6eXdV3dtA0McsDrA7P50OSQXBz9eO2NYrHYcYszxuUDamJF
    00grSjb312T6u1H+3miMDp+PaRVlyKrKEzt2UufxMLe2hn0jXiKpFJUFbmoOS5w3G40sa2pkv8uJ
    w3hs8D2/Zy93vrGBppJifnbux0mn0/lF9otHn8BWXsbzn72OhZUVRHIDyI8lX15+EhdPnQyaTkd/
    P8WFhZQUFmA3FPAfTz6Dv6eHCo+Hi2ZOx2g0UuPxcO3SRSybMplgMknnwCDzKsp5YMcuJIcDl9HA
    /533cfqHhlg4fTq7OzuZM306V86aSb/Xx8pJLcjofOahRzi3dRpPt+0hGk9y0YxWIukUgq5TbLMx
    t76eqKLSNjTMJ/74F2wmid5AAHLvRQe29PXxq9fWsWdoOMtIqyqfW7aU6T/7JaUOO62VFei6zvTS
    EqwmExo6SxsbaPf5qS0tIZROc8A/ymX3rYJYnM+ffy6/vuRCRqJRDvhHsRsNTCouPsKMTsgZBsIh
    9HSaEocD2zEspr0jI7QNj1Dv8TC75l/XX/5A9cRb+we49YUXeWz7DhY21B8R3/tbSanTwbKmCaxs
    amJ3fz8XfvcH7E+l+NS8uVz3wIP88MXVOMzmI2p5LZLEGZMnceXCBVQ47KRTqfyQtIPloW3b+d1T
    zzKYlvnKKStIpVLZPOhMBoPLyRmTJzG7qpKSAjfRWAxFUfKJCoe3f7WYTLzS0cltL77Ixu4ezp0z
    mwe27eSBzVvY0ttHWhTpDAa5bski9o54efXAAWYXFTGrqhLFYGRwdJSucIQX2vbittu5fHorrcXF
    KBmFV4eG+N+1r7NndJRGt4upLicXz5rJr9ZvoMPrx2yS8EZihNNp9gwOcvsF53LFnNk8uGMnIVWl
    2GLhQCDEml1t7OzvpzcQzIePUDVimQyTysvZ2t0NBgPoOgsa6nlp23Z8sTi7+gdJqyp3X3Au1y5d
    DKqKQxRZ193LS+0dJJQMKUWh0OGgqaaa81qn0h8O88s167jp6WfoGg1yxcIFR3zft3v7uOiOP5HW
    daYVF1F8jNGk33txNV/93R0MZDJcNnfOuCZ+P9I1Osof1q2HZJLr5s7BrevIqorFYsmHDhRN5+2e
    HroDAU5raTkq+/heZX5dLfNzSQ5v7NvP8uUnsbS2BkkUsZskbEYjgq4d6TsIwpF9mcZCGgf5XJ+a
    N4c5NdXYjUbkVApN07DZbHh9Pm4564ysr5ZMkpZlSkpK6O3txWw2Y7VaEYAHNm9lS38/KxsbaPZ4
    uPLe+5Fjcax2G98ZGuL29W+ysa0t608ajWzqOMApv/oNQ5EInf5R5ldWUFvg5un2A4TDkeyuIEkk
    ZZk/b92OxSBiMkpcf/+DZNJpXu3uobGwEE3XWD6xiTsu+wRfe/BhXt3fkU0PlSRKnQ6cdjtt/QP8
    ftMW5LFiB0HAYDFT5XQSVRSCiQRkMiyor+Pflyyma2jooBcocu/bm7P53CYTb3R1E0gl+cWGt2mu
    qabObOL05olYDCIui4Wu4WEQDXx+ySI+PW8ujaUlzP3p/7JtYBC7JFFf5DnqxueLxejcvp0tpSUo
    xxjBoioZWktLOGvFck6d1PI3A8jmvj4+/+AjnDV1Cl9feTK2D+FUiQ8E4kUN9Txy/TVkVJXW+joc
    Vis62XpQn8+XrYNNp7nt+Rd5ctsO1v3X11jqaDihDzBvQiP3X3EZmVQKn8/H7Z+8lN7+fsxG418d
    PzrrRz/BH43x36eu4OolixFzZl29x0N9bvePx+MIgkAikUDX3tkYLFYrsVgMQRCora2lvb2dlpbs
    Ynp8x05WbdrMrmlTuO3cjyMPDVPaUM8tZ36MvYOD1LucWGdMp8vnZyAaRTGZeGXf/jwB9Wb/ABv6
    B8hkMsxpaqQ7EGA0GiOVyfDMjp2cOqGROqeDTDwBRgOKqrLf7wdd5/LHnmD/zf+N0yRBKoUuSdR5
    PMx2OVh9oJP2waEsGSeKWRDn5iRNrqygbXiYYDjDJ2ZOZ0phAQvratm1f3/2OFUDIdszus7t4u3+
    ATAYyCgqD7btQdrfToXLyWb/KAeCYW5onYx9yULMJjMrpkyiwmbDYDRS43aTSCQZiMVZtWUrSxsb
    uXT2zIPcZJ15pSV0/+F2BEWl6CibfigUwmAwcPmC+Vw6by7S35D08kZjbNy0BY/NRkbT/vk0cZXb
    zYUzph/G6WZ7Qo2Zl6ZkMmuKZWR8fj8+hx1N1zEajX+1uufdiGQ0UpYrDfT5fNjtdurKyggEAkSj
    0eOO9BiORPDH4hjMFjRVJSnLeQtiTOx2O5FIhFgsRnlFBcMjI5TnOlg4HI58gX5TUxP9/f1UV1fT
    UFyE0WjkuV17uO28c9AfuDs7nDydZsov/4/BSJTbP3kp04qLOPP3fyCS04oFViuhRAJF07KaSdNY
    Ul/HzYvmc87d92OxWjGZTLzS00ujxcyfL7mAqCzzi3Xr6YgnQBRp9wf4f48+zp/OOZu1vX3csW0H
    FRYzD3Z0sqqzG5MkcVJdLaVGA72pNEaTia5QiBfa9mSBLUms2r0n46BDAwAAIABJREFUSyCueyNP
    aF00awahRII3urrYMjCQTdQwGJhcVUlaVdk5NEwwkeTVjgPsGxompus8eskF1NXUIOWKRDKZDH+8
    +AKS8TgXPPgIOweHiMmH9rQeHh7GZrVSUFBAIBDAchARpqpqviPIWLugv+bA6bqOnrPC3o/Uud18
    7ZILmV9ff8xqsI+0T/yu4q6SxL/Nmc3N553D5Koq7HY7DocDi8WC1+slmUwSj8eRZRmTyZSNbeZ2
    vDFfNZRMIuf6Mx/tY4wdZzKZCAQC2Gy2fG+o44UvylxOTp88mTkV5VSVFOfL4cbqgvOEmNlMRpbZ
    3dfHLatfYeeIl9k11ZiNRmw2G8HRUex2O5ogcuMjj/NUrnNFmcvJlQvno+s692/eSjQeZ213DwZR
    5KTaambX19Ht9dETDCJrGqmcab+krpZzJk9ih9/PG/v2059I0huJMr+ygpMa6nl0VxubhkfYOzrK
    aDpNdziMWZLQdB1V1zECBWYTD+zZx5r2DlKKgiiKJFUVVdPoDobY7RtlgsXM6i9+ji2dXewa8eZd
    CqvRmN1IRDHbL0vXCScTPHvFZTy7q43BeIKGwkK+umg+vmQKbzyOy2bj400TcJrNDMYTVNisDARD
    vNjdQzSRZHKRh4yqEgmHMbrdbOntw2W1ctKECfkca7/fD0BxcXE+/DSmDDKZDJFIhMLCwmPGkw+X
    A/5Rrrl/FV977Ak+u3Qx0ntssifLMjZR4JxZM5laUf43meb4kY4Ti6J4SE8mTdOy2ipXwidJEkaj
    EZPBwCf+dBeVLhdfO2UF0yqOniU1FqaIRqNkUikkQcBis5FMJo9a3qYDV8ybe4iJNjb0Kx6PY7FY
    DlksnqIiXu/s4v6166msreb6JYtwWSx4ozE64wkCvX1UlJQQ1DT29/WzYuIEHrz6KiSjkTVd3Xzx
    d3dw9ikruPPSi0hHo1R4PFTYbPzvWaeza2CAtkAwq/U0jeoCN9fOmcVrXd3sVzXWHejEbrNxck01
    VrOJ2VWVvNXXz+4RL7v7B5jXPJGrmhq5c+s2ulJpUok4N7/+Bn3+UZAkjILA9xcv4IYXVmcTP0QR
    gyQRUBR++9paYsaD4smKwpkNdTza1YNkMNBSXETPaACMEre/tjbbpicQpCsY5MbnV4NJglw980hp
    KYFwmE/NmkGR1cqtq18BTSM+fy7nNjWSUVXKPR4m/+wXRBJJRgMBHGYzSyc0oCSSGHUt32dLVVUk
    SUKWZVRVRVXV92y5hWIxtvT2MdjRcdx4/rHi06lU6iPR3+tDYx+MVQG9wzdlSKfT6ILAi2tfx1Ba
    wnlTJjHB5cBisx81rihJEuXl5dz65FOE0zIXTm9lYUM9SjqNdBhzfvhfO51OIpEILpcLVVVJJpPI
    h5nXCyZP4utnnY7TaMSTq/19ad8+vvLI4wiCgMdiJqOD2Wbjtd4+Lr//QYrMJiZXV4PDQTwSYXp9
    Pb0HDqBnMmzp72c0GMRx8L0JAlsGhrhn81ZmVlZwQUsz23r72BuJ8IM3N3JB6zSmlpexsbcP0WCg
    tbyMFYUFvNDZzb5wFJfdzmhGwReNZQGbUZhTXJTdBCa1MBwIsHc0SI3Lye5ojP946lkwGPBYLQQS
    ScoLC5mQi8tmFIUO/yiqphEeHeVWrzfrGo1pJKOR5iIPBwJBVF3n2b370DMZBlMpUplMvnNmvdOJ
    3W4nHgohWa0kZRlBFCgqLialKNz01DP0+kd55nPX519DPB5HURSEXM+vd6t9x4aNi4JAsc3KN08/
    lf4F8/IdP9+NxONxdF3/yDTo+9BmbEmSlP9wz33nW9gkiUklxRhNZkKhUJ60Orww3WAw8Jftu+ge
    GcFttRJVFHp8Pj57yspDtPDhIB4ryh8rtbPb7aRSqTywAcocDr5/ztm0d7STjsUwu1xYjBJxWSaR
    SuENaKCpYDSCwcCLu3aDIHCFUQJRwGI209PXR01NDbv27+cnG97ipa4ebjx1JSlgx9AwCALtPh8/
    8fuZVV3Fb847hyvuvg9vIgEIPL13H4qqMr2qkh39A0jorKws5+n+AeLJJPF0Ogu0nFavL3BxWmMD
    m3t76QqGiKfSIAoMxrILFZOJEpuV6YUFRBGw6ToOgwi6ji4IJMaaGojiIQz+2IYzFI1ltVzO98Ro
    pCcQyL1kAUXTWN+XDV/ZDSKD8QQ/OedsTILI5Noa3urs5D8ee5LoyAgm6QvvfA9VRZSM+UYJ78b0
    jcVi+dpkQRCoczq5rrT0PWtgTdM+UgkmH4m+000lxdR5PNhzoyvHwjkmkwmfz5dnkMfIjlQmw6ya
    akocDn78ymvcv20H/7f2de58cwMtpaVMyPlcR7MGrFYrXq8Xh8OBJEn5uUZj5rpgMKCm0ghadrTK
    1MoKVu/bT/foKEuaGvn0ogUo6AxHomiaxiOf+XfMAry4q40ZTY2c0TwRUdcxShLPdfeyq+MAA4kE
    ZoPIcDSWBweaxnAkyh82b2GHz08ykyW/NE3DJEk0l5bQPRqg3OHg0knNLKuuotnlwFVSykgwREpV
    qXG7Obm6inV9A7wVCNI9OkoixzdU2awkVQ1V1/l0UyOn1tSwat9+QskkF0+bwisDQ6QzmWO22Rnb
    DmVVxWW14raYsxuIIGS3yLG/03WsVgvntkyk0u3mphde5Cdr1iEYjVy3ZBEes4mVk1rw5eq4E6kU
    33v2eR7Z3cbchsajstOHg87v9yMIAk6nM+8GvZ8m82MdRhzHacz/YZS/u6eu6+9+ROfxTO+x6QRl
    ZWV539rr9TIyMsJ/rljOjy84jwUN9cRlGVlVsz2XozHSOc2i6zqarh+1p3JpaSkjIyNZkifXcSIU
    CqHrOtF0mqqbv8eEn/+KRzdvyZrZ9XWcOnkSF02fxndXnkyxxYoiy3xq8iTOn95KVY4hHwiFkVWV
    opIS/rjhLSaVFHPXJReyoraGncPe/PVtJhOfmzWD3y1fQiyVRhkj+nK/r7Xb8v8fSsusae+kQBT5
    c3snT+zYSSgHvr5olLv27GPN8AidwRBnTGjAntPSvYkkcu68v2vbx/Vr1hGSM4xmMnz3zbeJpFJH
    BbDdaMBjMoGus6K6kjK7jUgyicdkZlFFGRwehhEEUqkUOgJGSSKRUQgkEoxGo2jpNB67ndOnTqHD
    5+NLDz7M6n37Wd/bx/P72gmlUoecaiAU5rr7H6T1h//DgH+UkZERZFmmvLwcl8uFwWB4zznWY+tA
    zmTy7tNHrT3Q39Wcfru3l7ahESrcLj52ggP0drs9778Gg0EURaHe6eCJq68kmUrxelcXRe4CmnN+
    98/XvMbOgSEumTWDs6ZOORLIZWW83tZGbUUl5S4nbrcbv9+P0WYDiwWHSUI2GLhzzVqmV1Zy27kf
    Z8v+dq585HF2joyAwUBz0wTa9uzBYbWCSSIYi3HNo0/wZtsekEyAToXbTUtJ8SGseyKVYstogGA0
    mgewxWikzGGnLxSmOxJlJJXVet5kkvv6+vltxwEG4nEQBERBwCAIqLqOdhCYnu/szid4ABgFAUGA
    zEG4y2g6ncHgkabzmL+oqIhmIwZR5NW+AcyShM1sRlMyiIZD/8ZkMCCrKp2BIDO+fxsrZ88inUwi
    AK8e6GTH0DAzG+rzsXlRFKn2FHLO5ElEcimXh2hdJcP2gQF2tXeQ0VSqDiJG36/88rV17B4aZnlD
    PZcvnAcfwXYDf1cQ/3nDW/zm+ZdYNG3KCQfxwXKwn1ySTJJIJCizZ0v/TAYDaBqrNm/lrZ27qHK7
    jgriZCbDSbf8gIqmRp757HXMqq6mpKSEgZERFkxoRM1kWL2rjXu2bScTCFL/nW/x4Jat3L1mLUhG
    VkybwiTJiN3hQPB6KXIXkMgo6OEwjdVVJDMKQ5EIQ/E4Q5HIO+ASRUrsdoYjUTaEQvmkjJSi0BMM
    5U3qMbY1o2n0JhKMhiNZXxiwGQ2sqK2lJxxmh9cPuflRBoOBEquVUTlNRs5wSmU5HbE4B6IxyLXz
    cRqNNJeWsGlw6JjvN5nJUGy1EpXTLKisJKLrbO7pZf+Y75zbeCaVl+X9fOx21uxvz/f1cpjNFOW+
    UyQSYfUXPwdAOBLBaDRiFsV80/kxcRglrl44n1NbmiktKDxmWEnTNdwWK6XOv+5Pr9qylTd37UY6
    43QuXzj/A689XyzGjsEhqtxumktL/i71BH9XEE+rqGDZjFZmVVaCquYX3YmWjKryWscB9o14Obm5
    iVqPh3s3buL2da9z7oxWfn72mZxcVUmt08G8mhrSySTSYW1Sw8kk2G0U22zEQ2H0ykoEUaTQ46En
    FGZ4xEtnOILbbObslcuzg6/7+ikoKcZuMdNSUMDMygokSeKezVtJyDLBZBJNUVn3xc+ypbeXLz/y
    BKLZxITiIsqLiujPDfVOKgqgM7GoiKScpj+RPJh+BVEkkfM/04pCOpo5RHMmFRVZ13FZrUfQeFaj
    ATGTJa5cZjO+0WBeM+tAqdXCt1qncMHA4DH9YUXTGE0m8ZjNtIfD9IfC71xf05hRWUFnMMjA2Oak
    66BpaIoCOd5C03UOeL1UWbL8xoaubkKRCI/v2Us4meTaRQtZedBGn06ncZkkrlu65Ljf/tr7VxFL
    p7lk1kz+85QVx2eyMxm+tGwpF86YztzDhpm/X1l3oJOLfncHVy1dwm8uveiQ0az/FCC+bO5szmmd
    ikk0EE+l0HX9XbOP74lhVBR+v/5NHtq0mQllZbgsFjr8fqIjI2wt8qAD/3HKChRNQxJF/IEARlGk
    O5mkMximzOnIzhYWBMrcburKShnxeikvL8cmSfzfJRfw+NYd3L1lKyUOO2dNmYLLZGKL28VwLMaC
    hgZOa2rkyfYOXuju4eUDXWhjvreqEE2ncZlMoGs4LBaunzaFZS3NROQMP371Ndb7R/HFE2Q0HVVT
    82RChdOJxWigK6eRD5bmAjcD8QRxWUYVRV7o7MJkNB4CblXXs2WFOY25MxQmcnA7HF3Hm1G4Y8cu
    zEYDaUXLQnsMiGOkG6DoOt5UCpLJHNGhIxmN/OcpK4hnMuxct56olq3+8uTaAl08cQLfX/8m3ZEo
    8XSawUCAvQ4Hazu7+OWatUTTKfp92YSPUydNOoS8UlUV+7uYR/xmVzepdBqHycQXl590zCyrWCyG
    rut88gQXTkiiAavFgsNs+iuk4EcUxG6rFfdBZWWJROKQEM4Jo9wFgdrCQqbnShPTkQhfPP00vn/e
    uSRTyWxHCpsNl90Ouo7Nbgclwy2rHuG3G99iRfNEvnv2maDpmCWJ4oICrJLEpXf+kVcPdJJRVU6q
    ruaqWTO4Z8cufv36eh66/BOs7+jkmiee5q3uHq5ZOJ9PP/wY8cN7RhmN/OS1dXxm7mzQNJbV1fK2
    f5THOl/k/JaJuG02YvFeMBgIpd9JSXSaTNy0eAFrBgbp8vqyYawxE1wQ6IxEEQWYU1XJ5sEhEIRs
    kYOuYxQFFFXLm+b2XEZWeyR6BAkVTaV4YTiNputYJCM1Lhftfj/VDjspVcOfTh/BVE4rL0PVdPZ4
    vTyweQsZTcuazZqObhC447yP48rInDRnDj/bsh0iURKZDLeufoW6Ig+v7Gsnk06DonLTmR9j4aQW
    WnNJPclkElVVcTgcJDMZfvv6el5o28vNZ5/Bovr6Q0z8R7Zs45IZrdy9cRPdgSDRVJqOiI81HQco
    dTjyOdqJRAJBEE64AtFUlQUVZQz84LuYDIb3FJv+yMaJbTZbHsgHj0P5wOc1mfjJBecCsNfrZTgU
    xpxOExv1Y3W72RNP0Nvdw2mTWjAbjZgUBUkyYbJaQFF4dd9+Lvf7mTShkWc2beY/PYX8+pKLSKoa
    /mgMdB2L1UqB3Y6eS5e0SSYaysuZWl7KAa+ftfvbSanqIRMNBUAXBIa8Xr791DNgtfL03n35+97r
    H6XE6WBqZQV7RrwczPMmFYXtg0P84OSTuKqijIe6e9kUDLHTP5qNx+Zo/81jPujYZuZycs7EJkbD
    YV4ZGGQokSR+jIb2Qs6kVnUdURBIqwq94RAAEx12Ygj4vb7D4usiDXYbF01u4connqUzFM4r7BuX
    LOCGxYto7+6msqICg9FIud3GnrHnHfGyd3gE0mnuveFaLpoyGS0XEQCIJxJkFAVXLmarqBobu3t4
    cdNmrl68kIV170y+GPSPcstzz9Ph8zOxooyljY04zCYeaO/gy488xpyaai6dPZN4PI6qqidccei6
    zmggQGlJyb9enHisntcfCtEbiRJKJLFIUpaAOgFSbLdTX1RETVkpmqqyrbOTax56jD++sJqLZrbi
    ttnwR2N8b/XLRFJpInIGTdcJJpIE4nF0g4F4WmZDTw+LGhp4cXcbTpuNT8+eQZ3DQSIWZV5tLRNK
    inlizz5ebu/Al0jwyq425tTV4jGbiWUymA0GltXVMhiLU2G30xdPZCcQ5nKTEQRSqoo/HscXj6Mf
    JU10TzBIgd2OMZ1iZnUV0Xic/kQim5Bx+Aao6zgtZj7bOpUJukbUauPlzm4EUTwq/+qWJApMJmKZ
    rD9+ZkUZ7eFs+yGPxUKp2YwCDCQSh17DKHFGQx2XLTuJN7u66A+FEUSRMqeTkyc08tLuNn6wZRt7
    QxFmVZRTZLPxRq6rijaWIKKonDFlMi3FRazu6mZDdw9FVgv/9eTT3Pjs80iiyOyKckaDAeScBdHu
    9TEaSzC/upKBgQEiiQTbvT50QeA7Z57Ot8/4GJLBwM7+fjpHA0wsLeHCqVPIqOoJT+TQNA2/35+f
    DPlPzU4fS0wmEz3xBAu/+W0Ej4dnPnsdZ06ZfOLN+YICmurrWdHchMlooKKoiFQ8xmgqzW9ff4Nk
    RuY3n7iUIquF3QODaKpKj9/PQzt3s3t3G3fd+2e+dv8qTAYDiiBS7XZx00lLmN7QwHVPPcOqNzbk
    e2NhNrOptw/RaODUxkbKzSaunT6NS7xetnp9IAhM8BTSXFDAyz29yJkMF0ycwIv9g8QON1lzIE6k
    M9y2dj01VgveRBKr2UxcUY8KYASBSDLFzevewG6xcGFzU5aBFgREIVvnfbCfOxYzziJc4NmBobw/
    HUineW1omMVVFYfyZIJARFH48RsbKbTZGcqx3A6zCW8szo3PPA+CCGKWUX9xXzuiKPKtZUt5dM8e
    Ng0MZadEqApyRsbtdvPlh39B7+AQf77yCiKpNN2BAL5wGIPBQHVFBZ+tqKDD5+dnz79IKJ7gy0sW
    YLFaqXO7efqz171juSQSqKrKNUuXcM3SJaAqxFLpE25Cj415LT5GAtG/DIgBCqwWFs+dQ7Hdjvtv
    yOhVuFz87OILkdMy97z8MorNhg24Zt4c0HWunDcHq9lMrLEBkyDQEwxis1p5u7uHwZ4e6os8dEdj
    3PLCS3gsZiZ4PFTu2EVzZQUzmyaw3edDlzMH4cnAm319VLtczOjuod7lIp6WiWYyBJMpOoUwGV2n
    zOmg2GJhgqeQ7QOD+U4a+eSJXCZUQpbZlxuPEjxaRpWuM6+kmLfHQkuSREJTeamrG0QRTdfR9Oxx
    VTYrMUUlpaokc//lJQdglyRRbjEzkkpzIJY4ehjVZOKPO3bS7feD0Uh0bKC4KOZTQIciUW58/gVA
    4A8fP5ORRJJEJIrF6eDshQuYXleHNxbj/OaJxOtqqbBa+eopK7h84QKaPIV4/f6sn24yceaUydQW
    F1NkNhGMRLA7nPmh7NFoFF3XCcsyJrMZLZFA1DTMZvMhI2xOkBGdHY3jcPxDG/l9aIaMy6qKNxrD
    IArZrKKc2XMi/OQ3Ort5ZPt2iu12vvGxU/M/v+TOP/LYjl38W+tUfnraKdy3czfbA0EunjObMye3
    cNNjj9MXjbNjeBhfJEKZzUZ3KEwomXwnRJZjbutLiik0m6l2ufIL3SGZuH/7DkCgxOXklOoq1vUP
    4EsksprvIE3okCQyqpoPSYSTSSZ5CplRUECRWaIrHGVUlmmLRIkdb0ibplFis+FLJg8FeE47v0NV
    q1zXPIGpLif3dPfz9ujoURM8pricOCUjW4JhMqp6TMbVIAjcvGg+G0Z8PLO/HYCr587m0T37CI7d
    i66DLPPHT1xMNJFkXyrNb1a/zHmzZzESixGTZYokiTk1NdywYhk1DjsjwyMYjEZcBQWYTSbMY7Fj
    TSOam1JhNptRVZVwOIzdbkfVdT55173sGBjk+2edzuXz5h4zeeX9yE9fWcPdb73Nl05awmeWLP6H
    Y+dDo4lNBgPVB03vUxQFv99PyfsgCpb/4td0+Pz870UXcMmsGWwfHOBnr6yhqaT4EBA3lpYyqbSE
    jUPDfOXF1bSNeNk2OIQmipS5nDywq42uXO9pNJ2hSAxBFDitrpazp0/jD5u3sWtkBJ3sOBRzcTHf
    XL6U2uIS2rq7+O7rb2ZtT1FgNB7n0fYOZFl+Jz5+ECBi6TTkRn6KuaSPjkiEnmi2e4hJFCmxWokd
    L59Z15laXMxANHrkMYf/22Dgz129THQ66EkkjrrIXZLENc1NrBvxstE3elwg6MDDnd34wpE8C37/
    rjZSh/nrZrsNj9XKFv8or3YcAIOBN7u6CCRTKDkisCsUZnZRIYXTplJ3UL/pUDLJtx9/ii39/dx5
    4fnUVZQjCAKxWAxZlnG5XBiNRmRVZcfAID2dnYTS8gkFMMB+r5fte/YyMHPGhwI7H9oqJqPRSElJ
    CV6vF6vNhs1uR+DddWiIplKEk0lkNauxyp0uljdNoDpXWqbn2MQfnftxvrViOXN/+gvu39WWHTFq
    MnHXxre4a8NGCh0OzJJEhcvFJ5saeKKnj70jXoZiMX64Zi0hOUOBzUY0lcJlNjOaTvPkvnYuNZkQ
    7Q5e7+vPZ0tpuo6cSvGz00/lq6+8higIFFmtxNNpEmmZC2e08lTbTkQBVF1AQ0TTNWRdQ9UFEkBE
    TmEUQdVBP8iulQQNg6CR0kSUjJx/RwIgCDq6DkZBxyBogIAOKLqIrGnszsWNRUFH14V8XrtB0Ilm
    ZG7avgunZEQ0iFkz/CgiCgKarrN9eOSQDSN50IYjALqmsbi2hsGMwu1vbiSjKGAw4I0n8n8niiI2
    i5npU6fwpn+US370U9K5CRRTaqqpdLt4edNmbP9+RdZfDwQwm835MtZEIkEoHOYPF52HxemiqfiD
    d48ZWy9jb7yluJjFM2cw8R/ARH+ozenjyTcefZy+WIxzpk3lE7lxKMeTJ3bsIpxM0uwppMrj4TP3
    r2L17jbOnt7KTStPpieRwBuJsnNwCFXXiKdS7BnxsqOvn//P3nnHx1Vfaf972/TRzGhGvcuWe7ex
    TTUdTA0sJCwJCSywG9I2yWZDstkkm+wb0jaNFAibQgoBQkJIgdBMccW9N9nqfaTpfW57/5iRLLnK
    YHYhy/l8BAKNZkZ37vM77TnP+eSlF/OjtevIqRpum414Nsvdy5byr0sW8nJnF//0x6dhlCMsCKBp
    3L50CerIMI+0d1Hh9RLP5cgkk2C1IosiumEUwGGanOUvpTWZQpFEpvh8bOzpRZRkbp09hY97vsBi
    b5gvH1rKFw+s4KaavVRaUvygcwkeS5rPTtlItTXFcyONbIlVkNYVBMHk8QV/YXlFkHt3nctPe2YR
    Ut0s8wfI5veyzDtAxpB5d+VBrqnqKt6R8NG9F/Kz3jnYRZVZ7gSzXQO0p71sj5cRUu18qnELh9I+
    /jLchKpLJ6YUGwbL6mppC4WIFHNhvZjLn11dSdyEvYNDNAf8qJpGTyQKRebWKPjtxZ63KAr85NZb
    uGXRQkKpFL/bsZN/efKPpPJ5ME3OaaznwqlT+POBVn7ynptpdrsoLXKuAaKRyBippvQECpmvx7b3
    9rGuvYP5NVXM9fvfcnPGb4sNEE/u3U9r6yF6I1E0w+CCqVOO0ZQeb9fPmwPAhs4u5nztm8TjceY2
    NPD0zl08vWXbBJ4yosCHL1zBjbMDvHfJYpyaRr3XS2twhHgx9w1Fozy+azd15WVUuJ2EsrlCXxYQ
    ZJmHN20pPJcsI2Fik0QyVmuBZeVyEkymyBkGGAYel4t7p7fwXGcXq4sV4Dqniz1da4jMdJDTwpzr
    7ePz01ezzDOAboqMqHaa7DHunbYNVLit9iBpVaAn62Z1pJblviBqBm6pPMAKXzcP9V/HZ2d0omXW
    cq6/AFzTgGyx3mQV4caKdtxynuWefq6tiCOIYQbTVjbGqtiTDPCB6r1ENBs7EwHaU6XACUTiBIH+
    ZJIra6qxyRLbonG2DQyCYXDV9On8YPtO7IpCiazQmU4jSBLm6FZFQSCvqiyoq+Wy2bN4atcetnX3
    ohsG33rpFbpCIZxKYV4bw+Cc5mbmVFdz3wsvcfPDv6TrS19A1TQ2HG5j/eE2Lpk5g2llgTNegX5i
    +w6++stHuOmyS3jsztvfcvh4W8wTp9U8leXlvHq4jUeff4HNwyNs6+nFrig0nUSyJZ3Ps6690Lu8
    Y+li1rV3FnbcKvLY4D6iyOaOThpKfXzxisuYU1VFiUXhL/v2o8gylU4nmwYGeHnfAeYG/FxZV8uG
    /sFCJXc0tB9VjiwSEjKahgG4bVZiufwRlURBQDUN8rrBzlCIlKoiCALRrErCtHB5oJO57giN9gSX
    BPqY6ogxzRnl5urDXOLvJaMWQmnNAEkAvyXHWd5hMnohDK6yZpjmjHN+QGam8kdqbDHyeuHx+rh4
    SzehwpLgPF8fc90R8kYGzQCHrDPLFeWC0n5skoYiGAzlnWwM14FogAluRT7SjhptM2UyqCYEbDaS
    qkZPIgmSxMaBQaymicNq4dDwCFntSPVbkWTUfJ5STwmL6xtIqyrP7d3H+kOHGclkGEkm6Y9EcNus
    JPMq/7jsLC6a2oxFlBjOZFnW2MiF9bWkslm+t2YdX33kUWrqarl45owzPnQQSyZxeL28/9yzaSkv
    K4TWb6FxxbeFJ/74hSsIp9O0lAVoGwnxyJatrN21hxJgqiyhCQI1NTVjogCjNr28nPOmNLOxq5st
    Pb2g6wiyPBbajoFQknhm/wGe2bOXdy+Yz+yyMiiKzoVzOUoKb3xuAAAgAElEQVRkGWd1NbUeD1fP
    n8d/rH9tfKMQZBlRECiRZaL5/FgltqWkhKwo0hWOFAbmRZGeeIKe2JGpJdMQWBZQubNmDZcHOlEL
    WEHVx4esR1q4o6wqA9BNkWC+Cb/UVsxzC19eYTu5UyRJVrHwpLlxeDRMJvyeJBjEVCtoFhB1ZEkj
    ezzZVlFkfzTK/mi0IB9b9LTpXA4UBaGY+44PwWeVeTFEifOnNpPUNH7+2kYQRa6dN5crmht5sfUw
    bUNBBpMpUDX+6/pr0Siom148cwa5dBqr3Y5kGJxVV0PP5ZdyVlPDGQdXJpPhwoZ6OiNRvvTMs3z5
    medY98mPvRNOn65ZZZmqkhK+sPKKQtujsoKhZJL5AT92lxOPr5RUKkVXVxfZbBa/34+iKHgdDmyS
    hCgIRDM5PnjRhTz46mqQZVw2G6l8vkCJNAxSmk4uk+HevzzDjzduBknCME2yqspHlyzi3KoKRJ+P
    z7ywCr/LRTQeR0HghrOW8MTOXSiCwHk1VfylrbPgJgGbKCJKElVuF7LLyYFRMbzxN5oAUVVgXaSa
    Ga5hzvYMoh0FwFHw6khkDA9WMYGIioiBiTAG7LEPVcievBACZE0PAjoWIXniME0wuaXqAFeWdVBh
    SbMhWs2/HTz/+Plx8W9Sx5dYxsv7jAPwgqpK/uOaq3nXDx+gMeBnIJ4YO1S39fQSTyY4UKSTomnc
    e9UVYyyrx7Zt51/+8EeCyRT5b3+DWCzG7eeew/vOXo58AjXU12uGYTA8PEx9fT2RXI4NHZ3FdbC8
    A+LTtayq8v5f/YYnVr3MLz72Ye697JKxcHlkZIREfz8er5eG5mYskkQ4HAbD4PmDB/n5a5vQ02le
    vPsOLnzwIRClY27o+TXVNLhdbBscoj0aI6dphaKLrrO0toYV9bUMJ5Lc+evHCoWT4o2Sw+SxbdsL
    M76CwMv9gyCJhYF8YE8sPpZXf+biFaSSKX60fedE5UXB5GBS4mBsLg32GDPdTtxCGwYSAgYCJh3q
    heSlaup4GocYLugoAy+nPsci289OWynFBKxC7NSHp6hzUWkPFhFWhWr48uHlILzBOqgosmNwiHc9
    9BOwWtnR38+00lJsikJW0+iLRgmnU1w4pZk9g0P0DI+workJ4a57UFxO/MUeuJ7JjjGlDgaDHBgK
    Fj7H4obJN1yRNk26urqoLe5Evqi5iZJ3XcdbsQ78tgCxIAiUOh0I/lL0XBZNVZElCecH7qKkpprb
    z1rMS62H8DmdPPehf8Lr9SKKIlXZHLZiiD2kG6QNsyBkh0gynR4D446BQXb06pDfzT+fdw7fXHk5
    q7u6WdfVzfSKcoZM2J5I4lCUQp/2KA80WmVN5XLIoshVtdW8OhQszCQXC2jbOrqQJbHwkmaBHOGW
    5YKUzjh4yUKObu0cdMFNpbQFGyGalVcY0ZswBOsEIC60PYxHGmDyewkENNOCLOSZrEiShoSCTl/W
    hW6O67eawtghdAyRZJw5ZBnTNCeywQRhjJGWV1WuntLEnsGhsZ5yRtX46559eN1ulrVM5R8eeRQc
    dlTTZDCZLKYrBq6SEnKaxoNr1/Pdx5/gvjvv4F8vvRj5DYJY07SxRQCGYRCORLh4+jQunjlj7NoL
    74D49EyRJD58/nn83fx5nN3chKaqZDMZGqe30BIIYJom7ZEo6e5eBgYGEBUFh9OJYpo4LQpYrcy5
    72s4FQs3LVmM1+kkFo9hlSTUvFpQ+ti7H6xWGhWZRSVuzjn/XJTLLmUwHOYLzz7Pax2dNJeXkxzN
    eQFn0XvIokBON5EliRkeD3FNI6cbEwpfz3d2jeXfUCBSXFVXzSOt7bitGpWOCBWWNA5xGFFswscO
    FOJjKXFA6hjtDo3dSB6pbwKAhXE/O74HFsmYpbiEIYRJgVhAN22oZp6pzijvqTrAr7sXgKiywBMk
    Y0gcTHpZ7C9l6zhxgfFW7XQQUBS2hqOoxlGhqG5Qqij86UAr4XF6WqIg4HA6SeRybGxrP9JJOOIm
    cRV1yxRM6jwl4PWytaf3yNz267R8Pk8wGKS8vBxVVdF1ndKjFFWFdzzx64jABIHp5WX0RKI8tHY9
    K2fPZGZFBX+4+x9wKhbah4dZ393DflFEzed5bPsOnm/v5MY5hWH9wumqI1kEPjB7Jstrq0noBk/u
    2g2SRLko8Piu3SBZiKgqn3p+FU63m/Mb63muo5uuVArBbqd/1AuMz1UFyBWHEEzTpNxmpTOZPIap
    dLSKSUbX2RaJATLn+bp4X/U+ZjjDCGaOOnHNBMByggbPEWJG4b1E9QoUIYtNiBUEM4+6ny2Cjl8e
    ID9p122iCCkME5Z5h1jqeZ7FniF0/Fzl38K+VIC791yCLJyYEdUWi+MsL8NrkRnOaGOeWBFFzmmq
    YWqxu7C+r/8Yjy4KAnpx93Ja1bArCpquo+by1Pq8/H7rNg4NDdEaCoMgsK69ne5whCpPCc7XsfhM
    0zQikciYyqkkSW8L6dq3HIgT2RzPF2dsL585HXdxF08onebbL7/Cqh07Kb37TmZWVLCgpgaAhtJS
    nqisIKNqTAn4ObxhE+u3bWd2ibtQGS0uA8vqGgOpFJIo0trewX++upYcJjdPa0Gx2lDTacKZLBuC
    wzAwSE8kwsGREHU+H3ZZJn1UKD06bTSlxE1bIoluGKwdCk7Im49ndkkio+scLK4S7c262Zko47ry
    tgmtoEkdcEBHpgSnlKfGHkQ3BJKahCSYuBRjQuy3NRpAM0WWeoLkjAKZTGSUAXZyUwuXkA/V7ydr
    enGLceySTrklz8bRItQJDpoD4Qiarhd48RYr8VQKwWKhxGJlcamPfDrNbJ+XlKbROhLG0LQj9FJd
    5+/mzuFXO3Zis1n58PnnFngDwWHueuwJYtksMyor+P6tt/DDNWu57IcP8LnLL+Ouc5afdg4cCoXI
    5XKUl5djsVj+V4ca3tYgHkok+Pjv/4AgwCv//JExECuiREtZgN7GBmRNZXBoiIC/sLjMIksT+sWf
    vvxSbli4gGUNdXz7pVdIF6uWmWiU7pEQd//paYK5PJVOB4qi8MfD7SgWhXuvvJ4Vfh+43RwaHuG5
    w20gCDS4HKTz+QKIRwFqGCyoLEczTPaHwlTa7TT6S3mtq7vQgz5Zoa6YHxoAgsHuWBkG8P9a1p82
    iCUBNkSq+H73IvxKhlurDpA1JD598AI+ULMPMOnIeDARaE35mO0Kcf/Ml6ixpRnKWWnPeFlUEsQu
    FZCcM04WjoNpqlgYRjNBFgwEjFPK0OSKCiOldju3tEylrCzAl15YxTMdHbxQTDNyuk5TwM/KmdP4
    /MUX8v11GzDzKg6Lwq6hIRBFItkcj23bTiibRdM0MqLIkoZ6FtfX8ZHzz+WxrdtYd/AQI6nU6d93
    Q0MkEgkaGhqOaVW+A+LTLmIVcmDD0Atazz4fCAIBl5MH3nPzhMeGw2E0TSPg8/HnAwe55ee/IJtK
    Y/74h8ytriKv63zrxneNPb6zo4M/b9vO/W0dxCNR1t5zF3ZRZPGDP0G22Vi1bRu91dVcVVvD9sHB
    Qu5qs1KtyCgeNy9ns1zR2MBznV2IksiOwSAz/aXopsm8ulo+tHghq6oqeWrvPnoy2ZMW6o5UOUWu
    rLXw+fpXTshNPm77AwHNVFBNneG8g62hAiHDI+fZHi8nrtr5fueio17Y5FDSz0uheqY5wwzlnfSk
    vAVXbYhM8wzxu4V/YpojchLPbI4dHuvj8xhWSyf9wYazWX60czdgIskyumGicyRPbh8eoT04zKHh
    EN+/+AL6M1k+9uLLBcZW8aA4HI2BrrOosYGtn/6XQkQUi7F682YGwmGcTgfTyifHaf7Lnn3c9/wL
    WESRr196EcvmzOHtaK8LxLphEM9myWk6XrvtjCr6TQkEaP+Pfx8rYIyEQiiKgtPpRJKkCe2DMX6s
    qhIKR1BEiayi0BEMsr67hz/s3M35U5vRdZMdfX1YFZl4OkM8ncbl82ICdQ0NWJ1OLILAhqFhNvT0
    EQ2HWdfWgWixcMPMGdw8pYlkOMz6kTDPdXZxeV0NwXSGHSMhuork/Q09vXREo1xeHmBhVSU97Z0n
    itsKGwNT6TFSyN5wmJYZwWIeLBa828kKfSI8Hazj0eF/4Pf9AmougSDnMYEn+meCYBS+jgsmg7hm
    ZUu0ugBIUR8DWX/ORXvaQ4sjcopyF8RUmXA2gWBmAfvpFDgAgTneEnaGoxM2RTSVuInmVWK5HN/c
    vJVDkWhBo+yokcoqn49yh51kIkH/wABer5et6Szt0RiL6+qOWbd7IhtOJNjW04vLaqWyppa3q70u
    EIdSae57/gU2dHRx33VXc8m0ljfNLQcCATRNI5lMIssyLodjQqUylc+zvqOTJw8cKIS7gsB537kf
    v9vN7n0H2NzZRXcoVBiuF0Tm1FaDJPHFiy7AYhhouTwP3nAdv92yjb92dFLl9eDxeDi7ZSqhRJJf
    79zFC53dGLpOvki13BGOsKAsACOhMenYRCZDIpUil0rR7PVglSRyun7cv2kMwEXv2JPyMJh34pA0
    QEYS8hPDUbMEqxA/EkKL8O3Opbw6HAXRBHlcrVmcDBnBPLbfK5gksy62xSu42N+NIpz8IMmbEr8e
    mM1ItgQk9bQ/2kPxiUVCDIP3Tm1mfyrD1pERXmrvAlmasK0RSYK8yqcvOJcF5WUkk0mcFRU8sP41
    NnR0cv7UKSyoqT51FGMYaKrKvMoKvvGua1EkCa/d9rYF8eviTgeTSb781+fYuWUri1pamFHqQ1VV
    rOMWQp/R6rQoYrPZsEgSP1y9lu+vWYtVLuwjGozHuePXj7J29x7M4pB+RtO5pLmJO88/l3lVlWzu
    60M14UML5/P+ubO5bvYsplVXI0kS+WyWudXVPHmwlfa+fnwOO12ZDJ9aupiFtTVsHhhkMBYnmcmM
    HR55XWcomSJnmlgkaUwjC1EkqRv0p9LHB/D4nGECpiRejVZzbmk/NbbEMaGsjhVZOCLZI4uwOlTN
    jkTFGe93/Oe0NTTZE6fsPYvALHeIVyLVRDXb6zqgx9pBhgmiwPZwlN5UmuFkEkMUJowx3rPsLLb0
    9oGm0h2LsXU4xD+edw4f/t2TfP/Flwjncvz6/e/jlsWLTqrPls/nx7YeNlaUs6yxgbPq69+yC8Tf
    NE9c4/Xw1N13EstmafR5cdrtYxQ1AJvN9uaU5iWJVW3t/GHdBnpHQqSyGRY3NjC3uppUPs9wIkk8
    m0UA6vx+/vnCC1h74CD3mYVZVovFwr+9soaLpzbTsXsPB4PDrJw+jXg2SzSb5dJ5c3ix9RAMDtHb
    1MCeRJJ6u41PLVrAYwdb2T4YLPKaDVRVBwFU4dhUQwesUrGYdiqaniky39vPB2r20mSPHjcvnuCF
    BRjM2DiQKj3DABawyTl+3T+L2a4QjlN4V5ukM98dxC2rx6QLxz2ojj6Yio+ziiLNfi/t0RiJfJ7E
    2FK2iVn4tsEh1n/qE6w93Mbnnv4r+wcGeWr7dtYUhQVimQy90ShL6uvI6zr/9OhvCbic3L58KbMr
    C/K3w8PDBZkel+t1LVz7mwKxRZKYXnGsst+oWFgulyMYDGKaJrIs4ykt5Zk9e9nR188tixcy7Q0M
    U9+6ZBGpXI5VrYfY88ST/PUf3s9Df/cuHE4n7//Vb/jVps0IgoBXEmg7fAghm+GuBfMIRiJYrTba
    BgYot1vpiifpj8fZ0NbO7uAw88rL+Pp1VzPjg3cXROmSSc761ncJpjM4TJPpPh/bBoMIwFyfh/k+
    H1lD50+9/QVix9EVWd2YHMZMkSZHjPdV78Mlqads9QhAWLOxK1HGG19NNxEpTlnl1qr9OCfxPmQB
    NseqGMnbJwD4rKI4/5Zi7/aEIW0RxDnD4GD4yBzwiUT/NvX0ctkPHkA3jYJMkNXKbY//vjBRJUkY
    uRybu3q4ft5cdMPg4Vdexef3c8n06czw+xkJh6moqHjLTSC95arToxfHZrNhK2ohGYZBW3c3D7y6
    mmc3b2VWRQXTAoGxm+R0Lqhpmty0YD7RdIbXOruwyRL7+wb4701bcFksPLVrFwCfuOhCrCUlTP2v
    +3F7S6hyOpnf1IhDlnjvBefx4t79NDkc3DN7BrLDwS927KbKonCotRWLYVDi8fDonn30JJI4JIlP
    rdtANpvjXfPn0hmOsKN/gF2R2Ck9jjlJ9OimQFK34JBUTGzIQuakdaFw3ophnvkb0TQFBGGy7xqk
    Cfoiha5C1tAJZnOT2n5gEUXcikxC1ciPY1qVWq2UWBQ6Y3HqAn56I1HEYv1j4kE5LsqxWLjvz3+h
    ezjIZy9cUSwAiqQTcXK53NjmzL81AP+PtJhEUaS2uobpVdXsqB1CUFVC4fDYiWi32ydsgRcEAfkE
    +Uk8l+PDv/0dqw+3seafP8y82lpePHCQh37yc2bOmIZFklAkCZtpkEokcPu85AyD1nCE1pEQL7nd
    fOW8s3kkHGYokeC1gUEQYFpFOc91dpNWFEoOtNKe1+gKhQpb50dvLkXhT/sOIInCsZpNRW9hEUXU
    URWPsUqyeGSe+HgHn6jz58HpPDvcxLZzf8l0Z/SkvWIRaEv7UM0zTEQQTMKZElaFGlhYEjxlYUs1
    4OLyfpocGfpzLsBENQx2Fw83WRTHhBNO5o1NE86rqmBjcJiUWmB0hbNZwsUaRM9IiFpPCTNKStgb
    jhAsrnRZVFvDwZERUnkVoRgdWj0efr15K79e/xrYbHhsNhqrKnG8CauC/k+BGMAmS/zHVVfw6Usv
    xu9yYh2Xj6TTadLjBMlHQTx+WsQ5Tmo0msnSE4kSTiQgl0PNZvHX1hBOZ3j3/HmkUin+uG8ftU4X
    ty9eiEtR2Nc/QG80QoXLhaRpLK+vw6Io7ItEGEkkaR0JgcXCzoFBEnkVU9cnDPqP3XSAcXTSaprU
    ul24RJFau40Xh4KMT1ZtpwCxWQTQB2r3MtMZRT2FI1RNONfXx8KSINvj5WhnGMxN9tgpATwa1qMX
    VsQcfWibpolNkkhp2tiBZhFFfBaFUC6PZpqUWiyEs1kigNdixS5JpFSNMruNKquVuKbRl06jIhLN
    ZEnYbHhtNuJqnlQ2x0eXLuErr67hcDaMx25n5bSprJw9my8+9wJ94TB5TaM7GqUvFmdRHe+A+EyE
    2V67Ha/92H6i46glWaZpju3KGQNuNIoA5A2dT56znGtmTuf+da/xy2076Y5EsCsyvd09LL32Kqb4
    /VzwX99hh8XK2fk8Hz5nOVN8Xpp8XiyqRmcsSk7XyZkmt8+fy4HBIf7S1lFQqMjlj0zYnIYFrBYq
    FIWhbL6A9HG4SpxMXhbAFPBYUizzDBwzR3wi0LvkPO+pOkB3toShrPONjweOg2ZSt2CYAuIpntMi
    Qkd+Dn1ZzzGJgyAIx8jqmkzkcgsC1HtKcEgSg+n0GItNEUSWlPrYGI6M6WMnNY2NQ8Ejs9hWC//0
    1J/JF04NVNOgPRrlxdZDWEWBB2+5mV9s20Gpw0Hl24D7/LYA8ekC3nmUyPdouO0ALp7rx9fRyT2/
    ebyg/Qy4PB6uPvcc3LrGs7t2g6ygKAVv/pWXXqUnFuPBlZfR4vPx8O59bB8Kgm5w57w5bMrlJ95Z
    p+A9H69VsjcS5YAgFnI08TRzLkPi442tvLf6wKRyaIsAzww3873OxYRV2wkBrIgiVlE8uUb10aG6
    ksU5iZ6vKEBrys0nDsymPW1h/HjGcaeITBNV0xgep7UdUzWqXC5aI1HaSIxFK0OZDKuGgvRnsggC
    VDvs9Geyx6QwpijgkWRiqkoqr7Kxu5eNnYX1MIubm7hu4QJMoMRqewfEbwVTjmKE1ZYF+LerVxaU
    N/J5ymw2/uGC86gsKWH3M8+CrjG3opZPrziPzz2/imQySTabA0lmz2jVVJH5wbadDGWzE1hDfqeD
    SCY7djPai155dB5WAHwWC+FxrRDVMFF5vYoPQrGFZCDCqWeDRdgWr6A/50I/SXFLNQw0wzi980Sz
    0Jt1oSMgnuJIyRoKfVm5kF6c4tyqdzqY7/Xw554+AiVuApLEgVicveHIMUP2ummOseBGr+3xDoVS
    h5NSUSQ2+jmMRlC5HKm8iv+Mb3t4B8THtX2DQ9zw3z+jdf9+wj95EN8p9s+2h0Js7+nD57TzlWuv
    PvLBaxqhcJi8xcLlM6ajXnsN3131El967kUOhcIgK/z7K6up9nq5uK6GjNLEhs4uDkajhZtIEHAr
    MnUOO0srK3n0cDt5XafMai3s4C166Gk+H62RCOFcnhKLBdM0J4TLp5rnPT6GdZ4ZbqI/5+SqsnYu
    8/dMIqyefAX59MrTAtXWJPIkcmIRE6ekooj6MUU2scgNNwFZFAhYrWwIhQuhb17FtIljqdNYWiVL
    pDUdURDwKgpxVUU3TYZzueLIpzChDTWUTDEETK+qRBZF9vb1g6bxr1ev5OzGBpK5wm7k327bwYNr
    1zOzspyH33frOyA+Y60MYDAeJ5rJUOv1MFBTg5rNMaCqhNIZSmw2ar2eY/SS/rBzF5/66S+oaWrk
    0Oc/i73IEJNkmfLycpLJFHNKfdSfs4z7nvg9O3P5InUPBjNZFDnJLy5egdfn44ESN2v6+lnT0Vk4
    2W02llZV8fv2DnKaRq3DweXVlTzdP0ha00jk8wi6hkWSyOs68XGk/FGzSxIuRSaYyU2eiCFAzpCI
    a1byhnTqXzOhzJIpDvW/Oe2SUz2raUKVLclXp6/hT0NT+H7XoiNALs5UR1SVnG7Q6HCyoizAd1oP
    gyAQU1VixWsnFiv6ecMokD9ME8M0kQWBd9VUsSEUpi+TpdRqYbrbzfrg8DFh9cGBwTFg+0pK2Nrd
    w3dfeZUfr13PgYOtYHeARUGWRP5W7U0HsarrHBoexjBhasCPrTjYfe2DP2FaeRlnNzUyo6KCgWSS
    J7bv4FuvruGaWTP58S03U3pUayCZy4PNRktZgFw2h6nrmKaJoihYLBZcLieGqhIaGuLuSy9GNHT+
    dLCVgXiCi5qbuHbaVOxlZRyMRLDoGvF4QXVyXn0duzo6eTgaG7tJml1OVEFgQamPg/E4CVVlKJfD
    oyi47DayhsFQNnfEOxR3+iqCyIkEXERBwCUXJF9HCznoMrdW7+fjzTvBgIx+ChAJ0J91YZxpABsS
    CCZfajuHpd7Bkw5BCAIM5Zz8v8PL2Z0sQxsf1gsCg5kjKYqJyd54vMBdF8eJ7lNga9U7HfQWdyYv
    LPXSkUwTzGTYFoniLLYaAzYbLW4n64eGjz1NAIfVimGazPCX8tLBVl7auo2yqirmzZ5Fi8/HuS1T
    qPf73wHx67VwOs2HHv8dOV3nl7e9l5ayALphsnXjRoIzZzIUi/Fadw8liozb4cCqyAiYjIyMYGQy
    mEBZWRk5XacvFgdManw+dF1DxUTXCyOLVkUpeFVZpjkQIJrJ4LVZEQQRDIN5VZXcOn8uq/Yf5Hvb
    dtCdTDEYjyOLIp+dM4uPBIOExpEUVg8MwsDgkVaTaRLN5LAoMl+cPZ2HO3sKpIbRPN3poD+TPcKj
    PsHNv8zvoyuVpjVe3Jck6jwTbMYr51jsGWK6M3LykUQTmh1RJMwzBmSrqHOevxvNFHk1VM+uRNkp
    xhELa2CG8k6GMu5jhy7G/f2dqTQ96SLv/Khrk9F1+jOZ4jL2AsvNwARRpD1ZmAlWJJHudAHYiAKo
    KhaLBdU0cRYjtu5QCEkQ2dDTC6bJZ264nvOnTaXO5calyDQVaZejIfw7jK3TtLyms6mrm0wuR8/g
    AD5MAmVl7P7B92gq9fHR3z3JSxs3Ec3nuWvxQi6oqsSuyJTY7Xi8XnRdp+3wYTTDwG+RwWbjkc1b
    +NqlF2EtyqgIgoBNEvnC86sYTKexKzI9g0MsbG5iXnUVg8kk2/v6GR4eIZnPcyAULqhQAjrwxfWv
    UWq3E0pninOuBhdMbSYcjrAnGsNvtTLXW4JNktgYivDV/a2M5PITVCtHcgWvnDnJ4INumGwMhQvb
    IMb0t3ReCtezOlLLEwv+xCxn5OTFLQE2RqvOaH84YMnw1KKnyBsSd+y5AoeonjLcTusyaV0+8cjj
    uEJVIVQGv81K3jBIqBqKKDLd7WJPNDZ2LfaN6nGPHgSGgUtWiOTzZEyTCpeTuxbMY+tQkBfaOsiq
    KolsDlMQC3uORRFUlc9ceAGqpmGzWHC53fzXqpf55abNrGiZyvdvuvEdT3y6VuvzEvnGfcVTtaAB
    PTQ0RD6Z5NJHH+ecpkb6H/wBFW43oiAwpb4YhqsqIyMjaJqG2+0u8LUDAawWC7lYDFEUcdrtE9g4
    cVVlpBgi15aVcU1VJduDQQAiuRyy14stFC4sZhNFzqmtYW1PL62xOBZR5OzyMgRZZmNwmLVdPQW3
    J4rkDIPBTA5ZFMgaOpGj6H8AWX1yleC4qh3nRhfQdQvb4xVcW951csKHAL1Z95lkTSNi4rIY5FSD
    R+Y9g4h5yuePadZii2vyiXY4lx97XtUw2BePH6XBLWCXpcLQiGmy2F9KXNMK11sQCKbSfH3DpuKO
    5UIO3Rc7iv5qmki6TjgWIy3A/a+8yo82baEvHKH+KMG7vxV707N9gYL4u7W4JQGgoqICq9vFa7v2
    sOHQYfp6enl5507mf+2bXPvAQ+zr6CASCuFwOKirreWRvQf45NPPoqkqI5+7l85Pf5KvvfgSzn/+
    Fxbe93VCwSChcJiuL3+B39z2XsqcDgIlbupbWujTdUxg9+AQD65eW9B6kiQMw6AjmRr78POahiHL
    KLI0FnaNhrVJTeNAPM6eaOy4U0knuo+l0wrbBNxKflLgdI31ct94WGgVdWySNlZwlAXjlESPwt9m
    4pLU4qbFyRc0J6Thx3kZv8VSCLtNkxUN9dR6Ssb+UkWSMDEn9KIFQCqG6jZZpq6inMt+/it2JZLU
    NzVzOJ6kLzjMJTNncM/55/5Ngvh/bReTYRjEDZMr58zmykULCCZTfPXFlxAQ+PtFC/G5XKTTaUaC
    QT781B95eccuPF4v18yagdXpZPfAIFuHR+hPJNkxOI5DpG4AACAASURBVEQJJtHeHiolgeG8ygu7
    9/LnHTsZTKZxWyw0lfrYMDjE862HEMwC13Y4lRoL2wBqHHZMXaM7ncEmHcX9PXpzwzg75rFFj1Bl
    t5HSdRRRnCgYf5TZJQ2PkuFcbz/n+AZPyp02TbigtJcN0RqaHFFcsoaASa0tiVNSsYk6WUMeN5og
    4JJU8qZUuOVN4Qj2DZl76rdzT/1OWhzRSet7mUCFNc0NFYfpyHg4nPZOXLVapF6+HovnCwdUs9fL
    +VUVHEok6YonsEgSiyorkESRaDozVtSyyDKVJSXEMxl+fP01/LX1MG0jIyyorWFLTy+/2LSJ+Q11
    fObSS7lq9sx3QHwmzWW1cu3c2Sypr0MSBBw2K02lpVw1ZxbLpk7B5XDgdrvxlZbSl0jg9Xq5dHoL
    SxoaQBSZUllBpdfDywdb6RgZ4aWDrdR7SqiuqCx8yDYradMklUpx5bQWHrr5Rr719HNkJQm/zYrH
    ohDOZAnYbVxQW02d3cY0V6FPbRFF+tOZSTO3jkv0FwQSqopVkrBJUiEPPkFV+LzSXr42fQ3nlvZR
    Ip/cGwsCJHQLF5X28Plpm7ky0MYy7yAfadjOyrJOLi/rJKzaOJwMgCGDKbGy4jADORdWUcdvzZDI
    O0CzYLNkmOse5qMNO05boM8lGwSsOfYlfKyP1BwpspkmAZuVEkUpzF6fJuFk9LCMZLOs6uyiq1gA
    1A2D3licaDrDlIoKGv2FTociy/SPjNBSXcX7Fy9kqtdLVtd5rbOTZ/YfINjeyb+963puWbyQv1V7
    yzC2Ak7nCcOdL668At0wsMoyFlnGAgQUhYDTydUzp/P0wUMMZ7N8d/deTI+Xv18wl0unNPO7vft5
    8XAbTtMgODzMh84/h5/v2UdvOgOazuXTptJZ3KurqRqbwlGSmoaEcHIAmybSKbzr6A2Z1fUj7aQT
    WEpXuK68E5OJC86Oi3kT6opeN6dCvS1Jkz2JYcI0RwxRgCmOKBeW9jCYc+KUVJZ5B7i0tBtZNJjj
    GqE76yZnSFRa0zhElbQhYxcnR8+UijK3v+lv4YnB6XSkPROr5IJANK/iKO6yeh0h2pGOwDimXr3T
    wXAuT0ZVyWka6XyevK5R6nQwGA6j6QZ+h4NPrryC3myO7/zhj1yyfCmfvf39NPn9f5NV6bFL/nZY
    Mn4829TVzUX3/5AypwNREOgIhSGb5Y7lS7nvyssL6iIeD9u7e7jxpw+T1zUevv5ashYrn/jLM3T2
    9fHrD7yP763bwOau7nEkAuGk6aYsCNzWWE9a03i8o+uU8rSTOkkFgyprit8u/DOLS4KT8ooCJ2Zk
    CYBmCmimiCiYY3ROgcI2xKOZZfokBy/sEgzlrPymfyZPDzezKjgFRO2UAxinw+O+oqaK57p7jwyh
    FPnWNotC3jhS0Bq7drKEphvIoojPbucnN9+IVRTZFgrz4zVrmVZexu/uvIMS298uh/ptKyyk6jrp
    eIJ+XUcpcma/ccN1XDF7Nr6qKr7+/Is8f7CVKxsbiBVlX1oTCSQzjp7LgSzz2+07C9XN44wdntBR
    APvicaZ5vVzb1MCq/kHS+hvblKeZIj2pUtK6MulS1SlayUiCiSQceV/SaQD2uHm7DO/efhVPDk7F
    RCjgaJICeaphTDqsjpjgcTqIZbI4ZZmza6pY6inhvt37jwyXjPusNL0Ack3TGB4ewVAUrpg3l5J9
    +/luPs8L+w7g+fineOZjH2HlrBnveOL/KVvf0clPN7zGxS0t3LJk0QmrvH/dt5+rvvU9mqqr+Piy
    s7hh/lxCySS9wyN8c9MWVh9uA13nm5dfwr+uXgeahtNmw2uzMpBKYYzuSyo+/2j1/JRhoDluidiZ
    CNF0mU9M2cgH63bS6Iiflv70/5TZFKhddRehvJ2s8ebpU1kkkQVlZWzqHwBRLNAyNQ2LLHNVYwMr
    aqv4/vZdtBdzZUUUuXvpEq6cNZP9oTA3LZiHV5LojUT45pr1rGlvp2skxLMfuYcrZv5tgvgt6Ym3
    dvfws6f+ROSyS7l+zizsslzIQYuTOYJQ2MlbXuLG4nYRy+dxBgIMqypXPfIYQ4NBsFmp8XkxRZHt
    g0NjudYMr4f5fh9ZVaM1m2NL/8DY666oqyWczbJreOSY6qokCGM5sCAICKL4hpd3jbesIdGe8dDs
    iGO8RT6H0ePJIkJeg3ubNnEo7ePRgRmM5BxveI75mMEGIK8bYwAGChpaoohNUagsK+O7u/bRFYky
    raqK1uFhBEHAoyhMcbmQDIPth9uYXVfHXzu6WDl7Jksa6nFYLMybhJTtOyA+g7a0oZ57brqRc+rr
    0fN5NnR2sXdwkFmVlZzVUE88lUJVVbymSb3Py+GhIHc99BMaq6pIplJYbVam+Uu5ormJPbE4vakU
    M8oC1JeV8Y0LL6DG5SSrqqzt7ePvf/eHsRO91mZF1rVj803DZH6glJ2RCLpp4pRlKmxW2hLJM+OJ
    JY1zvP002N86AAZI6m40M8dfh5vZHKvkUNoHJniVXAHEp2tH0S4FQZg4fjlO+vdoS+Tz/PfmzejF
    5XWzKyto7ekhb7WiyBLfWruen726BjB5+K47+MzDvwSXC4fDwfZ7P0VVSck7IP6ftGWNDSxrbBj7
    76dXr+Vrf32Oj644n2X1dTgcDiRJYiQWY0ogwOGBQSorypldU8UlDQ3UB/xUuJysOdzGs69tojrg
    5/fvvYVEXuUL619jSX0dnzzvHP7rt78fNxNs8HhbO9L4HudYyAzBXHZsrCFgtZxRL4wuc7G/mzJL
    9nXnrG+GWUWV4ayNb3acxb5oVbGIBeIkilnHA3C5zUZUVQvetZi2GMVD0q7IzPGUsDkUPnLtx4HZ
    NAz0UYAbBn9Ys47ffPAuVnd08qMNmyhx2Kmpq2UkmeQ76zbQ2NyEIQh079tPeJz80zsg/l8yn91O
    rd9PhceDpbh3WBAESkWpIKqXSHD2kkVs6e5he28/r3ziY7SUBZAdLj6UU/Hb7UwpLSUuivxp9Ro2
    1tSw2Oth7/BIQR2kWAkVTMjkC0vTRElidqmP3UURgd5UegzUfacYdDh9ECvYRQ1FAM08PR6WRSw8
    fnxrSqBQA5KEQiFr9GCwFVNZzSj8P6tUCDnUcVsRR2EjChDOwYGUv/AzSQWhQOnwKLbjUk9PHjsL
    RFT1GKECUYAKtwtTgP3FPNdts1HldtERDqMW6awei4WYqo4Ns3zyA7fRHYuyYyjISDTKVfPm8s13
    Xce7f/YLXt2zh+vmz+Mzy89CvPnGCQ7hncLW/5IVVBELwD16xljTdTLpNIPpNBfc/0MGY3Gevu1W
    zp7SjGaaeLxeLMV+4562Nm7+1W840NfP5665imlOB9v6+vnea5tBFLiisQGbotCbz7O1uxe/1UJo
    VDnCNKl3OelOvUmnuinw5Wlr+fyUzWSNyV4Xgd8NTWMo5+Sf6nZgFQu/GNGsdGY8HEr5WFASZJaz
    MFZ43sb3UG5J84nGrcxwhvlK2zJKlSw3Vh7CKuoYwEDOhYRJW9rLv7WeRzDvOGE7q0RRyOj66RM6
    xplLlrl3Rgv9qsqDBw9PeB2voqBhklQ1Zpf62DsSAtPknOYm3ju1me9s3cHhSAR0nasXzONXt72P
    nmiE4WSKhlIfnaEwn3j8Cab4SvnGysvYNBTktl8+wqIpzTzygduYcRzt9Hc88Ztk4kmqwHJxqKLa
    7ebX77+NLz/7HE93dDKtugqn08kTmzYTSiRYUFNDhdXCypapHBgY5LGt22gLhQveVZZA1wk4HaR1
    nW09vSAUONNT3S560hlymk6N3Ub3OL71mQWxOGktaQEIqzb2J0t5ZriJjrSXS/xddGVLyOkyTwan
    siVWSUesnPqSYVK6wixniNfCtVjlPJcFOvnPtuW8NNTCv7Ss4z07ruFgqpQSOU885wRTYLG/h2DO
    BYJxwnZWvLiv+WQ961Pm3ZrG53ftLX7QR8JnSRSIFp9fFgRimsZUfymHwxHWd3Syvr3jSGtQlnl6
    0xYeam7mUxdfiFQtoefz/HnrNvb09KGJEk3NzRxIpUGWsUgSpmlMyNFVvaAqIhXfg24YxYj+WMfx
    DojfDCcG2BWFaq+H1fsOsLGzi49ccD5bu7q55w9/JBGNFcTaHvg+04PDYJq0hSITbhpRltjZ1z+h
    55szDA6PFq9EgU3hyJsD4CKIJzteaJXhknXvZn+0utirNVm89o4CL9oszCjb5DwoObrTHgDWZF1Q
    lNL5/KHzyOgyyHl+2ju38L0pEFetY73frdGqU44Zjl58v83CSC5/qmgaEWECy81anBCbCF6ROf5S
    /FYLm4MjJDIZbpo7h8d27S6kPeMOdFdRDEAQAKsVRdfpGRggks3hdbs4f9YsPr4yTZ3PhygILG1u
    4tE7b8dns+EWRWLxGHa7g4yq8u1X11Dj9XB2YyNzq6t4YsdOusIRljc2sGLqlHdAfEbSRsMgnsuh
    GwY+u33sxBwLLQ0Dt9XKLWcvRRZEfr5uPes7u5hfUUHS50OxWvnSH/7IT7fvAIvl2H0/JuyJJ45V
    uyx+X1LUfTqeOWX52BWcp5UvSKwo6+B8X++kZGsx4T1VB/iPZGAc0CYCbqyXO1qAKhI/DFMgmh9V
    yTSPfH/MlkRj0keobbzEr1n8x1HXQkRAFgUMvTDmGLBasErSxA2RFFp5MwN+Zpa4ORiJkkhoVHvc
    x/bkDYOLWqawcu4cvr3qZVwWC0vqaxlKpfjEM89R6/Fw/7uu5TvXXk0ylSIejeKxWrl59ixUTSOr
    5hEEkVAoRF8kwpcf/iWUlzOlvIzVH/ogX332BXa1HuKj11z1lgfx20Z4qCcS5QO/fIQV3/0BvaMz
    pONaFaYgUOv18NP3/j3fuukGdo2EWLf/AAF3CV+9/loev/N2OnI5BoaCCOKxwm42RZ4A2qNNM47S
    0TDNsVk6zTAoGbdd3iKKpzckKMB3Z77EJf7JgTinw2eaN8HrFQYYD9g3qlktCAVJ2aL5bRamlriQ
    j7qOummS04uhuWEwz+NhYJyMz6jldZ3fHWjl4QOtKFYbK+bM4dvrXjtCwzTNQs/fMMikUpxltdA3
    EuLgUJDheJyOkRAbtm7nmT176BseYXB4mOFIhEQiQTqTwRBFkCQUxYLVasXv99NYXc0/3ngDHrud
    tp27iGSz/PNFF/DFd/8d182d/U5OfKYsWVyilu4foKu/H0eRhztK/BidP86pKiOhEAtLfTwnivzl
    wAFCqSQ/vfYqPr5oAQ1WK/etehnGrWF1KgoLK8pY3dVz3B4lMJFaaUKl3U6D087OSIysYaCME2I7
    epXLqbzwNRWHqLfFT7n9Ybyzs8mc0X1qb7TwOGqxvEpS0048HGKazC8LMNXj5qW+fpAkhNG2XrGt
    pJkmHeEoAbuVhc0N9JSW0h4pFOdKnU6WlPmJCyJdyRTrIjFuX7qEte0dvP+3T3LTgvl03P8tBASq
    3K4JEVs+nycSiUwYhFAUhYDXy/duuoEvrLyCdD5Pk7+U2bU1E673Wzkrftt44qaAn+c//EGe++Ln
    WNzSQllZGX6/H13T6AyFeeFgK+va27FarcydM4e7L7mYKxcvQjcMBuNxstksduDCaVNpqasbmyEW
    gGWV5Uz3eidMqQuAT7FMINuP/+FgJlPwQMUbIjlOseP0JGsNVoXrGchNfl+QIsD6cMUZ3Pxw5kwb
    73FP8ph/XLaUJ6+5kmumTsFUNeZWlHNWbQ2YhTZWrbeES+fM5iMXX8zLd98Bxe0c4XSa59s6WFRe
    xqHgMJ995llKPR7m1teTTqXYNzREY2kpDaU+LIqCJEljX3a7nfLycsrKysa+HA4Hg4ODxEIh5GyG
    EkMnl0zyledeQLjno1z1wEO81WefzpgnTuRyfOOFl8hqKh9dcf6kpFB008QwDCRRPKYCaBjGBOqj
    RRCY5fVgGgbpWIyUaSKKIhWVlTy0YxdfeOy3TGlu4uC/L8E0TSpL3Pz879/D3s5OAj4fHkyS+Ty/
    37WHwXgMSZYxRrcOxOIMx+MwzptaJInHr7mClwaG+Nqa9aDIx4SRPUflc683tM1kXbikyfVdbSI8
    3DeDO3ZfeaxA3dvBBAHR0Pne+o3EJZHdIyMggM9mQyl6Y1MQ6I0neGzzVh7bsBHJbufW5WeBJLOh
    s5OO4RHW9vTicdhxWiyImQyzXE48JW62dHYhfOhjmD+6n4F4vCDx63YhnyDCslgsVI4T0hvz2tks
    stWK9W2wfPyMvcNULs93X3mVZDrNtTNnUOV0npTVpBkG6zq7eLWtnWtmzmBBTTVisQ8sAOlMBm1c
    CCtJEr7S0mOepy8WY9XBVrDZUPN5+np7yZgmH/3LXym3Wrh6ajM1NiuusjLMbJbOUJhEPEFDZQV9
    0RiaYbA7euya0pyuc/njvwdRRFDk43uWSQJYFgQMzOMPNhgSy8s6sEuTm+dVTTivtJdZrhD7koG3
    Tkx90vj/SKHLJkosLS/np3v3j7WI7A47q7u6C15YkpBFETWXw+10YpgGAgKvdnQiahp1gQDDiSS7
    Dh3mkXv+keWNDZTYbQTcJSRFkQc2bMQiSRzq7WPmN7+NAOy+91O0lJchnQYgbz17GXPq66n2lPzf
    AbHbauWLK69ANQyqS0rI5XInvb2ymsYTW7fzk6f+hPX225gZ8CMKAsFkkvZQmOk1NTQcB7QTCjzZ
    LGv3H+DVjZuh1EdPMsnwyAjtwWFeWLceLFYOj4T4wcrLifT2smUkzKjIbNfwyJFiyYnAWCxWvRGY
    FET5BEzjBM9kiiwsGcKn5FEEyBgnz790E6Y6ktTaEuxLBP7XkjW/xYIoCAzncqdImA3KnA5yqkZc
    08jqGmuGgiDLhQ2YAnzmwgt4YMMmBhMJJEHg+pYpZBCY7/NQYrOzcWiINZ1djCSShDWdlbNnMiUQ
    4KzaWmpcLjTDIJfJcMX0adR7PEiigCBJTPGXMpJK85tNm3nvksXUeD1j0Z0oisfs/Bpv08vLmV5e
    /rbIic+YPI9FljinuYnzpzRT6nZhsVqxnuIrlssRkSRuWrSQuXV1WK1Wnj14iDseexzFNJnlLyWf
    y5HJZMhns+zpH+DnW7by6p491CgyoihS6nRiuFzsGRhADUX4+ntuwu/x0jilGcFmY3lNNWc31NOV
    SnPHE78nmctxbpGGF83mTu49jCOjioVVnvrxdxOfwivr5knUI0WT/Uk/YdVGX9bJEs/ISfnTNhEe
    HWjht4MziGrWM3+qiwKT4fBJgoCOOcaDPu71A+5uaebc8nIG0mmGi7reoeJqHFEUqHa5qLTb2TUU
    JKtpXNEyhX9dvpTlZX4WVlWytLaGpw63sXcoWJD7yWQp93p46N03UenzIRcXB8iKQp3Xy1mNDSyu
    r0cSBab7S9kzMMij23dw9fx5TA34EUURSZIQBIFUKkU6nS7cX/k8iqIgHifs/j+TE5/+zSLynkUL
    uX7uHBxKYRheN02CiQSx9k6CCxfgdruxFSmTkiTR093D155/kQqHg9sWLiCTz/NsWwcHh4b43rv/
    jrb+AX6yeRtXzJ3DbXNmcc2UJpx2O+XuQs/RSKWICwJpXaemtJT2aOyE76/B7aK+ooKu4RG6YzFM
    0+TWlin8prgGdfRGLbPZUMSJbZbT26xoktIt3N+1iEv83dxZt/8UdQRocsQYfD1TRJMIe2vsDvoz
    meMvMhvfLTiJSodTlvFbLPSk0zzVO8Asj5u+8e2k4r8lBGLZHD/dsROt+Hormhr/P3vnHR5Xda39
    3ylzphdp1Ksty7IkF9lyARdcMNX03mIggYSE9HZzv5tKElJuckm5afTQEwIEAphmbIqxKe5yr+pl
    pOl95pTvjxnJkgsY0oDL+oPn0TAqc7zfvdde613vy4Lp0wn6hzB0g66uLoaiMTRNw2W1otglXtq6
    jd0HD6LoOgvvf4gZlZX87yUX0TSKRulxOlnY1IhtzVoIBEhrKorZfMR9eHQNJhKJkM1mcy1Lw8Dl
    cmH5ACiCfKDkeV47cJAfPPs8pU4nd111OalslsvvuIuntrTxmZkzeGT/QQaTSb44fRpfOGk+is2G
    22bjp6+u4bWOLqosFl7r6uLg4BA3tk7HLIr8YvPWv4OkYdDqLcBpknh5YOgQx9puZyCVOrY43lEB
    JHJB2W4ea32a1NvUq0QB4qpMzUs3ENNMfJDDIcs4ZZm+5DFECTMZ7rz8En718qtsHfKDIKDIMlUO
    O7++/FIsJhOn/OwWJlRWcPvFF7C4ufkIDsDfVayNRonH4yOgttvtON+HfscfiD7x1t5ebln1Mq1V
    lZw1uZldAz529vdT4/ZgN1tAkmhsaKA2HEEJR5icb0EFIxH8g4Ns7urmpd17aC0todXjZpzVwt07
    dpFMp2ksLGB2STGSKPDQvoPvDniiwMZAcGw6LQh0xo/kV5vyIgJH7Z8aIjPc/TzW+jRR1YpJSB7j
    zBZQZINpa64kpir/lhaTAFRYrQhAd+LI6vxxq6PkT/OYqo6Yq+n5O7QsiRhGzp0jEIny2fnz2DE4
    yK9efpWMYdAViXD2r38LgsAV8+dxfstUptSOIxwOk8mfpA6HA7PZjJHvYrwXkTyn0zkGtKlUCl/e
    jAByPWaHwzEmBZdGsdd0w0DLi1gcqzqeUTXCqSSiIFBot7+n1P19CWJd18lms7l/AEFgW3cv9zy1
    gg1TJxNLZ2g/2E5LYQFGZTk3zJnF2ZObaC4v4/PzTkSQZRLRKB1+P92BAHokQjo/ieRLp/npogXU
    uV185dV1vNnTw65giF2BYK5HLEvHJHu8XcvkeF5720kfQ8Atp8noHBPAuUWhgJ5lWfEBBHJjgv/q
    6rQB+DNphGMoglqlnLPj24riHX7dMAwKLWYmuN05jyWbjf1DQ/RmMqwd8nP3BfPpDYVY1d5BIJHA
    F4uDoiAJQs6C1uGg2Dm2z55MJgmFQiNps5yvTIuiOCaNjudHKq0m0zsOOlgsljHptaZpxGK5VD/3
    zy5gsVgQBAGTJLGuo5NVe/dR5nbxqXlzyWazY1qnsiSyas9evvDoE5S7Xbzwuc+gSNIHE8TxeHyk
    Z3skHgSaS4r5r49dQUtFBZIocvcbb/Donr1s8Ac4vWkS9WVl1BcW0uvz8Vp3D754grX79rGus5sz
    6utwm0wUOxyUud3YHA6imsZ/LlkIgsD5Dz6Mls3SUlFGVzjCnsGhIxanw6SQ0tR3bdp9/MgQmOoc
    ekc8SkKapAa3TFnDuP0RvrjjlH/LaXwsyxpZEJAF4Zgc8/zKp7KwgJ5RHkwgEMhk+WlzI2v7+tmS
    TBHMZEGW+evLr/K7M0+j1O3m6auv4qlt2/nBy6/SFwhSVuTl0XWvs2fAx1+uXU6tx40lD1ApD95h
    g3o1v6kYuk7bgI8Nvb3YTCae2r4Tiyzzw3OWUeJwvKvnIEkSbrf7qGvZJMDzO3byg788RmFpCRfV
    VpMRc4ST4b9JMAwyWXVkI3kvAP6HVqePFZu6u/nZylWkVJWJJcWo2SzhcJhkMkkymSSVSmEymZBl
    eaRyqCgKVqsVs9mca8Z7PCyun8DWnl52+3zUFxfzhxdWsqGzk79t30kgGmFuVSUmXeNLf3uKe7e0
    sbmvn1AwiMlq5dunn0ZjcRG7fIOkdJ3JdXX8cdNmHtm+k2umTOZzC+ZyTcs09vsDbBnWd9IPSaOW
    WS2kde0diz3vHcQiCwq7OK2o8x3leYT8zvv4wATWBKv+7awtkyjiNplIaRqyKFJsMRM63Lt5uMqv
    60wtKaG1qpLt/QOjriBgqCoFFguPHuzg4OBgjo9us7Jo4gRaysvQs1mqyspoqCgnkckysaKccydN
    ojseZ3tXD9VuF+VOB9c8+GfuefMtSlxOmsvLyagq6XQaTdNyJ6au8+D69Xz+4UfpiUR4esNGNvb1
    8+kF8yi0/f3FQkXJcbJFWcZusVBTXsaZTY1Um80YgjBSBTcMA7Mk0lBaykkNE9k/NMQb7R00l5Xi
    fJfFtH/6SdzW188vnlrB4EnzaSkswOtyHbF7ScexA/miMe58/Q1Wb97CA9ddy8Of/TT/8cSTtPcN
    kMiqRMIR0jYbmmImq2lcMWsmS8fVoCXi1BUWkJYk1jz2BDsGh3i5o5Nt/QNEE0mKRIHmAje/2drG
    hu7uEQvOlpIivIqJVd299CSSR7/j/aPUPSSVX3bM5LM1m6nKC8G/XWErlJW5rWvau5g0+ueFqutE
    sllEQWBBUSEJVaMjb0s6fDpPKfayIxjO3WdjUfbvihz12mJSFCKp1Ej/PpZKs7a7h/Pu/xOPL7+S
    4liUvkiU1fsPsLG7h5Wfv5FNPh87duxCtlqQzBae3bMXLZvl8jmzEWQZqyRhHlWVliSJmtJSakuK
    mVpZQWtNNRbZhNtq/Yc/m1k11cyoqkRAyM1H59P7dDpNWlX52our+XPbNv7fkkWs3X+APQM+Lmmd
    TsVh+Pi3gTiZTBKORLBoOq2NjbTU1lBWWorlPdLYFFmipbICfzzOvEmTGOct5EfPr0Q2K8yvLKe5
    uorlf3mU9Z1dOBSFQgEig4NcOHsWFkVh+44dOe5tPM6efI8TSeRgNMaarh5uWftGznEgD8pt/gAn
    lhbn2zpHosptMrGg2MuaQX9ONubvrQNkzdRYY+84xaQb4FFU5hb08vTAhH/7SWzkn48kCNhsNgKp
    9JjNzWoy8eP5c/nSqpfZHY0ROtbcsSiyor2DMpeT/vwMt2EYxDJZ4tks7YkkM8xmzJYs0VSKTCpF
    OBTi/muWc/81y4GcF/YPz16Gpuu0VleOXMcOPySumDWTK2bN/Kc/G1EQEEf97oKCAlKpFJlMhvKy
    MtKCQNQfwGSxMK+ujlKXC4f53ff+/2EtJk3TMPLeR6lUCofD8baMmHe9KWSzfPmxx7l15arcyjF0
    vIUFLK0bz+WtMzh9Yj2fevxvPLBhU77sl+WClimc09zErq5uPr1gHma7nQvuexBV0yh32PEnU2Q0
    jd2+wdy9RBhrCpZVNQRJzBEbDp9M0nVmeAvZvEJP8AAAIABJREFUH4sd1a40321CFAScskw4k8Es
    S2SONRygySTO/PlxVSdFAfpSds7acCHbo8XvixN5+JkMV4KHNz4RqLTbCBsGkTzXvMBk4rLW6Ty4
    eStZXSelZjGyKl9ZuICEqvKHNWuZWFLMp5ob2RoIcd/OXZQ5ndxzyYVc+edHmFlZwdeWLGJm3Xgy
    8ThlJR88mZ3+gQHWdnbxZl8/CyfUsWxy87+uxRRNpxEFAavJhGAYpPOX8kQiMVLad7/LdOC4TmJJ
    otztorTIi8NiAQTOa5nK/5x/Lsl4nHA4xNkNE0nn741DkSgHIjE+t+I5EgMDfOvC8/nLxs18acki
    hhJJOnt6+MW6N9BUFQRxDC3HJIrUezzsHBqizu1mcnERL7Z35Ab/R8mqbnobpQ9FFElrOrph5E5q
    QSB9LA9jQ6S1oAd/xkqxknznerMBxebEiCXp+yGsskxa0yg1m4mpKtF8IUkHuobT6/yzCqZS/PLs
    ZXx8ZisP79jJ715bRzKrIkgSocFBMGBaTQ0fW7yIh99az8bBQepLS+gLBvH39mGurSGjqtz81Apq
    SktZFIuxJxxhZnUVdV7vB8Jzqay0lHlmMz976RV+eu8DrP/FfzOzuvpfA+Ll9z5ApdvF9XNmMam4
    eOSuWPgOPOe/J7b29BJNpzl1wgSW1deT1TTSmopJ0xn0+ZBkGZvDyXkt05jb2Eg0m6G9t4//euoZ
    En4/S2ZMRxBFrvvD7WC35RaT2YzLakUym9ENHUmSCUSiIObaA/35Xq8gQFLTUA0DsyiimBWiw46J
    b7NY0tq7IXoI1NnCeI8DwEIeGF/btZiDCfd7SqdzrRTjH+o0YZJEktns0Ykbh38tinztyafxOp2E
    44lc1V9VMdvtLJkxnT9t2sKOvn5ufvU1llZX8YuliykqKWH1wQ4WtEzl7IZ6Xtqzl1ueW4mnyMun
    TpzDf9/3ID+68Qa+ccpSPii2aZJsorKwEMpKscjvnbjzrkH8xIursVdWcP6UycywWt99X/U9xLee
    WsHugQEWT6jDZBjMrq3m0qlTsVqtqIKAPKrv981H7+Ol/ftzo4tlZcwbX8snZ88ilUwys7mRRDaL
    L53GHwzzydmtnN7USH8wlDOxNpn4your6Q1HSORJCPsCQfblBfXmVlZgVhReau84os/5dxW4RI3H
    B+rZFi1mumvwbbnTsgAbIsXc3T2FlP7elAGskoiqG6SN95iGH0WCJ5JIIsrS8W0MgsBv1r4Ooogs
    y6iZDGe0TOOKhnpkp5O5zU2sO3CQnfv2M/PGGzixvp71u3fzw9UvU1fgobmyErMsc2ZrC16LhQWT
    Gnhj0UnUuZwEAn6KvN4PBIg9Vgs/Pvcsvr50CXVF3n8diO//2pdwW63MqK35lwAYw2B7by8H/AE0
    A/Z39/D5009hxd7HsZgUPrtwAXNqa3L3jGiUlXv2MhAOg6bRXFTEl0+aj80w8NhsPPH5G1nwi1+T
    VlVqy0u5o20H63v6+OXSxdSOH0dBSSnV1dWc+ofbSQzrKg+P0JnNOB12Vh9oP+JzexRlRJnxeD/T
    GCcEQM3YCKvmdzxFJBHaosWkNPE9F7US6t83h+wym3AqCj2R6Eg1/4wJ4xmMxtjgGzw0Hfa2K08e
    qW4jybx68CCn3HUPBgLRdK44VlBeRkNpCaVlZUwwIHjXfWyIRvnfNa/xm0su5oxpUwmHQhiaxqPX
    XYuaTqMoCqlIBIvLyft9dMEkSUwsLobiv+/nvGsQX/VPrurpeWUHQRCIRCKomsb+m74DwONb2/jh
    cy8wvbKSG159DZMkcVJ1JZMcdlRdZ8+Aj8FYDEUx0VI5jrn1EzgQT/KfTz3N1u4eyp0O4ql0ju4X
    CObc9GSZtMXCxJ//ikgyxZ2XnI9JPDIdTGUydPkGj2Re6TqtJUWs6u49rtPYJAosLSnl2b7+kfcb
    molvNr5Ey3HYmqY0+HjVDmrtZZy6rgXE3DeYJZHssPXn8Ryk7zEEYJLTgcOkYDIMfKk0CU1jdVcP
    fzh1Cb96cwObh/zvboMXIJ7JEs9kc/K3qsry+XO59fJLseaJEVXeQj52ysnc/+oa+uJJ+kNBSm02
    GurqiMfj9Pf3YzabKXY5kb7wVQpcTu667BLOn9nKhz3eV7TLtKry8QceoqG4mEuam6gvL8M1quR+
    /rSpnD9tKgCvt3cgCDB9wgTc+fSp3mzmmhNm4zSbmewt5IZ7HsgpcsgmRJNMMJ0mlcnkTw+dU5ua
    +PKSRdRWVuKwmPFHo3x9xfPcevqpfPnF1fTFYod2c11n4fha5L4B9gVDJDMZREHAEEVWdfWMLFqL
    SUbXDTLHsDvNagbr8q4Sw0WtRtcAZxYfxCVljutOrBnw8tCwD5Tx7u/gx3tSiAIeRWFw1ASSSRR5
    ayhwaMg2/3pa0/j4U8/mTuFRAC4ym8nq+hFtOFEQcFutBId55obB7LJS7DYbL+3Zi5xOoyYSaHY7
    gijiNJu5+Zxl9EcirHzzLbadNI+p48YxODiIy+Wirq6OaDRKZ2cXNrsdRTYxkEjQ2dtLmdeLJMvH
    xUf4IMb7SmMro2k89NQKbn9tHapJHtOkH443Ozp5aMNGvn3madx6+aVMLy0eYX8Vmc3cddXl/Ori
    C6jxeimrrGBWw0ROndLMZTOm871TTmZmdRUYBuVuDzfOm8upE+rY2dmJy2bDabfTWlHGSU2N3H3O
    MhiddqoqV8+ZTVt3D8n8gqxyODAPL9oRn1zt7emZAoRHW6BoMjfUbGG+Z+C4kj+TCK8EKnnC1/AP
    6RHb5GMv7KxujAEwQNkwm0g8SmFPlo94bSiVOmofvdTl5BMnzoHhvnI2yw/OXsb1s2ZS5fXSFgqz
    ub2dSCRCIpEglUpRYDZzUv0EpjU345ZlstksxcXFZLNZYrEYFouFcfX13Hf1VYzzuPn0727jkZ27
    UTMZ4vF4jkP/IYx/Ou3yXaVqgoDkdHLB9GnMzltSHh6f/vPD/PiBP5EURU6dUEdGzfWnjfzESCaT
    JZtVqS0p4dTGSQQTSSRJ5NOzZnL2pAYskszTe/dR5nTylXknAPDCgYPc+9wLlHkLuX7uCbSHIvxm
    /UY6AoFDp4oosmXInzud85Vpp6IQV1U0XeeShnrMkkRvLIbxTmn1Yf/fJmfoTduotZtwiJF3BPGz
    g+O4u3vq24P4eIpthkGN3UYwnTn2ew97PZzvp1skEVNemXI4FElEFMUcD14Qjl6ZzofFJOOLxphc
    U40/kSATCrOwbhxntk7n9Y5OXt6wiVkN9TQVF+FyuXIDMbrO5JJirpzVypTqarKZDNlsFofDgSAI
    pNNpUuk0M2prOBAIsS0U4qJpU5hWUUFGVclkMiN8BlmWPzQgfl99EkWS+N6yM972Pac1NiIIAvMn
    TMCkKMcUMts7OMjNz73Aszt3kRoa4uqZM8Fm48WODvR0GodiwlNQgNNiYZzLxdkL5qMZOr9YsxZf
    NEY4lRxboBFFNrR3jIDaZZLxJ5Ok86dMMp0hOoowIgDGcEp9tDRuRBFE5fGBerpSTtJGN98Y3z3G
    HO2IbEWHS8p2cX9vI+tCVUe/4RoGxVbLEafoEeDOm8O97XsOiwKzmWAmg6objNYrEYAqq40aj4s3
    fUOHCoPHiKFoDJMg8svzzsGpKPTFYjS43fz8+ZWs3n8ALGZ6wxGCkQiCKGIymXA6nVgtFlRVJRGJ
    oOk6LpeLeDyO3W5HURQymQz+QAC7JNJUVoYuihiyjE2SyGQyJBKJ3BRTvqX1UTr9b4hPzZ/LQ9de
    zaWzWhHfZjcNxBO8sGs3mq5z9WmnUFdSBKJIIJmCUJjvnnUmdpOJUDDI6VOn8NAnruH8qVMZjMYI
    xWLMrq0FTcNmMmE3mcAwKHW7mOwtBMMgklUxmWROLiuh0u1iRVc3+/OufuQB/KXJjZw2pTkn63NY
    lFgsTHI5Ge90IIkS68NlrBwqPp7Dk2KLgN3kPLqcrmFwVkX5UQEsCQJlVuuY70sd/rcZBrVvs7iH
    U2PVMLBJEqb8pmYAA6kU/kQK6Witq9F/a37GNpxMcvqtd9A+NMTC6iq8dhsRIBkKgW6wv7MTp9NJ
    MBDA7XSiJhKIuo7FaqWgsJCioiJikQiBQIBgXpdaURScLhf7AgHW7djBlp7e3By2omC32yksLMQs
    SfzipZep+fZNFP3nt9jt87G9v5/YBzTd/sDlFGZZ5njYpdUFBdx01pmYJZnr552Akgf8U9d/nPTH
    riCdzZJMpyksKiIRj/Pq9u3c8uIqgtEozVVVLKkbz67ePrJGbsEhCPhicQY4NPAfiUQ5b8kivvrc
    yiOnjwRYNns2yR07jwq2YosZVdcpt1goVsDCXlpcg2MOVgGI60XYxKFDRS3c3N09lVf8DhCOYmYm
    COyNxY76TDTDoD+Vevs0W9e5asI4ftS245jdg+GIHjYzHFdV2gKBo2cGDjtZTSelaZwzroa/7NmX
    68WrKve3bWdKcTGFDjsX1dZw0seuoMDjYbzXSywcZkswxLdvv5M/b9iEx+3m2euu4eaXXiWUSnHW
    lMl845STGRwcpL29nYqqKgygqayME5oaqS8uGpV4CMiyjOx0YrLaGIrHSakqjTfdTHmBhz99/BoW
    Tqj7CMTvl6hwu/jqyYuBnKztd59+Fpti4oJJDUwvLyWgGUy66XsQT5BnemCxWpEUhR29vezo7kYx
    m6lwu+mPREBVOWvqZJ7ZtefQQITJxI9eehXtsEKWSZIodjr42uqX2XqwHcFkOiLp3T6i7xVnVkGa
    26c9T5kSH2NragB2cWjkew3AKoaxi71YxalIgo3WAjev+YZG0nyvorBnVEbwduE0yQgCxLJajt+d
    p4e+1n9IvUIZdgk0DGRBOKbyiSAIlNhtBJMpMnmiDIBVlkhmVWQpRwRJhcLMaZhI3DBYsXtvbnMM
    BEhpKuFwmK29vdjdbuqtVmrKSunt7uFP6zfw6NZtYDZT5HRSU1HJxp4eBiJRGktLc5ticTERw2DK
    936A1eHgv889m+tOnMMT27bzZkcn8+vGjxn6n1xexvI5s9F0nRXbd1Dl9mAShA/kWv/Qgnh0+KIx
    7nxhJdjszCgpYe64WrRwmJnVVcTSGXTDwCZLLKgbz6Nb2+gPR0CSKHY6+fLihTz41np6wmGubplK
    W1cPHZFDo3Qm6ZAFic2skMio1HrcSILA1o4uGA3gw+6jw9DcGJaZt+4SPlnr58f1j40Bsm4I6ChI
    QhoR8GdN+LM2SpUY9e5aSswK6/wBzJJMSlUP+Sm/UxgGipgb7ojo6qEKsijy8oAP8oL+5flqdCCT
    pUAx0ZkY689cqJgIZLIj/tEXNjfxwoEDoKr4kymSaQ1EiUAiwZyyMkxlpezr7WdxbQ0rdu0BWeb1
    jk5e3XeAy2e18tst2+gIBllUNx6LYqLEbufRtu2IZjPnTG7mkwvmUeyw8x+nLCWRyTCtsuLQRxJF
    9oYj4BviwMAA+/r7+dzvb2fZ4oU8dO1yXKOuCWc0NXJGUyOQ8/mSBAHFyInlmfLOEcpRCqsfgfjf
    FDWFBfz0ystR0ymaS4rQdZ2mqioevPZqwokEsXgMt83GG53drOvswqGYSWka3UNDfPEvj+Kx2Vg8
    rpY/vrWRoWQSm8mEnieAuG02Si0WKu02Tmhs5EerX6YjGMoBe7SjhCxTbLUwEI+PKDseymAFMlIV
    SwrXHjGKqCrV7InOYpr1MdKGA7/q4ZftM9kTK2BuiUJ3MoVJyP2eiR43gqaxO38Sm/IeR+pRrWgE
    /KkMR3MwHGZh6YZBXyqFltcGi2aPHPafXOBhoz+IqutkNY1AIoFJkmgtL8NIZ5AdDp7fsxdFECmx
    WXHa7dz6xpu5DdBiyZFnNB23x4PL6cShKLjMZnb5fPR3dTNn+nQKXS7KPW6+d9opTB8/DoAvLDpp
    TMYiAEV2O7+++EI0XeeUKZM56PczbWYrU0tL6O3pIeXxYLfbMZvNY6rT1QWeUR9JJ5lMomkaqVSK
    QDLJy+0dmGSZC6ZNHSGffATif2EYhoGUTvOFeSdisdnQslke27KVVzo6WTqhjnOaGhHkSkRd56GN
    m9k14EMkp8XUVFHOzv4BYtkMe4NBtg/4MEkS3zlhFi9397CyvZM9/gBziov4xtKTsTnsPLZtBxt6
    eg5Dokqdt5CsrqPpRxfK07QsU+z7xjC2BCCeteEXZiEYjyGSocY8yHVVbXxrzwL+2tNLXNVy6b2m
    MZBIIg6D0sjpXemGcWy9KyH3HyGfUWRGEUYqbFakw61qDueLA+v9QRaWlqBms7w44OOFAwcwsiqB
    wkJ+dMoSnti5m3Qmw/jyMt7yDdE5sIPZdePYPOAjG4uB2YzX42JqWSm6YaBiEE4kOGtGCy1LFlPj
    dNBcU008EmVSRTmJRAKLxTJGnG74r/JYrXx+FLgr3C6evOE67IqCrGkMDQ5iGAbJZBJRFDGbzUeM
    yw6LyhuGgZrNMjDk5+uPP4nDbOaUhoaPQPyvjmEdYY/Hc4ito+s8t7WNO9/awNaeXtJZlWUNE+hN
    pXn5YDuJeIIZdePY1t3D3qEMaBpLGyexfSAnJ1PhdOC02zGRG0fMqiqv9Q/QH4kQ9fnY0NmZl/fR
    QZKwyhLXTp3M7kiEVe2dR5GtkRnvCHJ1xVoKTGMZWwZgM7polu9FA0QyWCQoMiXREcbOMQvCGBKJ
    RZZQRIGh9Dtzug0gexjja2BYWP9oqblhMM7pQETgQCBAVzSKxWwGXc85XUgSr+87gG/2TOwuF6gq
    +4b8aLqO3eXk95dfilsUeLOrm6seeAi7xYrXZmPQ52N563SS06czr7SYufUTsFitZFIpMoqC1Wwe
    IW1Yj0OJw64o2EelxHabDb/fjyzLI+n/4OAgAG63e0z6LAgCJkXBZbPRXF6K1aQgS+/PZo74YQFs
    OJXkJy+8yH89+TTtAz4Gfb6RtGloaIiBgQEGBgaIxmPMrBtPZXExr+zZy+1vvoXTUwCimCtQZTM8
    duVlTKsoxwCWNjeyuGEi2nCV1WbjS8+t5JmD7ZzX3JgTVzMMQpEI0VgMVJUKt4vbli7K0TIRiMoy
    g5nsmCq1y2Jh2ZQpkNWZ4fTxxXEbEPIQNkZxtxQhTpG0awTcqg6zPP2cXtwORu59wx5Wo0+mlKYx
    lM4c9/MbvXmIgnBsed18uF1uPjVjGo9fuxxdFNk4OMSC6kpuXXY6kwo8IIlc/udH+PELK2keV4uq
    aRiaTq3bxWAySUcyhShApbeI3lCIc+6+h93pDNU2G2lVRbLZScRimPLGdybFTDgSwW63j6S77/rE
    kmWKinLXKZPJRDwex2azUVycY/35BgY40NfP+bfdydxbfsXeoUGaykp56Quf45nPfOofosH1z4j3
    FWPr74n+cIQr/3gfL25t47T6OupKSognErhcLlwuFw6HA4fDgWoYzG9oQDQMuiMRplVUcM7UyWzo
    7CaYSIAkctvrb9Abi/P9M07l7jfX8/zWNmKZ3Ik2uaSYUCZLStdJahr+RBKPotBQUsyeYIg2f4Bw
    Ks2T+w9iiCKqrrO1r59EJoNmHAKLYBiQSTGUylJpjTLF6SeYtbA9VsrmaC111kGsopHTHxAgpJrZ
    G8/NbE+wRgmoJTw7WEmVw8plTY3UFHk5GAxiEgTKrdYRZtU7Np0Pf4+mce7EekqtVrrjcYxj3KcH
    wmFWtreztruH3vw9P5nJ0uR0sisWZzAeB1HEZTFz0oQ6xnm9JDWVz82awfJ77ufet9azuKmJGeWl
    bOjuod03yP2vvMaE2hp+/udHKCsp5uTx4xiIRlEcDq69537aBgaYVFZKeWEhwWAQRVGOarvytqeW
    KGK1WkkkEiOa0uFwGIfDgcvtJphM8NMXV7N7wMf1806k5H0oFv+hS6cNwyCbySCoKstnzySUStMy
    aVIOvBx+NVUJh0JomsaXli7hS0uXAPD09h2cfdPNLJk/l4umTeWbTz8DksS8mhoml5aSVFVCySRp
    Tec7p51Cu8/Hj9e+zo68SbYkinz/tXWQzR7iD49aXKIokNW0HMMpD5yUprEzEEaQBFYO1bGytx5k
    BYwsdnOMy8uLuKhsLyIGOvDEwERu3b2QBZXbWeLt4o/dExEEg0lOJ80OO7ft3ktW1ZhXUcZ+f/Ad
    n5sAzK4oZ+PAwKFCW77KbjGZKDR0vDYbvmgUWZJGNJ6N0cUvRDqDoZHPGslmuXfXbvpT6dxVQdM5
    fXIz/7XoJKaMHz+SibzSP4iYL8TNaplGTzDElr5+zLLMkon17D95EbNraykoLua6399GShB4Zv1G
    PMVFXDi9hXGFhXi9Xvx+P263+z1RKD0eD5FIJOe2WVBAIpEgm8mgSBLz68YzFI+T1T4Y1rEfaBDH
    43E0TUMURWqKi/ntZZcc872ZTIZYLJbTCnYdgrdmGDy1bQfY7ThMJuZNqONXF1/I1t5ervnTw1R6
    PHx22hSe27ufkKpysH+AxU2TuHPrNnbk5XoSmp4Tnh9eTLrOrKpK1vf0gqahyzJpVUMxmXBZTPgT
    idwJJ4r5+fossydUc9A3xFA6Q1w1c2dHK/f1NmKXNEAgrUtgC7HGX8OaoXEgqiAYvDXkpy0Uwpcf
    JNgdipDVtaOewnZZJq6qYBi4TSbmlZWwY2iImJ57rcrtQlU1Nnd1sTt8qI1W4XZTIEtsG/KPTbH1
    3NAHWu73pYEDkShoGovrxjG9tJR5DRPZ5xsklEwxpbICj92Ox2pFFHLfL2TTfPvUpWxu7+COTZvZ
    1NXNo9d/YuRC8cy+/aT8fk6f2crF06dROUok3ltYSCAUJiUIWM0KbovlHQXgR4fL5SKVSpFMJrFa
    LMgmEylVY1e/j+5wiA1d3cyoqvoIxP+MyGQyRKNRbDbbGIHwY0UwGMRkMmGz2Y5Iv4R8VdNbUsTZ
    06awuHESM4q8PG1WuPOZ5zjg9WKRJLb1DxBKpThlYj23rt9EIBLhvy84j65IhO0DA6zaeyCX9448
    WIN6byENJSWs2LgJTAoZQcBls/Kt00/lqbY2VvbkNa6zKj+YewKfXfEcQ+kUCGBVdEotHjriyTxH
    OQ8eUSNncJIv3qkqkWyWIpuNSCaDb3hA4yjhNSsjOmHlBR7u37WHhKohAJIoEkymMAyDiGGM4Xv7
    YjHSZvOhqwAGhqoxraSYWqcTi81KZzDEG719WPKkmSvnzGaOy8kBVeOKhx7mginN/PjM0/GYzdz9
    6hqQZU5uaKCwuIRYJAKiwMPrN/LYth2cNbGe6XXjc6f2lz5PRlWZ4C2kyGymy+fjiU2biaoaZzQ3
    ElazzLvl1yyb0sxPzzvnXQvAWywWNE2ja8jPj19cxZtd3QzGYtgUBbMkf3QS/zNS51AoNJICvdN9
    KJ1OEwqF8Hq9SJI00gPksCLOF+efyJkNEzBZbWzfs4eq0lIuPmEOk2pqOPU3v2dvIEilx01PR5h9
    /gC3v7YWDINNPh8TXDnR8lw5SkASRVTg9e5erprcxLdOO4Wz6+toG/Jzx/oNuBwO2rq7ebFv4FDK
    bVa45ImnMElSTiUknSGJRlcigXE8I/yGcYj2+TYnUXfikP6VpqpEM1l0w8iJvjsd9I1xZTj0s1OJ
    JE6LOXdLQOCMiRPZFwyxfHIjmwZ8dMbiOSlaVWNicRE3Lz4JKZXEUlbKw8+vJJVI5DjrgkhKzfLF
    pUsQdJ0T8hazDpeLqfX1fPyE2RQ4HDhMMkODgwiiyOyaseJxZqeTm//8CG91dLLua1+mwGbF197B
    7uKi95z+SpKEpChs6x9g885d/O76T3DVrNYRR86PQPx3xlf/+gTrDrZzTuMkPj/vBBwOxzuevMPS
    uaqqUlJSMkb9UBAEQqkUW7p7iGcyzKsoJ6Mb/ODFl1i57g1Wf/v/MdmdM6O26xrL6idwMBbn/ksv
    ZNotv+b2NWtHZmkf2rBp5I4rShKXzZzBpJJivvf4kyhWK0umTSPm93NqUyP7N2zC0A0CwRCCtzA3
    4ZQ/7QQBopkMHqsVuywzrdDDjkiMoaMYlg1nD8PQdikmFpWWsHHIT2AU3fGwOlTObnmUl/KeYOhQ
    rUDX6QtHjgrg1tISvtA6nY2RKLe+uR7dMPjestNRIhGqSkq49657aBvwIUsSPznjFK6dNRN/NEoo
    FELNZtkZTyCazSye1EBVcRHRWIwfnHYKK/buo/4r/8EFC+dzz/KrKHO5uOOa5fT19SHJMqIk4XK5
    GBwcxNB1CgoLMeW9iIW8m4RuGNR5vbzyw+/htdsofpen8OiwKibm1dWBIFBf7MVpsXxgBPfe9yDe
    4xvkrY5O5tWNx+EpeMf3Dw+Im81mPB7PETuuRZZ5q7ePc2+9g0hvH8Z9d/GndW+w8s314LBjVhSC
    iSSDQ0Mk0mkGVZW1b60ndvH5eO12LCYTWV0jHE9Q7HRSaLfTGQziMCtc3NzExvYO0A1mVVZwxrhq
    IpEo6zo6+Z8nngSHg85olDK7bcy9sspbSG84QiiRICpJiAmZWD7ttSv5FHh0j7aggI5wOOepm86y
    orsX8RgGZ2gaU2uq2drTe7zpzsjPsSsKhihy7RNPgSAiW8xIosh5d/6RxuJivtLaQq3TwbbB3IBG
    OJ4gkExis9kIxmJsbW9HkSQml5ZAJkNG03A67KRTaTKpFEphASUOJ1L+94mShMliweFyEQ4Gc1Vj
    t5uBWJz+rm7KnA5KCgv58dnL6BgaIpZO5ySa/gFDC0V2Oz87/5wPZG3ofd9isphkZtXUsLh+AuO8
    R8riBhIJVu/ZRzSVwi6KZNLpIxr3AP2RKA++tYHecBinohBIJPF4C/nEvBPpCAR4va+PK+bMYe64
    Wr7y1yd4bMtWzFYryXQGu8vJRc2NnDF1CqfXT2BKWSnPb9rMxxfM49HrriEYjdEfiRKPRdnSP0CH
    38/3zjoDwe/HW1hIQpR4YPNWit1uEqpKocvNLt8gCAJmxcQFk5voC4WIZtXcnTSVQhZFagoLuLJp
    Eq9394wxNq9xu/ElEhQ57CTyFVWLyUSwvIJ+AAAgAElEQVQ6k2UMjTI/PXTjgvm83tGZG0w4Spgk
    CZMkoakqE4q8iAgkVZWsrtMfiYIkMX9CHSk1i9lkYiAQRDEr1Hi97B0coisaxTAMAskEvmSSxXXj
    eaW9k2+vfRNZgO+dupTb1qzl9c5uguk0m3t6WThhApFUitaaasZ5vQgiKJLM8/v288sXXsTscFJt
    t7GlvYMzbruTlfsPcPkJc3BZLJx32x3cvuolNvQPcNaUZjz/BAuWD1K8r09iA7iwZdqYrw8/a9p6
    +zjr17/hwlkz+d3FF1Ja4Dnqz9ra28sNv/k99ZMm8ujyq7j1iktH7pENbjc3LJhPc2kpsUSCv67f
    AOkMO4f8lNhtXNYylT9u3kpS05hSVES7PwAmhb9u285QLM71J87h1y+8yM7BQayKAiaF5zZtJlJV
    SdPkyZiTuWKVJArIksjjW7flqtl5UUCzotBQWEhvPDGS7iqCwMIJE2gs9FDjdDKQTOYmiESRrf39
    oOu0eDysjERZWFvNVY2T2JdM8vqBdlaOktS1mxQe39rG+MICtnb3jBWwy6fXUyvKMcsy6/bspdBq
    xTDAP1w7EEXQNKwmE4FkiqyqUuBycdn0FnqiUZ7fu4+m6mrCkTDb+gbQDHBZrDyzfQc9g4P0IHDd
    o48TjsXY0NPH/WvWQjZD4rbfcd+GTazat59HNm+hqayU31xyEX/d0sZ9Tz/DUCrNOdddiyMWJ6Nm
    D0keCQKfW7yIH2Uy7Oof4OcvruY3l1z0EYjfryG8w9eQUwNxuVwUOR3YbMfekcsdDs6YdyItVZUU
    Oh2o8TjeoiJi0SiCLHPzcy/QUFLMvcuv4tkbrkcFrnzoYXoPHKTM4+H1g+2EIlFcTgepbBYUE73+
    AH/u6OTb5yzjpIaJvLr/AEkAWeLJA+282N3LGt8QvnAYdOgPhnNDEcO6Vvkhigc2bsY8/OE0HTCI
    qiqPb9vOs4JArdtFOptlYHiCyDBoGT8eLZMEXaextJSuZIpHt+0knBzLdW4Ph2nv7eMTSxayZ3Ao
    JxSYvyRXOZ2YRIEGbyFP7tgFsszG3r7hpnruzi4IIEms3LUbp91OVtMod7v4/ikn88mHHwHDYLLH
    DaXFvNXRSWttLc/t2cfmzi5+evEFpLMq33lqBZjNIBoUelycPakBq9lMtqeX9kiU9o5O9lRX878X
    X8gnTpzD1IpyWiorkEWRKTXVrLjuWgo8HkrzxItPzZ/Llp4ent+1+107CH4Y4x/mxfTvCk3XSWSz
    mEQJs0k+KtCHhfQMSUKRZRw2G4Ig0NnZiSAI+DSNWd++CUdpCU9e/TEWT5tKMBhEFQQsVitdwRC3
    vbaW376yJnciGAZVLhfzqyqwWG18/cTZxDNZfr5xE49t2IQ2DKJ88cptsTDe48YkCGzxDR5SwlRV
    rDYrdQUF7BkcQtV1Tq4bT71J5tZde3IA0jTm1Y1nfkkR7aEIdruNVCJBezzBmz296EBDUSFJVaMr
    GDryXpzN8t2TF7F20M8LO3eNqnLpXNNQjyHLiBYLf3xz/UihTTQpLJ/azLPtHQyEwnhtNtx2GwcD
    QQxNo8zj4QenLuWOtet4o7uHEpeT5668jNc6Orn5zfUEYnHK3C5mFBQwrW48P3lxFZlUngIqiZgV
    heoCD3u/80229fWx1zeEx2ZlycR69LxemjCaSmoYDA0NUVR8SKA5papomo4siceUaPoIxB+SGO4R
    C4IwQrnL4UclFovh8XgIh8OE/H48Hg/uwkJ8Ph8FBQVjquD7B4eY8uP/JpXNIosiF01v4Z6rryId
    jZLKZBCAPUN+rvjTX+gKBkFV+e2Vl3PXK6+yoaeXCpeTeFYdGVJwKgp/vPQiBKuVj91z/yFNqlEV
    ZEmAxdVVLGpqxJPNMrmyEpehk8lk+I+1b/BaVzcuRcEpS/QkjrRPGdZwbvv6V/ji355i1d59lLsc
    eG0Odvv9ZPN3ZIfFzK/PPotKj5tkOk1ZYSH2VJKp//t7UBQcZoVxXi+VHg8VTgd3r309dw0pK2Mw
    HicYi9H2mU/yuSdX8HJnF9VeL3cuv5Kzb70DQ9f51VlnMr6okNKSEh7fso2fvLCSQrudvptvOv7N
    WtMIh8P/VLugj9Lp91kMg9TpdKLmlQ7t+XQwlEwSCYepzhPfNU2jetw4otEo7e3tlJaWYgDb+/rJ
    ahq1bhcDfX20VJRj6DpPfuo6SvKV70SeqmgYBoV2G9fNnMG+IT9mq5XZJcXMP/8cpv/hDiTFjEM2
    kdR1MqpKlceNoWl0BYOUOp10BAK5FlBeSdIqywQTScIGfOdvK8Bi5sYZLSyrLOealavxxxPIssx1
    c2axZscOemI5rrJFUTAMg3Q2i9fp4G83fBJZFDiruZFVu3bjtNlZ3NhA55vrR/qq1S4XZ7VOJxYK
    0eHzYRUFCgoLKS7wIEky/UNDTG9p4b6rryIRi/HMzt3EMxkmV1ZQZreT1nU8BQWUOuwUeTxcWl+H
    nEpRU+AhkUoz3ltIndNJQ2kJJQvm0tbdTX80ym2vrGFhYwONx+FqKEnSCOfZ9j4dRPioOv0PjHg8
    PiJlOjw/OuzUuNs3yOf+9BfWdnZx+sQJOCwWrPnJGEEQKCkpIZvNEorFmPGzW/j1s88zqaSYc0+Y
    w/Xz5rJ8ymTceQCn4nGsNhv+eJyn27axYcjPirZtTK+u4tMnzSOZTpMKRygvKebpjZuYUVuLUzEx
    4PfTXFpCV8DPb9a9yXdPXcq6jk5SeTOySquVWaUl1BZ4OG/GDLb29hFPZ/CYFYptNl7t6SOhqkzy
    FnLZ5GbKzQoxSWJWSTFXzp7F9JJi+pIpwskUTrOZ7z+/kpf2H8AQRQwDeoIhekOHUu9rmhpY0NjE
    zS+u5jsrV7E7FKHI0JlSUU6hw8HG3j4MQeSq2TNx2GxcO3sWEz1uLmudznUL5nP2xAmkNY2pJcWc
    WFqC7HTy2MZNLJjUgM1k4um9+9ne04Mim5g5fhx/WLuONdu289T2HTRVVo7Y8LzjiSPLZDKZo3oO
    f3QSf8DDF42xo78fRRRp9haiKAoOh4NsNks6nR7TL+72+3l6y1a8Hg9fffpZTJLErKoKPp53lgdy
    dE6LhXOnTsEfjzN1fI4CGIlEkE0mVu/dx+1r13FuUxOXz5lFWNe58eln0QwdQmH82Sx/27ETHfjZ
    KSdzWuMkfvTCi3SFw/jCYW69ZjlGPM7ze/eRymR5cOMmRE3lmlkzeXjrVrqTKfq7e/jFyYu5ds4s
    QtEYN694ht5ojAd37mYwXzl2iyJ/3NJGqVnh89OmUFlSzLy68fQN+elJZ3jg1TUEE0lESSKeyfD7
    C8/n0395lP5QaAyt8vH97ex64CHaBgYIDg7R5fHwrf4BQuk0yUwGQZZpa9tGSlVxAUUuJ5fNbCUS
    i6GqKpooEgoEKC8q4uG27fz2rQ3EE0lOSGc4EAgwGAiyVpJ44UA7v0LglEkNTK0oxySKTC09fm9h
    QRCwWq1EIhEKCgo+Qu+HCcRvtHfw+UcepaaggBduvAGzoozIqzhGsXgMXafRW8jGb/8XpU4H47/7
    AzLxOFfMn8t18+eNfTCiyLcWL6TAW4jHYsUwDDKZDA6Xiw2dXTy04llsJgXJZGJTZydus0JTSQnf
    +MQSCmx2Lrr7XnyRCH/evpOfLzudH528mJteeoV0MMin5s9l/549nNzUyHhvIf+zZi1VLifXT27k
    lY4ODvoDaBrctPYNbtm4mcunt1BWUkx7NIY6XBTTDc6bPo3/9+jjlFRUsDMQYiiVZH51FbMqK3Om
    6ckUrdVV/Nfpp2BXFNrb2pg7oY6V23fk7t75TWt/IEBGELhq9kyKFp7E3t5ebn39zZzEjyxzWuMk
    JE0bs1j0fNFvWMe5tKyMg52dRCSJeDgCZgWrojAYCIIsU2C3E85mufr+B3n6U9dxcuMkktks6WSS
    dDp9VLePY6XVLpeLIf8Hx/3wIxAfT/qcydDhD2CSJMx5kkc8FsPucIzhVw/5/dSUlTGcvH3zjNOI
    p9PMrjkynfP5fEwaleb5/f5cUcUwOHF8LZ+9+EJaKiv5zyee5IDfz4s33sB4q4Xx48ZxzX0P4otG
    AQikUhS4nCytr2PyuFoG0hle27efId8gi5oa+dhJC4hnMowvK0NLp5k9rpaDgZzX0UAkwkBQxzLX
    yl1nncF1z7xAn9+fO0UNnatmtlKoKKQlmZ+/soZOf4DOUJhHduzCbZL5/mlLubS5kcLCQrz/+U0y
    mSx2xcS5M1pwyDKPtG3LSfKIIl2BAPe98RYvfuFGburqHnG5yGYypDWNXyw7HS0eh1Gb4rCUTTw/
    d+yw2zm9tobwCbN4aOs2DMNgQeMkROCSGS08sqWNl7du4441r7HXH+CG+XOR8pvjCJ3yOIHs9Xj4
    4YpnKXA6OLO5ibr/w4D+UID48pkzuHzmDPYODiF85vM47Ha2ffMbuCSJlJpjQSWi0SNSsO+ccdrR
    03Ofj5JRxRZVVZFlGUEQEASBBXV1LKirwxeNccuq1ciiiMduRxYFUFWCkQhoGuc1N/H1BXNJaTrj
    yspodbm4Y8MmFvzPL6nweulZMJ/GVIpvLVlEZXU1K1et4om27WAIXNoylbkT62nv6aHWYsbIZPl4
    cyPr9u9n3ZCfFDI1N/+UGVWV1Hi9fGN2K08f7CCUzbL2YDuKw47bbiOU33zSWZW0phGLpqksKGC3
    b5BMVoW84qXbZkPPi+PNrq3hra4u9vXnJIk0TeeNzm6KPW6qw2FmNzSMeV52u52MYdDw/R+BzYbL
    bmN2bQ2bunuYVzeeu664lEJF4XOLFnL/WxtYfu/9PLNnHzfMPYEMEE+mMFssmN5Fq0iQJL79xFO4
    XE7GF3o/AvGHpiKta9hsNgpsViRBJJpOc+19D7KuvZ1HP3Etc9+hPaHrOqFQaAyAAWKx2Ijfz+jw
    WMw8/LEriSUSFFktfOL+h1jYMBGzYgJZJhSLMauujkQ2i5KfVy1xOqkrKqI6TyE9cPAgVreb59u2
    81x3L+lUCiSJh19bx9Wt06kbV8t5d9+L1ePhB0sWUWi3Yw1FcNsUNF1jU18/GU3jqzNa+MX6jezL
    24pGM1kGkkmikZy306/OP4fd/QMMJFNMKS0lEE8g5oX0Wqur+O6pS+kJBPCYzXx5ySIKbFZ+8uzz
    tFRWcl7LVL7817/h6+nla+cso7W29ohnoSgKWKyIJhM1hYXMr63hrY5OVu3Zw+4hP7MrKxBTKSyy
    RFNZKXZF4X9Wv8TXfnsrX/7YFXxt/lwq8hrSxxsz6sbhsVpxmRX+L8eHCsQlTgc/OfdsLCYTBTYr
    GVXDH4/TF46gv0Ompus60Wj0iKGJbDaLJElHjD0OT0pNHVdLJhKhKxRi5Rtv0RuLM6m4CFSVg6EQ
    giCgaxouq5V17R10hUJ8YuZ0zm1uYsjnI5pKcfOra1l94CB9oRCy2UyVy4VcWko0FKSyrIyGmhoq
    CjzctbWNHXv2ctL0Fr55+qmEQ2Ha+vooUBRcFjMOkwlXfrjCoSi5FFUUiEajXNzcRLZxEvds3UZX
    MMjlM2fw3M5dhJJJTqqro7WinIWjjOOvPWEOy6dMJpVOkxAELpg6mQOVFcyZUEc6kyGtqmNaPSZR
    4qazzqAnkeS2l19lW28fV89q5d431/PZPz/CH5dfyeyKchaNq2X1Fz6Lbhg8v2s35c2N1JcUY7VY
    3lX7yDAMNn7ja7k23zEcL/6vxIea7JHRNDZ2dROIx5k7fhwFhy2QaDrFuoMdmCWJyUVeClyuMa2L
    YaCaTCbM+aH4EXZkJoOkKHz/mefY1dtLvdfLQDxBrdtFc3kZYU3Da7PRZFaQnS4y6RQ/ffU17nx5
    DZPKS3ntM58Ek8KO9nb+f3tnHh9Vfe/995kzc2bfJ8uEhJCVJWGVTZBVWcQdNxQuVq1a9Upbe62t
    eq3P1brcar23vV301VqrtYp6FYsgiIAguLIatkQhISFkmZlMZt9nnj9mEhMIS1ie+1yYz5/JmUnm
    zO9zvr/fd/l87lixitp2B8hk6CWJx+fO5oqxF3BwVw2jRo1i64F6frRiFe6Anyp7PndeNJnrRo3k
    cGMjBbk5tDQfJsdu560vvkTIJIjUKhWFWg2KcJic3FxkMhnuUIhxv3sBfyDIx/cvYem2Hfxh3XqQ
    yVh72y1MGV5Nm8OB3W5HzBiQdXR0YDSZ6PT7kZQqdBoVYjJFe3s7BQMGHLEViuOIRHhu7XpMKhW3
    DRtM5W9+T6HJxOOXzWPG4ApksVh6YEOjwekP0OzxYDcYyNXraG9vJycn56TPxl0PX7/fj0ajOaec
    Ds/bSHwkJFFkYmbwvC8c7Ojk5pdfYaDZzPK7vt9N4O/8s9Nzq12ZUwEY9sunMarUPHjJDOLJFM98
    tI5gIMC0inL+dt01dAaD6NVqzCYTsViMprY2/rzzU97cspWKHBuXlJcyqqgQa14+jY2NyLQ6kl2a
    XOEwQVGkJD+PUqOeWG4uJq2WITk2dtfVobXZuGnUSEYVFLD7cAtxn58PD7fy1mefc8u0KYzMyyU3
    x4aAgFKpJBqL4e7spLOzk+IBA1DK5fxl0U14QyEKjAZmlBYjMY1yi5WiHBsutxulKCJmppUVmWju
    9/mw5+Tg9XiQIyAq5CQEgWU7duIIhrj9wgn8dsMn/Gb9x9xy4QSevvJyIvE4QiLBOwuuJ6nR8uN3
    30O/dh1v3/49LHI5fp8Pm16PTfed7rPVaiXs96HWG076O5bJZMjlciKRSJbE5yNiiQQudyeS2Nsc
    WwBm/+6PWDUa7p8+lXE9NLn21h9EodPiDoUJx2LEEwnyrRbunDoFjUaNzmhEEAR0ej0OhwO9wciB
    tlbaHQ7uuHACw80mFAoF9W3t6CWJTqeLQDgC0SgbHvwX7KkkA4uLCYTCaNRqBFn6bF9dXo5FrUKt
    UnPjy6+w/VAz7y5cwOb6BlbW7KKqtIRJE8YRjcYQM8Lo8kzbqEqh4Bdr1rKz+TA3TRjHbRMn8MvV
    a/jl6jVMKhnEE7NnIZMkgh4P/77pU178cgspwPOrp9Cq1Xi8XhKJBAajEVdHB2/V7Obr5mZW1NbR
    eLiFhWPHcKizkwNNh2gcnE56KeVykMsZV1HBl9/up8HlQqlQEEukXQ0dbjfPfbCaEquFK6qrMGs0
    iKLIkvfexxkKs2T6VGZUlJ/U96hSqfBnatbnI5HPaxIPzs3hq4d/lrZYOUIVYs2u3eSazdw+aWKv
    nz+94Hq0koJBVgtPffgR0UCAu+fO5uZxF9DS0oJRK+ELBiEcIZ5MoZSLVNntrEztoNBq5d4VK3F1
    epheUUZHOIJVoyGSTIAkMe3Z5ym0WWl67BGCoRAbm5r5bNOntATD7Dp0CJJJpg8dgiSKpFIpPm9K
    y+xOHjqYMouZwrw8nnxnGX/aup3Zo0bwx+uvpaWlBZUosv1wC+u/3c8FJYMyESwtJZRIpYglE2gz
    RwaZXJ62bcnUo59c+zGdXg/3zZhGud2OwWTi2Q0b2X+4BY1WC0olyVSKK6qrsGi1jC3qLaej0+kY
    nJ/PR3ffRTAYwKJWkUgmScrlPPbaG4ysrup11PnyUDO1be1cO2pkv6IxpPursyQ+z6CRJMYeoeHU
    hXtmTMesUVN0RKLrwUtmpjPWGVWJtOBbBUGfD6UkodFo+On7K3ljyzZ+PmcW1wwayJTSEhovHE9V
    Xg4VZjMaIBCJ8XVDIyNKihmRn8+hzk46gkFy1GpCfj/hQIAVu3bx+s4aSKUwGfToFBIVOTbun/kD
    fvLaGzz9wYdUlwxi5z13EJeU+GMxajs6qHe7+fpQM6FwmFafj0q7nWmFAxigVjE5I0B3z5TJTCop
    4advv8MDK1bxq+vmo9dqmVVaQnVBAU6nk2a3m+c3bCSSSHD18GoG2WwIokhlTg4GpYrf3zCfiZlu
    tqnlZUwtLzs66SII5FrM5BoNdAYDvPbFV4SBy6qGUjComMrcnF5TSDePvQCHz8eQfnRyAWkf44wW
    tfC/1N0wS+IziBTwuxuOP2iuEEXunHwhs4cMJpFM8tmBeix6PRarFW84gisUwtnpQSGKTC8p5uLi
    IjQmMy9dfQUHW1p4avvXkEpycXkZv77mShocTrYfakYSBNrb2hAEGZdVDWNjUzPNDgdzKiv59dxZ
    5BUUIMpkjC8uoiUWo1CnZcX+ehKiyKpvvuW/99UiUygwqtU8uXY9b361haWLF/HDaVMQUil0mR5y
    o0pNrl7Hlwfq2d7axs/nzkat1XBxdRWxcBhRoWDvwUbunDgemVxOZVERXp8Pk9nMK4sXkkgkkUgR
    CgZRnyCjHEvBlwcbOej1ce87y0h2eshdci+vf++fGGS1UNjjQdn1kOx3hjYzpRaPx0+owZYl8fmQ
    sj+Ja5RyOVcNrwbg7qVvsXrPXiaXlvBqaQl3jruAeeWlTKkahk0QUGq17G1r56+bVlJ76BD/PH4s
    DR1u8AfQa7UgVzDIbmeQ3Y7D52PBi3+myGzmueuvpbiggC9r6/j84EHuee99fjRnFtPLy7h02FDm
    XziRH73xFnctW067z8cVw6sRBBmCICCJIuu++Za6vbV4QyEiBj0mk4k2p5P8/Px0Bluh4KG5s8mz
    WChQqxHkCvY7HMhEEaMoMsBm5d8vn4dMoUj7TsVidLrd2HrU20OhUHcd/VjwxaI8t+ETVuzaze2T
    L0SZSvHge8sZZs/jV1dfxUCzcNIP1+NdaTAY6OjowGazZUmcRf9Q09JK/cEmTJmINNyez0WDK1m1
    r46XN3/KxNISZg4ZzKtffMXh5sPcM30aby9agKTVYTf0zsQ6A0HWbd2OrsDOj71eLiot4aLSEq78
    9X+yfMdOZlQNY3p5GUq5HLVCwWuff4EmY/05raKcz+obAPjFvDnk6vR8MrwKu0aNxWLB6XQSDoe7
    /9Z/bdjEsq9r0CiV7D7UTFKU8VXDQQ57fbQ+8Rjujo60L1MiQSgYRKvTdSeRukirVqsJh8N4vV4M
    hr6zyolEkhavF1rbuGfKRRSplPx+5Wpcfj8HMh1psYz8jiAIGAyG7mha297OZ/UHMaiUvaSajnU2
    ViqVhMNhVOeR4keWxGcAr86/CvGGaymw2/F0dhKNxUAmY2tTE0u3bSeQTHLD2AsYV1RInSRh0+so
    Vqmw9BExrFoNi2ZOY6DVRn6GFI+u+IDldd9QWVHRfe7sKu9LOi2ReIwHLpmJWq7AGQhgUKkYU1hI
    LBJhUnkZCklCzCTDepKwxetlv8MJgkBN82GKLWY6giF8rg5kosjnLa089tJfWTRxPLeMGU0gEECr
    1eJ2u0kkEt0lOZVKRRJ466st7HI4uXbkiF7m3wVGI5/95Ed4vN706KdaTcMzT2DPz0c8wgwulUrh
    8/kIhUJoJIl/bN3GQ6vWUGqznpDEXWfjtra2LImz6B+cbW2MGzeOaDRKMBTqVp9QKRSo1WqUokiB
    0cDrC28kHA5jtuXQnrHUPBI5Wi3/etEkNGo1eT0sSxAEZlSUM7rwuwYLSRS5fsRwXvv0c5qcLqZW
    VDC7ohyjTseyr2tYvrOGq0ZUU6FSovf5kMlkJJPfWZguHj+WccVFfNPupN7l4srh1UwtL0tPQAGO
    UJgtu/cwoqgQ1eRJEIvhdrvR6/VHdbcpFAqW7dnH39d8xACjsReJBSH9v+aYzTidTlQqFWqFAqmP
    mWBBEDAajd3z3zazBb1Khe4kp5y6iHw+iQdkSXyaCAQCAMQTCb5paUGjVBFPJlECP5k5nZ/MnE4i
    mLkmBZF4oivkHP0w8AcIRiMEkkmsPcyv548cgU0hp+qIDql4LMYoqxX51CmoBIFDnZ1cXF7KTRPG
    c9fSt3n/443YTEaKS4rT5txGI7X7D2CNxVBKEnOGDmHO0CHHPH/m6/UMrqjAqFJxyO2myGxGn7EE
    FUUx7WGUkTsSBIGh+XlUDRtKgdFw3HOr1+tFo9GccNubAm6dOJ5bJ47vjtInk3nWaDR43G44T0h8
    zmtsnW007N+PSqMhoVAw+w8vcmFpKY/Nm0thRjrXHwiyp6WFFDC0wE4kECAnJ4fOzk50Ol2vuuaS
    t99hS2MTo8wmnrrycow9JnM8LhfRWAyt0YhapcLr9eLzegkDr9fs5ouGg6zY8TWEQ7y35F4643E2
    NRxkZnkZ5Wo1WklibUsrL27cxK3jLuCHM6enE1YnQDgW44F3lrG5/iD/cf18ppaVEovF8Pl8yAQB
    0ykM57vdbjQaDX6/H+tZmD5KplKsrNnFzubDTKkoZ1JpCXKZ7Jxdg9lIfIqIJ5N8ceAAuw7UM6u6
    Go1CwZ6aXURT4I9+Z+7d4nLy8w9W4wmFeGnRTRRk5p11Ol23UF8X1tbWsafuG9TDqzEeIYguV6lw
    +3xIajUupxNRFPF6vQTVav5t9RriDieLL51Drl7HuMpK7AY9iydO4I2vtvIva9YyUKNGazJRs3Mn
    9UMqkaWSJ/eQ6ujg/b37aNi9l+ZZM7u3zhaLhR379/Px9h0EYnEenjPrpO+dyWSira2tO4qfaSSS
    Sf7y1VbeWb2G26++knHFA7MkzqKPbXQ0yrPrN/KPml18WFzMlNwcHl90M+X5eb22k9F4gq1Nh/AE
    g/gjUYQMiQVB6HU+BVj+gzv42xdbaHQ6WPD6UmYMHsz3J0+iw+kgEol0b0NFUUTIRMFCq5W/Ll6I
    Oxjk2pEjMWvUvc6a+5xONuzdR7HZxI5HH+b748eRr9OyYs8+Dvn9jC8uZnTRse07S6021i25l8aW
    FkaX9W7mcCeSPLtuA80uV79I3JWBDoVC3cmyMwmZIDBv2FCEZJJpZed2FM6S+DQQiURx+HwkvT7i
    ySSSXM6tY8eg1+sx9Djn5ei0OJ9+gkQyiVIux+V0HvM9i81m3t65k5qMr3EkFmPOoIEMKizsdnPU
    aDTd53BBEDAoldw0ZjSpzOI9EreMH8vk0hJMahUBl4vRA4uQJRPc9+4/eHfbdv7PVZcfl8SSXKTE
    YqEkI+Xb87MZ1WouGFiERa3q5SGkVjoAAAnlSURBVOF0MlCpVHR2dp60LE9/IMpk3DpxPAtGjSAZ
    j2dJnMUxokk0wpvXXUPB/T8EwOfzYdDru0s/8WSSWCJBOLOI5DIZq/fW8u7WbQwfWMRdkybiCASQ
    dDq0ktTdxKDOdFsN0GiYVFrKgIJ0llcul6PVatHpdDQ0NqJWqynLtDwKgnDMJogSq5USq5VPGxp4
    q2YXc1UqhlitKEih1WhQ96O7SSaT9UoujSkq5L07b4dEgn21tZSVlZ10t5RMJsNkMuF2u4/bKHI6
    0VirVqfdGc/xLq4sifsBh99Pi8dLKhHHplCgNxq7J2cCgQAWi4VYxsFww7f7+dOnn6OWCby0eBEA
    6+rqeGHDRsZWVmDT6Vjw5NPMnzuHv/7TQnTKNJFfvfnG9Fiiz0tJYRGKTBSJx+N4PR60JhNv7qwh
    IUncZ7ORn7E2ORF+/Pa7fFlbR0M0zn9cOY+l37+tO3KeqBOqCzabjZaWFux2+xGhTyQnJ4fGxkby
    8/NPfnssytnndLHN4WRW1bCz4gesUChIJZPn9LrMkrgfWLptBw8vX8GQ3BxeW7gAi0KBKIrdNclA
    INCt47W3tY03PtmExmTipczrZ1ZWkLzsUkptVmw6DSXV1QwfYEcupnt+w+EwpTYbgiDQEouR7LEN
    lMvlaNVqmlpbeaFmN+1NTVw2ZvRJk3he1TCKTCZG2vNIJFPEIpHu8k5/xgV0Ol2fpSGTydQtopBK
    pU4qujoCAe7/4EO+/moLtX/4LZW5OWf8O9NqtXy2rxaFRkOR+Ts/pyyJz1ekUijlchSiiFwuRyZL
    9yl7PB7sdjsul6v70gmDBvLwNVeh6kHEnnVZbzjMih/cgUWrRZZMEs4oPsoliUgkgj8QwByPd58Z
    k6kUh31+AtEoD0yZjCfjsHCyeOCSGcTjCRKRKJFoFLl0arpUx+qIEkURjUZDIpHA7/cjCMIJI7JC
    FCm3WdlTWIAkP3ti8M9s+ISa1lYenjOL2yZOyJL4fMa8shLmLLkHhSjHqk2rQ3bZwxxZbh9TVMiI
    goK0EbjHg95g6NWoYFCpMOSrSCQS3U0TPaWBRFHsnukFaAsE+N7b73DY40WSi0RicR6/5qqT/t81
    ComUAup9PpocDgbb7eSd4tieyWTqs1dakiSCwWB3n3ZP76u+kKfX8ccFN/Bvsy8m9yw2Zuxpa+fA
    gQZaPN7sdvp8htfrpTA3l87OTiwWM4FAALXRiNPpxGq1HkViUZAhytNRWJZKdftC9UQsFiMQCKDX
    64/S9tLpdASDQXQZ7exUKkWr30+bx5M+M2fO3v1BMBrloRWrWLppMy/fsohbpl50SvdCFMWjymNd
    W365XE48HicvL4+Wlhb0en13C+VR7yOTkaPTkqPT4nA40Jylfuc9j/yMVCqFeI5mqbMkPpnFHwoR
    iMUIJZNIkoRMEIjE43SGw4QyZDpePJMkiWg0SiwaRZHZxsbjcYLBIMaMnM+R6NnnHEskeOT9lYwe
    UMDt48fS1OmlMxIhGgoh9aNhQgDUkgLUal7ZtoOWcISfzb6k/4tGLkepVPY5gihJUrevVX5+Pi6X
    i1A8jsloRBLFPstgPT+v7CwQ7VwvMcmyFD0xmjs6uPGV1xjx9LOsO9jIvrY2/rj5M4oefpTffrmF
    eDKJPxolFIvTVw+rIAhoVCrq2trZ1nSIZrebUDCI4YgtNqQlZiKRSK9taDyZ5G8rV7F6Xx3jBg3i
    5ZtuYNkdt9Hh8fTrc6gVCp6ffzXPXzefdbV1/H7TZk6151aWKZv1RXAgXZ+Vy1FqtSx67Q0eem95
    ehzxGLBarbS1tWUXWzYSn+k8VopQMJAe34tGCScS3PH3pRjVKnJ1OoKxOOFojHpXB4+vWo1Jklhy
    8Uwqco4eMRQVCp74aC3Lanbz0CUz+Nd5lx7zb6ZSKdSZGmdXJLnvuvmYNRoG5+cTiEQQE4l+Ry1B
    EDCp1VwyuIKHL7sUq1bLqQrZKJVKlu/YyZZDzUwfXNlL1E6v1/Pulq1EZTIGWays3fQp3wyuZOGY
    UQw4xta663U9hyqyyJL4tBEOhxEEGXazmWfnzUWt0/Gz5Suob3cwf9RInpgzi5K8PJq9Xv62ag3q
    /DyuGTO6TxIDqBQSaoUCnVpzwoeHJEnEo1FkGdmZ31w3v9c1HR0dCKdY/6y223nicvtp359/7N3H
    n95fievqK49Spnzm441sbTrEL6+4jIcWL2SA0UCewYDb7e5lp9OzRq3T6Whtbc2SOEviMwe/309O
    Trp2OarAjsVq5YWrLsdoNmPWaIgGAhj0eowaNX9ecg8GlYoqe/4x3++5a67iicvn9Wpd7Ov8Jspk
    RJJJvu1w41MdZmThgKOus5jNrNm+nbxYjPL8PDSK//dWJiVWCwPLyhjUhz1OodmMOxQmT6/np5fM
    JJZIIJfJEIDNdXXUOJwUmUxcVjWs1+sMBkPWSLyfyI4iHgNtbW3k9fAGcjqd2Gy2Pn9utVq7u59O
    R2exoaODAw4nVkkiKAjc+NLLNLW1k3rhd31/eXffR4HVwquLFzGzsuJ/5LjRS2i/B5KZZdXX7576
    8CMeevc9Lq4axkf/fPdROQGPx9MtrJBFNrHVL3zjcPDI+yu59ZXXyMvJOeo86XK5jpp/TfVcrKf5
    9//wyWYufvIZnvx4I3KZjFgi0e0h3Bc0Gg1GtTotnfs/EQEy0jp9ZteP8zudUonVaMTSx7Y5Cexs
    c/CnTZuPmwjLIrud7hMHO9w8v/5jgm3t/GXxwiPDDslU6qyKkw+yWhhaWcHwAjtra+todbq4deqU
    Y16/66EHkYsycs7wKN9Zjd7AfdOmcN+09Ofq9Hgw9sjSh2Ixnl6/ng/XbeDdRx/i6hHDswszS+KT
    R5nNxqNz5+DKjPpBesA8GI9T39pG9TFsRc6UWPni8eO4btRIbFot97753+D1HXfKqMT6v2/LeeSd
    UmeE+7oaYeQyGaMGDGD/8OqjXDmyyJ6JTwn1Lhe/WLmKT/bVseexR44iVUdHB0ajsVfH1ZnArsMt
    7G1rY2heHtUF9nP6HkciEaLRKHq9nmQqRbvPjzccptBkRCNJ2UWYjcSnB38kyqb99TQcOEC8j5KO
    xWLB7XJhPsNaUdUF9nOevF3oGvLweDwYjUbyDXryDfrs4stG4jODQDTKntZWOoMhZlZWHNV/+43T
    ye76BioHFjGsn073WfRGNBolGAweZfSeRZbEZxW3vPp3Xtm0mXmjR7LiB3dmb8gZIHIgFEKuUiGX
    yVApFAjZ23JcZEtMp4nBebmMryin2m7P3owzAEmSqHW6uPrFP/ObDRsJn8K0VvZMnEW/sGTaFO6c
    fGGfbgZZnBpaAwHWrV1PCrj7osn90gHLbqezyOL/A7T7/Hyyfz/FFgujCgec86OEp4v/C6h6ZFqo
    ho+QAAAAAElFTkSuQmCC
    '''\
    .replace('\n', '').replace(' ', '')

_default_css = u'''
                %s
                .code {
                    font-family: monospace;
                }
                .box {
                    border-style: solid;
                    border-width: 1px 1px 1px 5px;
                    margin: 0.75em 2em 3em 2em;
                    padding: 12px 1.75em 12px 2.75em;
                    color: #444444;
                }
                .quotebox {
                    border-color: #6FA6A9;
                    background-color: #E4ECEC;
                }
                .codebox {
                    border-color: #A9C98B;
                    background-color: #EFF4E9;
                    font-family: monospace;
                }
                .small {
                    font-size: small;
                }
                body {
                    margin: 6ex 23.61%% 18ex 14.59%%;
                    width: 61.8%%;
                    color: #646567;
                    font-family: "HelveticaNeueLT Pro", "Helvetica Neue", Helvetica, Arial, sans-serif;
                }
                a:link, a:active, a:visited {
                    color: #646567;
                    text-decoration: underline;
                    text-decoration-style: dotted;
                }
                h1 {
                    color: #6FA6A9;
                }
                h2.omnipath, a.omnipath:link, a.omnipath:active, a.omnipath:visited {
                    color: #6EA945;
                    text-decoration: none;
                }
                h2.base, a.base:link, a.base:active, a.base:visited {
                    color: #646567;
                    text-decoration: none;
                }
                .align-middle {
                    display: table-cell;
                    vertical-align: middle;
                }
                .nav {
                    padding: 12px 7px 7px 7px;
                    color: #FFFFFF;
                    background-color: #6EA945;
                    cursor: pointer;
                    margin: 5px 5px 5px 5px;
                    display: inline-block;
                    font-weight: 600;
                    text-align: center;
                    font-size: larger;
                }
                .nav:hover {
                    background-color: #FCCC06;
                    /*margin-bottom: 5px;*/
                }
                .nav a {
                    text-decoration: none!important;
                    font-weight: heavy;
                }
                a.topmenu {
                    text-decoration: none!important;
                }
                #logos {
                    display: table;
                    width: 100%%;
                }
                .logo {
                    margin: 10px 1.8%% 10px 0;
                }
                html {
                    padding-bottom: 1024px;
                }
                .paper-link {
                    font-size: larger;
                    color: #FFFFFF;
                    background-color: #007B7F;
                    padding: 12px 25px 7px 25px;
                    display: inline-block;
                }''' % _fonts

_header = u'''<!DOCTYPE html>
    <html lang="en">
        <head>
            <meta http-equiv="X-UA-Compatible" content="IE=edge">
            <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
            <meta name="dc.language" content="en">
            <meta name="viewport" content="width=device-width, 
            initial-scale=1.0">
            <link rel="icon" type="image/png" href="%s" />
            <style type="text/css">%s
            </style>
            <title>%s</title>
        </head>
        <body>
        '''

_footer = u'''
        <br>
        <p> <a href="http://www.ebi.ac.uk/~denes">Dénes Türei, 2017.</a> 
            Feedback: omnipath@googlegroups.com </p>
        <p>
            <a href="https://validator.w3.org/check/referer">
                <img 
                    src="%s"
                     alt="Valid HTML5" />
            </a>
            <a href="https://jigsaw.w3.org/css-validator/check/referer">
                <img 
                    src="%s"
                     alt="Valid CSS3" />
            </a>
        </p>
        \t</body>\n
    </html>\n''' % (_valid_html5, _valid_css3)


def get_header(title=''):
    return _header % (_favicon, _default_css, 'OmniPath :: %s' % title)


def default_template(content,
                     page_title,
                     title='',
                     logos_bottom=False,
                     above_title=''):

    logos = u'\t<div id="logos">\n\t\t'\
            '<img class="align-middle logo" src="%s" alt="EMBL-EBI, Hinxton"/>\n\t\t'\
            '<img class="align-middle logo" src="%s" alt="JRC-COMBINE, RWTH Aachen"/>\n\t\t'\
            '<img class="align-middle logo" src="%s" alt="Earlham Institure, Norwich"/>\n\t'\
            '</div>\n' % (_ebi_logo, _rwth_aachen_logo, _ei_logo)

    page = u'%s\n%s\t<h1>%s</h1>\n\t%s<br>\n%s\n%s\n' % \
        (
            get_header(title),
            above_title if logos_bottom else logos,
            page_title,
            content,
            logos if logos_bottom else '',
            _footer
        )

    return bs4.BeautifulSoup(page, 'lxml', from_encoding='utf-8')\
        .prettify().encode('utf-8')


def main_page():
    notebooks = []
    notebook_dir = ['pypath-docs', 'notebooks']

    notebook_names = [
        ('pypath_build_networks.html', 'Quick start'),
        ('intro.html', 'Introduction (directions, enzyme-substrate, subnetworks)'),
        ('pypath_mapping.html', 'ID conversion'),
        ('pathway_extraction.html', 'Pathway annotations'),
        ('node_neighbourhood.html', 'Extracting signaling networks'),
        ('TF_location.html', 'Finding transcription factors')
    ]

    if os.path.exists(os.path.join(*notebook_dir)):
        notebooks = os.listdir(os.path.join(*notebook_dir))

    above_title = '<p><img src="%s" title="Welcome to OmniPath!"></p>' % \
        _omnipath_logo

    with open(os.path.join(common.ROOT, 'data', 'main.html'), 'r') as f:
        doc = f.read()
    
    tut = ''
    
    tut += '<h3>Check out our tutorials:</h3>'
    for nb, name in notebook_names:
        if nb in notebooks:
            tut += '<p><a href="%s" title="%s">%s</a></p>' % (
                'http://pypath.omnipathdb.org/notebooks/%s' % nb, name, name)
    tut += '<p>Special thanks for Luis Tobalina for providing some of the tutorials.</p>'
    
    doc = doc.replace('<!--TUTORIALS-->', tut)

    return default_template(
        doc,
        'OmniPath: literature curated human signaling pathways',
        'literature curated human signaling pathways',
        logos_bottom=True,
        above_title=above_title)
